/*------------------------------------------------------------------------
--                                                                      --
--         H E L I O S   I N P U T / O U T P U T   S E R V E R          --
--         ---------------------------------------------------          --
--                                                                      --
--              Copyright (C) 1987, Perihelion Software Ltd.            --
--                         All Rights Reserved.                         --
--                                                                      --
--  cofuns.c                                                            --
--                                                                      --
--          Extra coroutine library functions, and the linked           --
--                                                                      --
--          lists library.                                              --
--                                                                      --
--  Author:  BLV 8/10/87                                                --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89  Copyright (C) 1987, Perihelion Software Ltd.       */

/**
*** I do not intend to document this code, since it should work if you have
*** got the low level coroutine library right. Anyway, it is all simple.
**/

#define CofunsModule
#include "helios.h"

/**
*** The low level coroutine library, written in assembler language
**/
PUBLIC  WORD fn( InitCo,   (void));
PUBLIC  WORD fn( CallCo,   (ptr, WORD));
PUBLIC  ptr  fn( WaitCo,   (WORD));
PUBLIC  WORD fn( DeleteCo, (ptr));
PUBLIC  ptr  fn( CreateCo, (VoidFnPtr, WORD));
PUBLIC  WORD CurrentCo;
/**
*** Coroutine stack sizes : these can be configured
**/
WORD Stacksize;
#define CoDeleted (123456L)

WORD InitColib()
{ Stacksize = get_int_config("Coroutine_Stack");
  if (Stacksize eq Invalid_config) Stacksize = 5000L;
  return(InitCo()); 
}

void TidyColib()
{ 
  free((BYTE *)CurrentCo);
}

void Suspend()
{ (void) WaitCo(FALSE);
}

void Seppuku()
{ (void) WaitCo(CoDeleted);
}

PRIVATE VoidFnPtr co_function;
PRIVATE Conode    *new_conode;

PRIVATE void new_coroutine()
{ VoidFnPtr my_function = co_function;
  Conode    *my_conode  = new_conode;

  (void) WaitCo(FALSE);
  (*my_function)(my_conode);
}

Conode *NewCo(fun)       /* create a new coroutine and put it on the */
VoidFnPtr fun;           /* waiting list */
{
  co_function = fun;
  new_conode  = (Conode *) malloc(sizeof(Conode));

  if (new_conode ne (Conode *) NULL)
     { clear_bytes((char *) new_conode, sizeof(Conode));
       new_conode->cobase = CreateCo(func(new_coroutine), Stacksize);
       if (!(new_conode->cobase))
         { free(new_conode); return(FALSE); }
       CallCo(new_conode->cobase, 0L);  /* first CallCo loses argument, fun  */
       return(new_conode);              /* must always WaitCo near beginning */
     }
  return(FALSE);
} 

                      /* call a coroutine, check to see if deleted */
void StartCo(conode)           
Conode *conode;
{
  if (CallCo( conode->cobase, (word) conode) eq CoDeleted)
   { DeleteCo( conode->cobase);
     free(Remove(&(conode->node)));
   }
}

/*------------------------------------------------------------------------
--                                                                      --
--  Lists                                                               --
--                                                                      --
--      Usual linked list library, but containing only the              --
--      functions which I can be bothered to use.                       --
------------------------------------------------------------------------*/

/* create and initialise a list header, returning 0 to indicate failure */
List *MakeHeader()
{ List *newptr;

  newptr = (List *) malloc( sizeof(List) );
  if (newptr ne NULL)
    InitList(newptr);

  return(newptr);
}

/* initialise a list header to the empty list */
void InitList(listptr)
List *listptr;
{ listptr->head  = (Node *) &(listptr->earth);
  listptr->earth = NULL;
  listptr->tail  = (Node *) listptr;
} 

/* add a node to the beginning of the list */
Node *AddHead(node, header)
Node *node;
List *header;
{ node->next = header->head;
  node->prev = (Node *)header;
  header->head = node;
  (node->next)->prev = node;
  return(node);
}

Node *AddTail(node, header)
Node *node;
List *header;
{ node->prev = header->tail;
  header->tail = node;
  node->next = (Node *) &(header->earth);
  (node->prev)->next = node;
  return(node);
}

Node *listRemove(node)
Node *node;
{ (node->prev)->next = node->next;
  (node->next)->prev = node->prev;
  return(node);
}

Node *NextNode(node)
Node *node;
{ return(node->next);
}

word TstList(header)
List *header;
{ if ((header->head)->next eq NULL) return(FALSE);
  return(TRUE);
}

#if (ANSI_prototypes && !AMIGA)
void WalkList(list, fun, ...)
List *list;
VoidFnPtr fun;
{ Node *node, *node2;
  WORD arg1, arg2;
  va_list args;

  va_start(args, fun);
  arg1 = va_arg(args, WORD);
  arg2 = va_arg(args, WORD);
  va_end(args);

  for (node=list->head; node->next ne NULL; node=node2)
    { node2 = node->next;
      (*fun)(node, arg1, arg2);
    }
}

word Wander(list, fun, ...)
List *list;
WordFnPtr fun;
{ Node *node, *node2;
  word result, arg1, arg2;
  va_list args;

  va_start(args, fun);
  arg1 = va_arg(args, WORD);
  arg2 = va_arg(args, WORD);
  va_end(args);

  result = FALSE;

  for (node=list->head; (node->next ne NULL) && !result;
       node = node2)
    { node2 = node->next;
      result = (*fun)(node, arg1, arg2);
    }

  return(result);
}
#else
void WalkList(list, fun, arg1, arg2)
List *list;
VoidFnPtr fun;
WORD arg1, arg2;
{ Node *node, *node2;

  for (node=list->head; node->next ne NULL; node=node2)
    { node2 = node->next;
      (*fun)(node, arg1, arg2);
    }
}

word Wander(list, fun, arg1, arg2)
List *list;
WordFnPtr fun;
WORD arg1, arg2;
{ Node *node, *node2;
  word result;
  result = FALSE;

  for (node=list->head; (node->next ne NULL) && !result;
       node = node2)
    { node2 = node->next;
      result = (*fun)(node, arg1, arg2);
    }

  return(result);
}

#endif /* ANSI_prototypes */

                /* This routine just frees every node in the list */
void FreeList(list)
List *list;
{ Node *node, *node2;

  for (node = list->head; node->next ne NULL; node = node2)
    { node2 = node->next;
      free(Remove(node));
    }
}

                          /* this routine is needed by the debugger */
void PreInsert(next, node)
Node *next, *node;
{ node->next = next;
  node->prev = next->prev;
  next->prev = node;
  node->prev->next = node;
}

          /* This routine is used to maintain the order of nodes in WaitingCo */
          /* It puts the node just after the list node given                  */
void PostInsert(node, before)
Node *node, *before;
{ node->next       = before->next;
  node->prev       = before;
  before->next     = node;
  node->next->prev = node;
}

          /* This routine returns the number of nodes in a list. */
WORD ListLength(header)
List *header;
{ Node *node;
  WORD result = 0L;

  for (node = header->head; node->next ne (Node *) NULL; node = node->next)
    result++;

  return(result);
}


