/*------------------------------------------------------------------------
--                                                                      --
--          H E L I O S   I N P U T / O U T P U T   S E R V E R	        --
--          ---------------------------------------------------      	--
--                                                                      --
--             Copyright (C) 1987, Perihelion Software Ltd.             --
--                        All Rights Reserved.                          --
--                                                                      --
--      Debugopt.h                                                      --
--                                                                      --
--  Author:  BLV 10/12/87                                               --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89	 Copyright (C) 1987, Perihelion Software Ltd.	*/

/**
*** This header file is used to control the debugging options available in
*** the Server. There is a set of flags, debugflags, and a debugging option
*** is on when the corresponding bit is set in debugflags. The various flags
*** are declared. To keep track of all the debugging options I use a structure
*** debug_options, each option consisting of a character and the corresponding
*** flag, and I have an array of these options.
***
*** For example, when ctrl-shift M is pressed to toggle the message debugging
*** flag the system looks down options_list to find an entry with the letter
*** 'm'. It finds this, and xors debugflags with the corresponding flag
*** Message_Flag.
***
*** To add further debugging options, define a suitable flag, add an entry to
*** options_list, and in the code use lines like
***    if (debugflags & Silly_Flag)
*** to check whether or not the debugging facility is currently active.
**/

typedef struct debug_options {
  	 int 	flagchar;
				word	flag;
} debug_options;

#define	Message_Flag            0x1L
#define Search_Flag             0x2L
#define Open_Flag               0x4L
#define Name_Flag               0x8L
#define Read_Flag               0x10L
#define Boot_Flag               0x20L
#define Memory_Flag             0x40L
#define Keyboard_Flag           0x80L
#define Init_Flag               0x100L
#define Com_Flag                0x200L
#define Write_Flag              0x400L
#define Quit_Flag               0x800L
#define Close_Flag              0x1000L
#define HardDisk_Flag           0x2000L
#define Log_Flag                0x4000L
#define Graphics_Flag           0x8000L

#define Log_to_screen           1
#define Log_to_file             2
#define Log_to_both             3

#ifdef Server_Module
WORD debugflags;
int  log_dest = Log_to_screen;
debug_options options_list[] = 
                               { { 'm', Message_Flag    }, 
                                 { 's', Search_Flag     },
                                 { 'o', Open_Flag       },
                                 { 'n', Name_Flag       },
                                 { 'r', Read_Flag       },
                                 { 'b', Boot_Flag       },
                                 { 'x', Memory_Flag     },
                                 { 'k', Keyboard_Flag   },
                                 { 'i', Init_Flag       },
                                 { 'c', Com_Flag        },
                                 { 'w', Write_Flag      },
                                 { 'q', Quit_Flag       },
                                 { 'p', Close_Flag      },
                                 { 'h', HardDisk_Flag   },
                                 { 'l', Log_Flag        },
                                 { 'g', Graphics_Flag   },
                                 { '\0', 0L             } };
#endif
#ifndef Server_Module
extern WORD debugflags;
extern int  log_dest;
extern debug_options options_list[];
#endif

/**
*** Where the bootstrap program is loaded, assumes T800 but not really
*** important
**/
#ifndef MemStart
#define MemStart  MinInt+0x70
#endif

/**
*** arguments to boot_transputer() in module tload.c
**/
#define debugboot	1
#define serverboot	2

/**
*** Various internal debugging options, similar to defines.h. They are enabled
*** only if DEBUG is defined on the command line.
**/
#ifdef DEBUG
#define memwatch      1
#else
#define memwatch      0
#endif

