/*------------------------------------------------------------------------
--                                                                      --
--          H E L I O S   I N P U T / O U T P U T   S E R V E R         --
--          ---------------------------------------------------         --
--                                                                      --
--             Copyright (C) 1987, Perihelion Software Ltd.             --
--                        All Rights Reserved.                          --
--                                                                      --
--      fundefs.h                                                       --
--                                                                      --
--         Declarations of the functions shared between modules         --
--         If your compiler supports function ANSI function prototypes, --
--         please disable this option to save yourself from a couple of --
--         hundred warning messages.                                    --
--                                                                      --
--      Author:  BLV 8/10/87                                            --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89   Copyright (C) 1987, Perihelion Software Ltd.       */

/**
*** Linked list library, the same linked lists as under AmigaDos and Helios
**/
PUBLIC List *fn( MakeHeader,  (void));
PUBLIC void fn(  InitList,    (List *));
PUBLIC void fn(  WalkList,    (List *, VoidFnPtr, ...));
PUBLIC void fn(  FreeList,    (List *));
PUBLIC Node *fn( AddHead,     (Node *, List *));
PUBLIC Node *fn( AddTail,     (Node *, List *));
PUBLIC Node *fn( listRemove,  (Node *));
PUBLIC Node *fn( NextNode,    (Node *));
PUBLIC WORD fn(  Wander,      (List *, WordFnPtr, ...));
PUBLIC WORD fn(  TstList,     (List *));
PUBLIC WORD fn(  ListLength,  (List *));
PUBLIC void fn(  PreInsert,   (Node *, Node *));
PUBLIC void fn(  PostInsert,  (Node *, Node *));


/**
*** There are problems with Remove because all ANSI and pseudo-ANSI libraries
*** including the Microsoft one for the PC have a routine remove(), which
*** clashes with the list Remove with some linkers.
**/
#define Remove listRemove

/**
*** A higher level coroutine library using the above linked lists.
**/
PUBLIC Conode *fn( NewCo,     (VoidFnPtr));
PUBLIC void   fn(  TidyColib, (void));
PUBLIC void   fn(  StartCo,   (Conode *));
PUBLIC void   fn(  Suspend,   (void));
PUBLIC void   fn(  Seppuku,   (void));
PUBLIC WORD   fn(  InitColib, (void));

/**
*** Some general purpose request handlers. For example, doing things like
*** Rename on the Clock is not sensible, so Clock_Rename is #define'd to be
*** Invalidfn_handler
**/
PUBLIC  void fn( Invalidfn_handler,         (Conode *)); 
PUBLIC  void fn( Create_handler,            (Conode *));
PUBLIC  void fn( Dummy_handler,             (Conode *));
PUBLIC  void fn( IgnoreVoid,                (void));
PUBLIC  void fn( Device_ObjectInfo_handler, (Conode *));
PUBLIC  void fn( GetDefaultAttr,            (Conode *));
PUBLIC  void fn( Device_GetSize,            (Conode *));
PUBLIC  void fn( NullFn,                    (Conode *));
PUBLIC  WORD fn( Ignore,                    (void));
PUBLIC  void fn( Forward_handler,           (Conode *));
#define Nullfn ((VoidFnPtr) NULL)

/**
*** Some utilities
**/
PUBLIC WORD fn(  convert_name,    (void));
PUBLIC WORD fn(  flatten,         (char *));
PUBLIC WORD fn(  mystrcmp,        (char *, char *));
PUBLIC WORD fn(  get_unix_time,   (void));
PUBLIC void fn(  NewStream,       (WORD, WORD, WORD, VoidFnPtr *));
PUBLIC void fn(  clear_bytes,     (char *, int));
PUBLIC void fn(  AddAttribute,    (Attributes *, Attribute));
PUBLIC void fn(  RemoveAttribute, (Attributes *, Attribute));
PUBLIC void fn(  SetInputSpeed,   (Attributes *, WORD));
PUBLIC void fn(  SetOutputSpeed,  (Attributes *, WORD));
PUBLIC void fn(  InitAttributes,  (Attributes *));
PUBLIC void fn(  CopyAttributes,  (Attributes *, Attributes *));
PUBLIC WORD fn(  IsAnAttribute,   (Attributes *, Attribute));
PUBLIC WORD fn(  GetInputSpeed,   (Attributes *));
PUBLIC WORD fn(  GetOutputSpeed,  (Attributes *));
PUBLIC char *fn( get_config,      (char *));
PUBLIC WORD fn(  get_int_config,  (char *));
PUBLIC WORD fn(  Request_Stat,    (void));
PUBLIC void fn(  Request_Return,  (WORD, WORD, WORD));
PUBLIC void fn(  goto_sleep,      (WORD));
PUBLIC WORD fn(  FormOpenReply,   (WORD, WORD, WORD, WORD));
/**
*** an 8086 has the same byte ordering as the transputer, a 68000 does not
**/
#if swapping_needed
PUBLIC WORD fn( swap, (WORD));
#else
#define swap(a) (a)
#endif

/**
*** On the ST, or any other machine where interrupt vectors are used which
*** need to access the global Now e.g. for the time field in a keyboard event,
*** the following macro is a clock update. This clock update is performed
*** whenever I need to do a system call that is likely to take a long time.
**/
#if interrupts_use_clock
#define UpdateClock()  Now = clock()
#else
#define UpdateClock()
#endif

/**
*** The various options for console/window devices.
**/
#if !(multiple_windows)

#if use_ANSI_emulator

#define output(a)     ANSI_out(a, &Server_window)
#define window_output ANSI_out 
PUBLIC  void fn( ANSI_out, (char *, Window *));

#if PC
PUBLIC  void fn( send_ch, (int));
#endif
#if ST
#define send_ch(x) Bconout(2, x)
#endif

#else /* use_ANSI_emulator */

#define window_output(a, b)  output(a)
PUBLIC void fn( output, (char *));

#endif /* use_ANSI_emulator */

#else /* multiple_windows */

PUBLIC WORD fn( create_a_window, (char *));
PUBLIC void fn( window_size,     (WORD, WORD *, WORD *));
PUBLIC void fn( close_window,    (WORD));
PUBLIC void fn( send_to_window,  (char *, WORD));
#if use_ANSI_emulator

PUBLIC  void fn( ANSI_out, (char *, Window *));
#define window_output ANSI_out

#else /* use_ANSI_emulator */

#if AMIGA
#define window_output(a, b) send_to_window(a, (b)->handle)
#else
#define window_output send_to_window
#endif

#endif /* use_ANSI_emulator	*/

#define output(a) window_output(a, &Server_window)

#endif /* multiple_windows */

PUBLIC void fn( ServerDebug, (char *, ...));
PUBLIC void fn( outputch,    (int, Window *));
#if use_ANSI_emulator
PUBLIC WORD fn( Init_Ansi,   (Window *, int, int));
PUBLIC void fn( Tidy_Ansi,   (Screen *));
#endif

/**
*** The following declarations are used in header file server.h and module
*** server.c to indicate the functions available on each device.
*** Many functions are no-ops or cannot be implemented satisfactorily, so
*** they are hash-defined to suitable defaults. Also, some devices can
*** share code to handle particular requests.
**/

PUBLIC  void fn( Helios_InitServer, (Conode *));
#define Drive_InitServer            IgnoreVoid
#define Drive_TidyServer            IgnoreVoid
#define Drive_Private               Invalidfn_handler
#define Drive_Testfun               Nullfn
PUBLIC  void fn( Drive_Locate,      (Conode *));
PUBLIC  void fn( Drive_Open,        (Conode *));
PUBLIC  void fn( Drive_Create,      (Conode *));
PUBLIC  void fn( Drive_Delete,      (Conode *));
PUBLIC  void fn( Drive_ObjectInfo,  (Conode *));
PUBLIC  void fn( Drive_ServerInfo,  (Conode *));
PUBLIC  void fn( Drive_Rename,      (Conode *));
PUBLIC  void fn( Drive_Link,        (Conode *));
#define Drive_Protect               Invalidfn_handler
PUBLIC  void fn( Drive_SetDate,     (Conode *));
#define Drive_Refine                Invalidfn_handler
#define Drive_CloseObj              Invalidfn_handler

PUBLIC  WORD fn( File_InitStream, (Conode *));
PUBLIC  WORD fn( File_TidyStream, (Conode *));
#define File_PrivateStream        Invalidfn_handler
PUBLIC  void fn( File_Read,       (Conode *));
PUBLIC  void fn( File_Write,      (Conode *));
PUBLIC  void fn( File_Close,      (Conode *));
PUBLIC  void fn( File_GetSize,    (Conode *));
#define File_SetSize              Invalidfn_handler
PUBLIC  void fn( File_Seek,       (Conode *));
#define File_GetAttr              GetDefaultAttr
#define File_SetAttr              NullFn
#define File_EnableEvents         Invalidfn_handler
#define File_Acknowledge          Invalidfn_handler
#define File_NegAcknowledge       Invalidfn_handler

PUBLIC  WORD fn( Dir_InitStream,  (Conode *));
PUBLIC  WORD fn( Dir_TidyStream,  (Conode *));
#define Dir_PrivateStream         Invalidfn_handler
PUBLIC  void fn( Dir_Read,        (Conode *));
#define Dir_Write                 Invalidfn_handler
PUBLIC  void fn( Dir_Close,       (Conode *));
PUBLIC  void fn( Dir_GetSize,     (Conode *));
#define Dir_SetSize               Invalidfn_handler
#define Dir_Seek                  Invalidfn_handler
#define Dir_GetAttr               GetDefaultAttr
#define Dir_SetAttr               NullFn
#define Dir_EnableEvents          Invalidfn_handler
#define Dir_Acknowledge           Invalidfn_handler
#define Dir_NegAcknowledge        Invalidfn_handler

PUBLIC  void fn( IOPROC_InitServer, (Conode *));
PUBLIC  void fn( IOPROC_TidyServer, (Conode *));
PUBLIC  void fn( IOPROC_Private,    (Conode *));
#define IOPROC_Testfun              Nullfn
PUBLIC  void fn( IOPROC_Open,       (Conode *));
PUBLIC  void fn( IOPROC_Locate,     (Conode *));
#define IOPROC_Create               IOPROC_Locate
#define IOPROC_Delete               Forward_handler
PUBLIC  void fn( IOPROC_ObjectInfo, (Conode *));
#define IOPROC_ServerInfo           Forward_handler
PUBLIC  void fn( IOPROC_Rename,     (Conode *));
#define IOPROC_Link                 Forward_handler
#define IOPROC_Protect              Forward_handler
#define IOPROC_SetDate              Forward_handler
#define IOPROC_Refine               Forward_handler
#define IOPROC_CloseObj             Forward_handler
#define IOPROC_InitStream           Ignore
#define IOPROC_TidyStream           Ignore
#define IOPROC_PrivateStream        Invalidfn_handler
#define IOPROC_Read                 Dir_Read
#define IOPROC_Write                Invalidfn_handler
PUBLIC  void fn( IOPROC_Close,      (Conode *));
#define IOPROC_GetSize              Dir_GetSize
#define IOPROC_SetSize              Invalidfn_handler
#define IOPROC_Seek                 Invalidfn_handler
#define IOPROC_GetAttr              GetDefaultAttr
#define IOPROC_SetAttr              NullFn
#define IOPROC_EnableEvents         Invalidfn_handler
#define IOPROC_Acknowledge          IgnoreVoid
#define IOPROC_NegAcknowledge       IgnoreVoid

#if multiple_windows
PUBLIC  void fn( Window_InitServer, (Conode *));
PUBLIC  void fn( Window_TidyServer, (Conode *));
#define Window_Private              Invalidfn_handler
PUBLIC  void fn( Window_Testfun,    (bool *));
PUBLIC  void fn( Window_Open,       (Conode *));
PUBLIC  void fn( Window_Locate,     (Conode *));
PUBLIC  void fn( Window_Create,     (Conode *));
PUBLIC  void fn( Window_Delete,     (Conode *));
PUBLIC  void fn( Window_ObjectInfo, (Conode *));
#define Window_ServerInfo           Invalidfn_handler
#define Window_Rename               Invalidfn_handler
#define Window_Link                 Invalidfn_handler
#define Window_Protect              Invalidfn_handler
#define Window_SetDate              Invalidfn_handler
#define Window_Refine               Invalidfn_handler
#define Window_CloseObj             Invalidfn_handler

#define WindowDir_InitStream        Ignore
#define WindowDir_TidyStream        Ignore
#define WindowDir_PrivateStream     Invalidfn_handler
#define WindowDir_Read              Dir_Read
#define WindowDir_Write             Invalidfn_handler
#define WindowDir_Close             IOPROC_Close
#define WindowDir_GetSize           Dir_GetSize
#define WindowDir_SetSize           Invalidfn_handler
#define WindowDir_Seek              Invalidfn_handler
#define WindowDir_GetAttr           GetDefaultAttr
#define WindowDir_SetAttr           NullFn
#define WindowDir_EnableEvents      Invalidfn_handler
#define WindowDir_Acknowledge       IgnoreVoid
#define WindowDir_NegAcknowledge    IgnoreVoid

#endif

/**
*** The console device is always optionally available.
**/

PUBLIC  void fn( Console_InitServer, (Conode *));
#define Console_TidyServer           IgnoreVoid
#define Console_Private              Invalidfn_handler
#if multiple_windows
PUBLIC  void fn( Console_Testfun,    (bool *));
#else
#define Console_Testfun              Nullfn
#endif
PUBLIC  void fn( Console_Open,       (Conode *));
#define Console_Locate               Create_handler
#define Console_Create               Create_handler
#define Console_Delete               Invalidfn_handler
#define Console_ObjectInfo           Device_ObjectInfo_handler
#define Console_ServerInfo           Invalidfn_handler
#define Console_Rename               Invalidfn_handler
#define Console_Link                 Invalidfn_handler
#define Console_Protect              Invalidfn_handler
#define Console_SetDate              Invalidfn_handler
#define Console_Refine               Invalidfn_handler
#define Console_CloseObj             Invalidfn_handler

PUBLIC  WORD fn( Console_InitStream,   (Conode *));
PUBLIC  WORD fn( Console_TidyStream,   (Conode *));
#define Console_PrivateStream          Invalidfn_handler
PUBLIC  void fn( Console_Read,         (Conode *));
PUBLIC  void fn( Console_Write,        (Conode *));
PUBLIC  void fn( Console_Close,        (Conode *));
#define Console_GetSize                Device_GetSize
#define Console_SetSize                Invalidfn_handler
#define Console_Seek                   Invalidfn_handler
PUBLIC  void fn( Console_GetAttr,      (Conode *));
PUBLIC  void fn( Console_SetAttr,      (Conode *));
PUBLIC  void fn( Console_EnableEvents, (Conode *));
#define Console_Acknowledge            IgnoreVoid
#define Console_NegAcknowledge         IgnoreVoid

PUBLIC  void fn( write_to_log,      (char *));
PUBLIC  void fn( init_logger,       (void));
PUBLIC  void fn( tidy_logger,       (void));

#define Logger_InitServer           IgnoreVoid
#define Logger_TidyServer           IgnoreVoid
#define Logger_Private              Invalidfn_handler
#define Logger_Testfun              Nullfn
PUBLIC  void fn( Logger_Open,       (Conode *));
#define Logger_Locate               Create_handler
#define Logger_Create               Create_handler
PUBLIC  void fn( Logger_Delete,     (Conode *));
PUBLIC  void fn( Logger_ObjectInfo, (Conode *));
#define Logger_ServerInfo           Invalidfn_handler
#define Logger_Rename               Invalidfn_handler
#define Logger_Link                 Invalidfn_handler
#define Logger_Protect              Invalidfn_handler
#define Logger_SetDate              Invalidfn_handler
#define Logger_Refine               Invalidfn_handler
#define Logger_CloseObj             Invalidfn_handler

#define Logger_InitStream             Ignore
#define Logger_TidyStream             Ignore
#define Logger_PrivateStream          Invalidfn_handler
PUBLIC  void fn( Logger_Read,         (Conode *));
PUBLIC  void fn( Logger_Write,        (Conode *));
PUBLIC  void fn( Logger_Close,        (Conode *));
PUBLIC  void fn( Logger_GetSize,      (Conode *));
#define Logger_SetSize                Invalidfn_handler
PUBLIC  void fn( Logger_Seek,         (Conode *));
#define Logger_GetAttr                Invalidfn_handler
#define Logger_SetAttr                Invalidfn_handler
#define Logger_EnableEvents           Invalidfn_handler
#define Logger_Acknowledge            IgnoreVoid
#define Logger_NegAcknowledge         IgnoreVoid

#define Clock_InitServer           IgnoreVoid
#define Clock_TidyServer           IgnoreVoid
#define Clock_Private              Invalidfn_handler
#define Clock_Testfun              Nullfn
#define Clock_Locate               Create_handler
#define Clock_Open                 Invalidfn_handler
#define Clock_Create               Create_handler
#define Clock_Delete               Invalidfn_handler
PUBLIC  void fn( Clock_ObjectInfo, (Conode *));
#define Clock_ServerInfo           Invalidfn_handler
#define Clock_Rename               Invalidfn_handler
#define Clock_Link                 Invalidfn_handler
#define Clock_Protect              Invalidfn_handler
#if SUN
#define Clock_SetDate              Invalidfn_handler
#else
PUBLIC  void fn( Clock_SetDate,    (Conode *));
#endif
#define Clock_Refine               Invalidfn_handler
#define Clock_CloseObj             Invalidfn_handler

#if interaction_supported
#define Host_InitServer              IgnoreVoid
#define Host_TidyServer              IgnoreVoid
#define Host_Private                 Invalidfn_handler
#define Host_Testfun                 Nullfn
#define Host_Locate                  Create_handler
PUBLIC  void fn( Host_Open,          (Conode *));
#define Host_Create                  Create_handler
#define Host_Delete                  Invalidfn_handler
#define Host_ObjectInfo              Device_ObjectInfo_handler
#define Host_ServerInfo              Invalidfn_handler
#define Host_Rename                  Invalidfn_handler
#define Host_Link                    Invalidfn_handler
#define Host_Protect                 Invalidfn_handler
#define Host_SetDate                 Invalidfn_handler
#define Host_Refine                  Invalidfn_handler
#define Host_CloseObj                Invalidfn_handler

#define Host_InitStream              Ignore
#define Host_TidyStream              Ignore
PUBLIC  void fn( Host_PrivateStream, (Conode *));
#define Host_Read		                  Invalidfn_handler 
#define Host_Write			                Invalidfn_handler
PUBLIC  void fn( Host_Close,         (Conode *));
#define Host_GetSize                 Invalidfn_handler
#define Host_SetSize                 Invalidfn_handler
#define Host_Seek                    Invalidfn_handler
#define Host_GetAttr                 Invalidfn_handler
#define Host_SetAttr                 Invalidfn_handler
#define Host_EnableEvents            Invalidfn_handler
#define Host_Acknowledge             IgnoreVoid
#define Host_NegAcknowledge          IgnoreVoid

#endif /* interaction_supported */

#if gem_supported
PUBLIC  void fn( Gem_InitServer,     (Conode *));
PUBLIC  void fn( Gem_TidyServer,     (Conode *));
#define Gem_Private                  Invalidfn_handler
#if Gem_Always_Available
#define Gem_Testfun                  Nullfn
#else
PUBLIC  void fn( Gem_Testfun,        (bool *));
#endif
#define Gem_Locate                   Create_handler
PUBLIC  void fn( Gem_Open,           (Conode *));
#define Gem_Create                   Create_handler
#define Gem_Delete                   Invalidfn_handler
#define Gem_ObjectInfo               Device_ObjectInfo_handler
#define Gem_ServerInfo               Invalidfn_handler
#define Gem_Rename                   Invalidfn_handler
#define Gem_Link                     Invalidfn_handler
#define Gem_Protect                  Invalidfn_handler
#define Gem_SetDate                  Invalidfn_handler
#define Gem_Refine                   Invalidfn_handler
#define Gem_CloseObj                 Invalidfn_handler
PUBLIC  WORD fn( Gem_InitStream,     (Conode *));
#define Gem_TidyStream               Ignore
PUBLIC  void fn( Gem_PrivateStream,  (Conode *));
#define Gem_Read                     Invalidfn_handler
#define Gem_Write                    Invalidfn_handler
PUBLIC  void fn( Gem_Close,          (Conode *));
#define Gem_GetSize                  Invalidfn_handler
#define Gem_SetSize                  Invalidfn_handler
#define Gem_Seek                     Invalidfn_handler
#define Gem_GetAttr                  Invalidfn_handler
#define Gem_SetAttr                  Invalidfn_handler
PUBLIC  void fn( Gem_EnableEvents,   (Conode *));
#define Gem_Acknowledge              IgnoreVoid
#define Gem_NegAcknowledge           IgnoreVoid

PUBLIC void fn( poll_gem,    (void));
PUBLIC void fn( restart_gem, (void));
PUBLIC void fn( vdi,         (int **));
#endif /* Gem_supported */

/**
*** The RS232, Centronics Midi and Printer devices share most of their code
**/
#if Ports_used
PUBLIC void fn( Port_Open,       (Conode *));
PUBLIC void fn( Port_Locate,     (Conode *));
PUBLIC void fn( Port_TidyServer, (Conode *));
PUBLIC void fn( Port_ObjInfo,    (Conode *));
PUBLIC void fn( Port_TidyStream, (Conode *));
PUBLIC void fn( Port_Close,      (Conode *));
PUBLIC void fn( Port_Rename,     (Conode *));
PUBLIC void fn( Port_Read,       (Conode *));
PUBLIC void fn( Port_Write,      (Conode *));
PUBLIC void fn( Port_GetAttr,    (Conode *));
PUBLIC void fn( Port_SetAttr,    (Conode *));

#define PortDir_InitStream       Ignore
#define PortDir_TidyStream       Ignore
#define PortDir_PrivateStream    Invalidfn_handler
#define PortDir_Read             Dir_Read
#define PortDir_Write            Invalidfn_handler
#define PortDir_Close            IOPROC_Close
#define PortDir_GetSize          Dir_GetSize
#define PortDir_SetSize          Invalidfn_handler
#define PortDir_Seek             Invalidfn_handler
#define PortDir_GetAttr          GetDefaultAttr
#define PortDir_SetAttr          NullFn
#define PortDir_EnableEvents     Invalidfn_handler
#define PortDir_Acknowledge      IgnoreVoid
#define PortDir_NegAcknowledge   IgnoreVoid

#endif

#if RS232_supported
PUBLIC  void fn( RS232_InitServer,   (Conode *));
PUBLIC  void fn( RS232_TidyServer,   (Conode *));
#define RS232_Private                Invalidfn_handler
#if RS232_Always_Available
#define RS232_Testfun                Nullfn
#else																									   
PUBLIC  void fn( RS232_Testfun,      (bool *));
#endif
#define RS232_Locate                 Port_Locate
#define RS232_Open                   Port_Open
#define RS232_Create                 Port_Locate
#define RS232_Delete                 Invalidfn_handler
#define RS232_ObjectInfo             Port_ObjInfo
#define RS232_ServerInfo             Invalidfn_handler
#define RS232_Rename                 Port_Rename
#define RS232_Link                   Invalidfn_handler
#define RS232_Protect                Invalidfn_handler
#define RS232_SetDate                Invalidfn_handler
#define RS232_Refine                 Invalidfn_handler
#define RS232_CloseObj               Invalidfn_handler
#define RS232_InitStream             Ignore
#define RS232_TidyStream             Port_TidyStream
#define RS232_PrivateStream          Invalidfn_handler
#define RS232_Read                   Port_Read
#define RS232_Write                  Port_Write
#define RS232_Close                  Port_Close
#define RS232_GetSize                Device_GetSize
#define RS232_SetSize                Invalidfn_handler
#define RS232_Seek                   Invalidfn_handler
#define RS232_GetAttr                Port_GetAttr
#define RS232_SetAttr                Port_SetAttr
PUBLIC  void fn( RS232_EnableEvents, (Conode *));
#define RS232_Acknowledge            IgnoreVoid
#define RS232_NegAcknowledge         IgnoreVoid

PUBLIC  WORD fn( RS232_initlist,       (List *, Attributes **));
PUBLIC  WORD fn( RS232_send,           (WORD, WORD, BYTE *));
PUBLIC  WORD fn( RS232_pollwrite,      (WORD));
PUBLIC  WORD fn( RS232_abortwrite,     (WORD));
PUBLIC  WORD fn( RS232_receive,        (WORD, WORD, BYTE *));
PUBLIC  WORD fn( RS232_pollread,       (WORD));
PUBLIC  WORD fn( RS232_abortread,      (WORD));
PUBLIC  void fn( RS232_done,           (WORD));
PUBLIC  void fn( RS232_configure,      (WORD));
PUBLIC  void fn( RS232_check_events,   (void));
PUBLIC  void fn( RS232_disable_events, (WORD));
PUBLIC  void fn( RS232_enable_events,  (WORD, WORD, WORD));
#endif

#if Centronics_supported
PUBLIC  void fn( Centronics_InitServer, (Conode *));
#define Centronics_TidyServer           Port_TidyServer
#define Centronics_Private              Invalidfn_handler
#if Centronics_Always_Available
#define Centronics_Testfun              Nullfn
#else
PUBLIC  void fn( Centronics_Testfun,    (bool *));
#endif
#define Centronics_Locate               Port_Locate
#define Centronics_Open                 Port_Open
#define Centronics_Create               Port_Locate
#define Centronics_Delete               Invalidfn_handler
#define Centronics_ObjectInfo           Port_ObjInfo
#define Centronics_ServerInfo           Invalidfn_handler
#define Centronics_Rename               Port_Rename
#define Centronics_Link                 Invalidfn_handler
#define Centronics_Protect              Invalidfn_handler
#define Centronics_SetDate              Invalidfn_handler
#define Centronics_Refine               Invalidfn_handler
#define Centronics_CloseObj             Invalidfn_handler
#define Centronics_InitStream           Ignore
#define Centronics_TidyStream           Ignore
#define Centronics_PrivateStream        Invalidfn_handler
#if Centronics_readable
#define Centronics_Read                 Port_Read
#else
#define Centronics_Read                 Invalidfn_handler
#endif
#define Centronics_Write                Port_Write
#define Centronics_Close                Port_Close
#define Centronics_GetSize              Device_GetSize
#define Centronics_SetSize              Invalidfn_handler
#define Centronics_Seek                 Invalidfn_handler
#define Centronics_GetAttr              GetDefaultAttr
#define Centronics_SetAttr              NullFn
#define Centronics_EnableEvents         Invalidfn_handler
#define Centronics_Acknowledge          IgnoreVoid
#define Centronics_NegAcknowledge       IgnoreVoid

PUBLIC  WORD fn( Centronics_initlist,   (List *, Attributes **));
PUBLIC  WORD fn( Centronics_send,       (WORD, WORD, BYTE *));
PUBLIC  WORD fn( Centronics_pollwrite,  (WORD));
PUBLIC  WORD fn( Centronics_abortwrite, (WORD));
PUBLIC  void fn( Centronics_done,       (WORD));
PUBLIC  void fn( Centronics_configure,  (WORD));
#if Centronics_readable
PUBLIC  WORD fn( Centronics_receive,    (WORD, WORD, BYTE *));
PUBLIC  WORD fn( Centronics_pollread,   (WORD));
PUBLIC  WORD fn( Centronics_abortread,  (WORD));
#endif
#endif

#if Printer_supported
PUBLIC  void fn( Printer_InitServer,    (Conode *));
#define Printer_TidyServer              Port_TidyServer
#define Printer_Private                 Invalidfn_handler
#if Printer_Always_Available
#define Printer_Testfun                 Nullfn
#else
PUBLIC  void fn( Printer_Testfun,       (bool *));
#endif
#define Printer_Locate                  Port_Locate
#define Printer_Open                    Port_Open
#define Printer_Create                  Port_Locate
#define Printer_Delete                  Invalidfn_handler
#define Printer_ObjectInfo              Port_ObjInfo
#define Printer_ServerInfo              Invalidfn_handler
#define Printer_Rename                  Port_Rename
#define Printer_Link                    Invalidfn_handler
#define Printer_Protect                 Invalidfn_handler
#define Printer_SetDate                 Invalidfn_handler
#define Printer_Refine                  Invalidfn_handler
#define Printer_CloseObj                Invalidfn_handler
#define Printer_InitStream              Ignore
#define Printer_TidyStream              Port_TidyStream
#define Printer_PrivateStream           Invalidfn_handler
#define Printer_Read                    Invalidfn_handler
#define Printer_Write                   Port_Write
#define Printer_Close                   Port_Close
#define Printer_GetSize                 Device_GetSize
#define Printer_SetSize                 Invalidfn_handler
#define Printer_Seek                    Invalidfn_handler
#define Printer_GetAttr                 Port_GetAttr
#define Printer_SetAttr                 Port_SetAttr
#define Printer_EnableEvents            Invalidfn_handler
#define Printer_Acknowledge             IgnoreVoid
#define Printer_NegAcknowledge          IgnoreVoid

PUBLIC  WORD fn( Printer_initlist,      (List *, Attributes **));
PUBLIC  WORD fn( Printer_send,          (WORD, WORD, BYTE *));
PUBLIC  WORD fn( Printer_pollwrite,     (WORD));
PUBLIC  WORD fn( Printer_abortwrite,    (WORD));
PUBLIC  void fn( Printer_done,          (WORD));
PUBLIC  void fn( Printer_configure,     (WORD));
#endif

#if Midi_supported
PUBLIC  void fn( Midi_InitServer,       (Conode *));
#define Midi_TidyServer                 Port_TidyServer
#define Midi_Private                    Invalidfn_handler
#if Midi_Always_Available
#define Midi_Testfun                    Nullfn
#else
PUBLIC  void fn( Midi_Testfun,          (bool *));
#endif
#define Midi_Locate                     Port_Locate
#define Midi_Open                       Port_Open
#define Midi_Create                     Port_Locate
#define Midi_Delete                     Invalidfn_handler
#define Midi_ObjectInfo                 Port_ObjInfo
#define Midi_ServerInfo                 Invalidfn_handler
#define Midi_Rename                     Port_Rename
#define Midi_Link                       Invalidfn_handler
#define Midi_Protect                    Invalidfn_handler
#define Midi_SetDate                    Invalidfn_handler
#define Midi_Refine                     Invalidfn_handler
#define Midi_CloseObj                   Invalidfn_handler
#define Midi_InitStream                 Ignore
#define Midi_TidyStream                 Port_TidyStream
#define Midi_PrivateStream              Invalidfn_handler
#define Midi_Read                       Invalidfn_handler
#define Midi_Write                      Port_Write
#define Midi_Close                      Port_Close
#define Midi_GetSize                    Device_GetSize
#define Midi_SetSize                    Invalidfn_handler
#define Midi_Seek                       Invalidfn_handler
#define Midi_GetAttr                    Port_GetAttr
#define Midi_SetAttr                    Port_SetAttr
#define Midi_EnableEvents               Invalidfn_handler
#define Midi_Acknowledge                IgnoreVoid
#define Midi_NegAcknowledge             IgnoreVoid

PUBLIC  WORD fn( Midi_initlist,         (List *, Attributes **));
PUBLIC  WORD fn( Midi_send,             (WORD, WORD, BYTE *));
PUBLIC  WORD fn( Midi_pollwrite,        (WORD));
PUBLIC  WORD fn( Midi_abortwrite,       (WORD));
PUBLIC  WORD fn( Midi_receive,          (WORD, WORD, BYTE *));
PUBLIC  WORD fn( Midi_pollread,         (WORD));
PUBLIC  WORD fn( Midi_abortread,        (WORD));
PUBLIC  void fn( Midi_done,             (WORD));
PUBLIC  void fn( Midi_configure,        (WORD));
#endif

#if Rawdisk_supported
#define RawDisk_InitServer         IgnoreVoid
#define RawDisk_TidyServer         IgnoreVoid
#define RawDisk_Private            Invalidfn_handler
PUBLIC  void fn( RawDisk_Testfun,  (bool *));
PUBLIC  void fn( RawDisk_Open,     (Conode *));
#define RawDisk_Locate             Create_handler
#define RawDisk_Create             Create_handler
#define RawDisk_Delete             Invalidfn_handler
#define RawDisk_ObjectInfo         Device_ObjectInfo_handler
#define RawDisk_ServerInfo         Invalidfn_handler
#define RawDisk_Rename             Invalidfn_handler
#define RawDisk_Link               Invalidfn_handler
#define RawDisk_Protect            Invalidfn_handler
#define RawDisk_SetDate            Invalidfn_handler
#define RawDisk_Refine             Invalidfn_handler
#define RawDisk_CloseObj           Invalidfn_handler

#define RawDisk_InitStream         Ignore
#define RawDisk_TidyStream         Ignore
#define RawDisk_PrivateStream      Invalidfn_handler
PUBLIC  void fn( RawDisk_Read,     (Conode *));
PUBLIC  void fn( RawDisk_Write,    (Conode *));
PUBLIC  void fn( RawDisk_Close,    (Conode *));
PUBLIC  void fn( RawDisk_GetSize,  (Conode *));
#define RawDisk_SetSize            Invalidfn_handler
PUBLIC  void fn( RawDisk_Seek,     (Conode *));
#define RawDisk_GetAttr            Invalidfn_handler
#define RawDisk_SetAttr            Invalidfn_handler
#define RawDisk_EnableEvents       Invalidfn_handler
#define RawDisk_Acknowledge        IgnoreVoid
#define RawDisk_NegAcknowledge     IgnoreVoid
#endif

#if mouse_supported
PUBLIC  void fn( Mouse_InitServer,      (Conode *));
PUBLIC  void fn( Mouse_TidyServer,      (Conode *));
#define Mouse_Private                   Invalidfn_handler
#if Mouse_Always_Available
#define Mouse_Testfun                   Nullfn
#else																							 
PUBLIC  void fn( Mouse_Testfun,         (bool *));
#endif
PUBLIC  void fn( Mouse_Open,            (Conode *));
#define Mouse_Locate                    Create_handler
#define Mouse_Create                    Create_handler
#define Mouse_Delete                    Invalidfn_handler
#define Mouse_ObjectInfo                Device_ObjectInfo_handler
#define Mouse_ServerInfo                Invalidfn_handler
#define Mouse_Rename                    Invalidfn_handler
#define Mouse_Link                      Invalidfn_handler
#define Mouse_Protect                   Invalidfn_handler
#define Mouse_SetDate                   Invalidfn_handler
#define Mouse_Refine                    Invalidfn_handler
#define Mouse_CloseObj                  Invalidfn_handler
PUBLIC  WORD fn( Mouse_InitStream,      (Conode *));
PUBLIC  WORD fn( Mouse_TidyStream,      (Conode *));
#define Mouse_PrivateStream             Invalidfn_handler
#define Mouse_Read                      Invalidfn_handler
#define Mouse_Write                     Invalidfn_handler
PUBLIC  void fn( Mouse_Close,           (Conode *));
#define Mouse_GetSize                   Invalidfn_handler
#define Mouse_SetSize                   Invalidfn_handler
#define Mouse_Seek                      Invalidfn_handler
#define Mouse_GetAttr                   Invalidfn_handler
#define Mouse_SetAttr                   Invalidfn_handler
PUBLIC  void fn( Mouse_EnableEvents,    (Conode *));
PUBLIC  void fn( Mouse_Acknowledge,     (Conode *));
PUBLIC  void fn( Mouse_NegAcknowledge,  (Conode *));

PUBLIC  void fn( initialise_mouse,      (void));
PUBLIC  void fn( tidy_mouse,            (void));
PUBLIC  void fn( start_mouse,           (void));
PUBLIC  void fn( stop_mouse,            (void));
#endif

#if keyboard_supported
PUBLIC  void fn( Keyboard_InitServer,     (Conode *));
PUBLIC  void fn( Keyboard_TidyServer,     (Conode *));
#define Keyboard_Private                  Invalidfn_handler
#if Keyboard_Always_Available
#define Keyboard_Testfun                  Nullfn
#else
PUBLIC  void fn( Keyboard_Testfun,        (bool *));
#endif																												 
PUBLIC  void fn( Keyboard_Open,           (Conode *));
#define Keyboard_Locate                   Create_handler
#define Keyboard_Create                   Create_handler
#define Keyboard_Delete                   Invalidfn_handler
#define Keyboard_ObjectInfo               Device_ObjectInfo_handler
#define Keyboard_ServerInfo               Invalidfn_handler
#define Keyboard_Rename                   Invalidfn_handler
#define Keyboard_Link                     Invalidfn_handler
#define Keyboard_Protect                  Invalidfn_handler
#define Keyboard_SetDate                  Invalidfn_handler
#define Keyboard_Refine                   Invalidfn_handler
#define Keyboard_CloseObj                 Invalidfn_handler
PUBLIC  WORD fn( Keyboard_InitStream,     (Conode *));
PUBLIC  WORD fn( Keyboard_TidyStream,     (Conode *));
#define Keyboard_PrivateStream            Invalidfn_handler
#define Keyboard_Read                     Invalidfn_handler
#define Keyboard_Write                    Invalidfn_handler
PUBLIC  void fn( Keyboard_Close,          (Conode *));
#define Keyboard_GetSize                  Invalidfn_handler
#define Keyboard_SetSize                  Invalidfn_handler
#define Keyboard_Seek                     Invalidfn_handler
#define Keyboard_GetAttr                  Invalidfn_handler
#define Keyboard_SetAttr                  Invalidfn_handler
PUBLIC  void fn( Keyboard_EnableEvents,   (Conode *));
PUBLIC  void fn( Keyboard_Acknowledge,    (Conode *));
PUBLIC  void fn( Keyboard_NegAcknowledge, (Conode *));
																																						 
PUBLIC  void fn( initialise_keyboard,     (void));
PUBLIC  void fn( tidy_keyboard,           (void));
PUBLIC  void fn( start_keyboard,          (void));
PUBLIC  void fn( stop_keyboard,           (void));
#endif

/**
*** Explicit declarations of the memory allocation functions
***
*** For most of the hardware on which the server is expected to run malloc()
*** takes an unsigned integer as argument, which may limit it to 64K. The main
*** server sources never need buffers more than 64K, so that is fine.
***
*** If the system's memory allocation is unsatisfactory for one reason or
*** another, I have written my own routines for the ST - see ST/STlocal.c for
*** details.
**/

#if use_own_memory_management
PUBLIC void fn(  initialise_memory, (void));
PUBLIC char *fn( get_mem,           (uint));
PUBLIC void fn(  free_mem,          (char *));
PUBLIC void fn(  memory_map,        (void));       /* a debugging facility */
#define malloc  get_mem
#define free    free_mem

#else

#if !(PC)                        /* The PC's header files declare malloc */
  #if !(AMIGA)
    PUBLIC char *fn( malloc,      (uint));
    PUBLIC void fn( free,         (char *));
  #else
    PUBLIC char *fn( malloc,      (uint));
    PUBLIC void fn( free,         (void *));
  #endif
#endif

#endif /* use_own_memory_management */

/**
*** Here are some functions to deal with the multi-tasking support.
*** The only implementation I have so far is for the Sun, and that is
*** a bit dodgy.
**/
#if multi_tasking
/**
*** These two routines are called when the Server starts-up and leaves
*** server mode. 
**/
PUBLIC void fn( InitMultiwait, (void));
PUBLIC void fn( TidyMultiwait, (void));
/**
*** This routine is called from inside the Server's main loop. It should
*** return when one of the I/O's the Server is currently waiting for is
*** possible, or after 1/2 a second to allow the Server to deal correctly
*** with timeouts.
**/
PUBLIC WORD fn( Multiwait, (WORD));
/**
*** This routine is called by the Server whenever it is waiting on some
*** form of I/O. The first argument is one of the constants defined in
*** structs.h, specifying the particular form of I/O. The second argument
*** (optional) gives more detail, e.g. which particular port the Server is
*** waiting on.
**/
PUBLIC void fn( AddMultiwait, (Conode *, WORD, ...));
/**
*** ClearMultiwait() is just the inverse of AddMultiwait()
**/
PUBLIC void fn( ClearMultiwait, (Conode *, WORD, ...));
#endif

/**
*** Bits that had been left out before.
**/
PUBLIC  void fn( initialise_devices,      (void));
PUBLIC  void fn( restart_devices,         (void));
PUBLIC  void fn( restore_devices,         (void));
PUBLIC  void fn( poll_the_devices,        (void));
PUBLIC  int  fn( read_char_from_keyboard, (WORD));
PUBLIC	 void fn( init_boot,               (void));
PUBLIC  void fn( boot_processor,          (int));
PUBLIC  void fn( tidy_boot,               (void));
PUBLIC  void fn( init_main_message,       (void));
PUBLIC  void fn( free_main_message,       (void));
PUBLIC  WORD fn( GetMsg,                  (MCB *));
PUBLIC  WORD fn( PutMsg,                  (MCB *));
#if debugger_incorporated
PUBLIC  void fn( init_debug,              (void));
PUBLIC  int  fn( debug,                   (void));
PUBLIC  void fn( tidy_debug,              (void));
#endif

#if floppies_available
PUBLIC  WORD fn( format_floppy, (char *, WORD, WORD, WORD, char *));
#endif

/**
*** The following are transputer-specific
**/
PUBLIC  int  fn( loadimage,               (void));
PUBLIC  void fn( resetlnk,                (void));
PUBLIC  void fn( xpreset,                 (void));
PUBLIC  void fn( xpanalyse,               (void));
PUBLIC  WORD fn( xpwrbyte,                (WORD));
PUBLIC  WORD fn( xpwrrdy,                 (void));
PUBLIC  WORD fn( xprdrdy,                 (void));
PUBLIC  WORD fn( xpwrword,                (WORD));
PUBLIC  WORD fn( xpwrint,                 (WORD));
PUBLIC  WORD fn( xpwrdata,                (BYTE *, WORD));
PUBLIC  WORD fn( xprddata,                (BYTE *, WORD));
PUBLIC  WORD fn( dbwrword,                (WORD, WORD));
PUBLIC  WORD fn( dbwrint,                 (WORD, WORD));
PUBLIC  WORD fn( xprdint,                 (void));
PUBLIC  WORD fn( xprdword,                (void));
PUBLIC  WORD fn( xprdbyte,                (void));
PUBLIC  WORD fn( dbrdword,                (WORD));
PUBLIC  WORD fn( dbrdint,                 (WORD));

