/*------------------------------------------------------------------------
--                                                                      --
--         H E L I O S   I N P U T / O U T P U T   S E R V E R          --
--         ---------------------------------------------------          --
--                                                                      --
--         Copyright (C) 1987, Perihelion Software Ltd.                 --
--                       All Rights Reserved.                           --
--                                                                      --
--  server.c                                                            --
--                                                                      --
--          The main module of the I/O server.                          --
--                                                                      --
--  Author:  BLV 8/10/87                                                --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89   Copyright (C) 1987, Perihelion Software Ltd.       */

#define Server_Module

/**
***    This is the main module of the Helios input/output Server. It contains
*** the start-up code, the main loops, and many utilities.
***
***    Please note the following convention. Server with a capital S refers
*** to the entire system, i.e. to all the devices together, whereas server
*** with a little s refers to the particular server for a device, e.g.
*** the console server or a server for one of the drives.
**/

/**
*** Header files :
***               sccs.h declares a static string containing the current
***               version number, its date, and a copyright message. This
***               is displayed when the Server starts up. There is a special
***               version of the header for the Synergy version of the Server.
***
***               helios.h is the main header file. It is a catenation of
***               all the header files, and should be kept in Ram disk if
***               possible. Using a single header file greatly reduces the
***               time needed for compilation. For details of which header
***               files form part of helios.h, see the makefile.
**/

#ifdef SYNERGY
#include "synergy\sccs.h"
#else
#include "sccs.h"
#endif

#include "helios.h"

/**
*** Function declarations
***                      The following lines declare the types of some of the
*** functions used in this module. PRIVATE is an alias for static, i.e. not
*** accessible from outside this module. Functions which may be accessed from
*** other modules are declared in the header file fundefs.h . Also, a large
*** number of public variables are declared in the header file server.h
***/

PRIVATE WORD fn( Init,               (void));
PRIVATE int  fn( Server,             (void));
PRIVATE int  fn( MainLoop,           (void));
PRIVATE void fn( TidyUp,             (void));
PRIVATE void fn( General_Server,     (Conode *));
PRIVATE void fn( read_config,        (char *));
PRIVATE void fn( tidy_config,        (void));

/*------------------------------------------------------------------------
--
--  void main(argc, argv)
--
------------------------------------------------------------------------*/

/**
***    This is the program start-up. It does lots of initialisation, and
***  starts up either the Server or the debugger.
***
***    On the ST I need to worry about the stack size for the main coroutine,
***  which has to be set in a very strange way under Mark Williams C.
***
***    If the server wants to do its own memory management then this must be
***  initialised as early as possible. The header file defines.h determines
***  whether or not the server uses the standard C library's memory routines,
***  and there are some memory management routines declared at the end of
***  fundefs.h
***
***    I initialise two strings, system_image which contains the name of
***  the system image to be booted into the transputer, and a temporary
***  string, and I set up a jump buffer. This setjmp() allows all the other
***  initialisation routines to exit the program safely at any time : exit()
***  is not safe because I may have zapped some interrupt vectors.
***
***    initialise_devices() is in module devices.c, and sets up whatever
***  devices are supported by the Server on the current hardware.
***  init_debug() is in module debug.c, and initialises some bits and pieces
***  of the debugging system.
***
***  If the implementation of the Server supports Gem, and Gem is currently
***  loaded, then there is a good chance that the argument vector is junk
***  as I discovered to my great surprise. Hence I need a special test to
***  avoid analysing the command line.
***
***    Otherwise I analyse the command line used to activate the server. There
***  are a number of command line options:
***      -Cfilename : this specifies a configuration file other than the
***                   default HOST.CON in the current directory.
***      -D         : enter system in debugger mode.
***      -{flags}   : set the specified debugging flags. The header file
***                   debugopt.h contains details of the debugging flags
***                   currently supported.
***      filename   : this specifies that some system image other than the
***                   one specified in HOST.CON is to be booted into the
***                   transputer.
***
***     Once the command line has been analysed I know the name of the
***  configuration file and I can read it in. This is done by routine
***  read_config(), in this module. The configuration file usually specifies
***  a system image, but this may have been overruled by the command line.
***  Some system image must have been specified either in the configuration
***  file or on the command line.
***
***     By now I have done all my once-off initialisation calls, and I can
***  activate either the Server or the debugger depending on whether or not
***  the user specified debugging mode. It is possible to reboot the Server
***  and to switch between the Server and the debugger many times, so the
***  activation is done in an infinite loop. I declare a string which will
***  clear the screen, and this string is output if the Server is entered.
***  If the debugger is entered there may be some useful information still
***  on the screen so I do now want to clear it. It is important to reset
***  all the devices by calling restart_devices() in module devices.c, e.g.
***  to flush out characters still in input buffers. It is also important to
***  output a copyright message, for obvious reasons. When entering the
***  Server, all screen output should go via routines output() or outputch()
***  in module devices.c, because this module is responsible for changing
***  ANSI escape sequences (the Helios standard) into the local escape
***  sequences if possible. As this may involve saving data off-screen, it
***  is important that the device driver knows about everything that is on
***  the screen.
***
***    Entering the Server involves booting up the root transputer and some
***  strange message interaction, see module tload.c for details of
***  boot_transputer() and server_helios(). When the transputer has been
***  booted its kernel and system tasks will be spending some time
***  initialising themselves, so I can use some time initialising myself
***  for Server mode. Server() can be found further down in this module.
***
***    The main loop is exited only when the user explicitly wants to exit.
***  I restore all the devices to sensible behaviour, and free the space taken
***  up by configuration file information and other odds and ends.
***
***    It is desirable for the server to exit tidily with a decent error code,
***  slightly tricky given all the longjmp'ing going on. However, exit always
***  involves going via the bit at the end of main() (unless memory
***  initialisation fails).
**/

#if ST
WORD _stksize = 20000L;       /* should be plenty */
#endif

#if memwatch
#if (PC)
PRIVATE WORD *memory_buffer;
#define memsize 1024
#endif /* PC */
#endif /* memwatch */

PRIVATE int return_code;

void main(argc, argv)
int argc;
char *argv[];
{ int i, j, k;
  char tempname[80], *curr_arg, *configname="host.con";

#if AMIGA
  extern int (*_ONBREAK)();
  extern int mybreak();
  extern void DebugMenu(); /* mean to say no menu for the time being */
  extern void ServerMenu(); 
  _ONBREAK = mybreak;	/* convince Lattice not to use their requesters */
  			/* cause more crashes than anything else */
#endif

#if (memwatch)
#if (PC)
   { memory_buffer = (WORD *) malloc(memsize);
   	 if (memory_buffer ne (WORD *) NULL)
      { WORD *ptr1, *ptr2; uint i;
        ptr1 = (WORD *) 0;
        ptr2 = memory_buffer;
        for (i = 0; i < memsize; i+= 4)
          *ptr2++ = *ptr1++;
      }
   }
#endif  /* PC */
#endif /* memwatch */

#if (PC)
     /* These are set up by the library but never used. */
     /* Closing them gives me some more streams.        */
  fclose(stdaux); fclose(stdprn);
#endif

#if use_own_memory_management
    initialise_memory();
#endif

  system_image[0] = '\0';
  tempname[0] = '\0';

  if ((return_code = setjmp(exit_jmpbuf)) ne 0) goto endpoint;

#if gem_supported
#if PC
  { bool i;
    Gem_Testfun(&i);
    if (i) goto skipargs;
  }
#endif
#endif
                            /* Examine the command line arguments.*/
  for (i=1; i<argc; i++)
    { curr_arg = argv[i];
      if (curr_arg[0] eq '-')
       if ((curr_arg[1] eq 'C') || (curr_arg[1] eq 'c'))
        { if (curr_arg[2] ne '\0')
           configname = &(curr_arg[2]);
          elif (++i < argc)
           configname = argv[i];
        }
#if debugger_incorporated
       elif ((curr_arg[1] eq 'D') || (curr_arg[1] eq 'd'))
           DebugMode = 1;
#endif
       elif ((curr_arg[1] eq 'A') || (curr_arg[1] eq 'a'))
        debugflags = (0xFFFFFFFFL & ~(Memory_Flag + Log_Flag));
       else
           { for (j=1; curr_arg[j] ne '\0'; j++)
              for (k=0; options_list[k].flagchar ne '\0'; k++)
                { 
                  if (((int) curr_arg[j]) eq options_list[k].flagchar)
                    debugflags |= options_list[k].flag; 
                }
           }
      else
       strcpy(tempname, curr_arg);   /* argument specifies system image */
    }

#if PC
skipargs:
#endif

  read_config(configname);

  initialise_devices();

  init_logger();

  init_boot();

  init_main_message();         /* in module tload.c */

#if debugger_incorporated
  init_debug();             /* This should be done once only.     */
#endif

  if (strlen(tempname) ne 0)     /* command line argument takes preference */
    strcpy(system_image, tempname);

  if (strlen(system_image) eq 0)
    { ServerDebug("No transputer program specified.");
      longjmp(exit_jmpbuf,1);
    }

  forever
   { PRIVATE UBYTE ANSIclear[] = { (UBYTE) 0x9B, '1', ';', '1', 'H',
                                   (UBYTE) 0x9B, 'J', '\0' };
     restart_devices();
#if debugger_incorporated
     if (!DebugMode)
#endif
        { output(ANSIclear);
          output(SccsId);
          output("All rights reserved.\r\n");
        }
#if debugger_incorporated
     else
      output("Helios Transputer Debugger.\r\nCopyright (C) 1987, Perihelion Software Ltd.\n\r");

     if (DebugMode)
       { int old_log = log_dest;
         int result;
         debugflags &= ~Keyboard_Flag;
         log_dest    = Log_to_screen;
#if AMIGA
	 DebugMenu();
#endif
         result      = debug();
         log_dest    = old_log;
         if (result eq 0) break;
       }
     else
#endif
      { if ((log_dest eq Log_to_file) || (log_dest eq Log_to_both))
          write_to_log("\r\n***\r\n*** Reboot\r\n***\r\n\n");

#if AMIGA
        ServerMenu();
#endif
        boot_processor(serverboot);
#if debugger_incorporated
        if (DebugMode) continue;
#endif
        if (Special_Reboot) continue;
        if (Server() eq 0) break;
      }
   }

endpoint:
  if (debugflags & Quit_Flag)
   ServerDebug("Restoring devices.");
  restore_devices();

  if (debugflags & Quit_Flag)
   printf("Freeing configuration information.\r\n");

  tidy_config();

  if (debugflags & Quit_Flag)
   printf("Freeing main message buffer.\r\n");

  free_main_message();            /* see tload.c */  

#if debugger_incorporated
  if (debugflags & Quit_Flag)
   printf("Tidying debugger.\r\n");

  tidy_debug();
#endif

  if (debugflags & Quit_Flag)
   printf("Tidying bootstrap.\r\n");

  tidy_boot();

  if (debugflags & Quit_Flag)
   printf("Terminating logger.\r\n");

  tidy_logger();

  if (debugflags & Quit_Flag)
   printf("Server exiting.\r\n");

  exit(return_code);
}

/*------------------------------------------------------------------------
--
-- The system configuration file.
--
------------------------------------------------------------------------*/

/**
***   The Server runs on a variety of different hardware configurations,
***  on different hosts and on different transputer boxes for a given host.
***  The Server can adapt to this, provided it knows exactly what the
***  hardware is. This is the purpose of the configuration file.
***
***   A configuration file consists of a series of strings like :
***      box = ATW
***  These strings resemble C environment strings. The following bits of code
***  are responsible for reading in the configuration and giving details of
***  the configuration to other parts of the system.
***
***   The configuration information is stored in a linked list, so I declare
***  a suitable data type and list header. read_config() is called by main()
***  and reads in the configuration file, one line at a time. I assume that
***  each entry in the configuration file is on a separate line. Storing the
***  information in the linked list is trivial. For details of the linked
***  lists see module cofuns.c .
***
***   Once the information has all been read in, I analyse a few bits of it.
***  Some items are essential in all configuration files, so I test for these,
***  and some other items are almost certainly present.
***
***   tidy_config() is called when the system exits, and frees all the memory
***  taken up by the configuration linked list. This can be done conveniently
***  using WalkList().
***
***   When some bit of the system wants to know something about the
***  configuration it can call get_config() with a suitable string.
***  Get_config() wanders down the linked list until it finds an entry
***  which matches the search string, and returns the appropriate part of
***  this entry. Hence it operates in a similar way to the C routine
***  getenv(). There is a related routine get_int_config() which takes a
***  configuration entry and converts it to a word, useful for numeric
***  arguments. Get_int_config() returns Invalid_config to indicate failure,
***  and can deal with hexadecimal and octal data.
**/ 


                        /* This is the list used to hold the configuration. */
PRIVATE List config_hdr;
typedef struct { Node config_node;
                 char name[1];         /* space for byte '\0' */
} config;
PRIVATE int config_initialised = 0;

PRIVATE void read_config(config_name)
char *config_name;
{ FILE   *conf;
  char   *buff = (char *) &(misc_buffer1[0]);
  int    length;
  config *tempnode;
  char   *temp;
  int    error=0;

  InitList(&config_hdr);

  config_initialised++;                /* There may be memory to be freed */

                                             /* open the configuration file */
  conf = fopen(config_name, "r");
  if (conf eq NULL)
    {
      printf("Unable to open configuration file %s\n", config_name);
      longjmp(exit_jmpbuf, 1);
    }

                                      /* and read it in, one line at a time */
  while(fgets(buff, 255, conf) ne NULL)
     { length = strlen(buff);
       if (length eq 0) continue;     /* blank line */
                        /* fgets() leaves a newline character in the buffer */
       if ((buff[length-1] eq '\n') || (buff[length-1] eq '\r'))
         buff[--length] = '\0';

       tempnode = (config *) malloc(sizeof(config) + length);
       if (tempnode eq (config *) NULL)
         { printf("Insufficient memory on host machine for configuration.\n");
           fclose(conf);
           longjmp(exit_jmpbuf, 1);
         }
       strcpy(tempnode->name, buff);
       AddTail((Node *) tempnode, &config_hdr);
     }

  fclose(conf);

  temp = get_config("HOST");           /* There must be an entry for HOST */
  if (temp eq NULL)
    { printf("Error in configuration file, missing entry for HOST\n");
      error = 1;
    }

  temp = get_config("BOX");                         /* And an entry for BOX */
  if (temp eq NULL)
    { printf("Error in configuration file : missing entry for BOX.\n");
      error = 1;
    }

  temp = get_config("SYSTEM_IMAGE");
  if (temp ne NULL) strcpy(system_image, temp);

  maxdata = get_int_config("MESSAGE_LIMIT");
  if (maxdata eq Invalid_config)
    maxdata = 2000L;     /* a sensible default ? */
  elif (maxdata < (word)(IOCDataMax + 100))
    maxdata = (word)(IOCDataMax + 100);
  elif (maxdata > 64000L)        /* enforce an upper limit */
   maxdata = 64000L;

  if (error)
    { printf("\nPlease edit the configuration file %s.\n", config_name);
      longjmp(exit_jmpbuf, 1);
    }
}

PRIVATE void tidy_config()
{ if (config_initialised) FreeList(&config_hdr);
}

        /* The following code allows the server to examine the   */
        /* configuration file. For example, to get the entry for */
        /* HOST, use a call get_config("host"). If there is      */
        /* an entry HOST=XXX get_config() returns a pointer to   */
        /* XXX. The function is fairly easy to implement using a */
        /* list Wander.                                          */
        /* If no entry is found the routine returns NULL.        */
PRIVATE char *test_name(node, name)
config *node;
char   *name;
{ char *config_name = &(node->name[0]);

                                  /* compare name and config_name */
  while ((*name ne '\0') && (*config_name ne '\0'))
   { if ( ToUpper(*name) ne ToUpper(*config_name) )
       return(FALSE);
     name++; config_name++;
   }
                    /* succeeded ? */
  while (isspace(*config_name)) config_name++;

  if ((*name eq '\0') && (*config_name eq '\0')) return(config_name);

  if ((*name eq '\0') && (*config_name++ eq '='))
    {   while (isspace(*config_name)) config_name++;
        return(config_name);
    }
  else
    return(FALSE);
}

char *get_config(str)
char *str;
{ word result = Wander(&config_hdr, (WordFnPtr) func(test_name), str);
  if (result eq FALSE)
    return(NULL);
  else
    return((char *) result);
}

WORD get_int_config(str)
char *str;
{ char *result = get_config(str);
  WORD base = 10L, value = 0L, sign = 1L;
  WORD temp;

  if (result eq NULL)
    return(Invalid_config);

  while (isspace(*result)) result++;
  if (*result eq '-') { sign = -1L; result++; }
  if (*result eq '0')
    { result++;
      if (*result eq 'x' || *result eq 'X')
        { result++; base = 16L; }
      else
       base = 8L;
    }
  if (*result eq '-') { sign = -1L; result++; }


#if (SUN || PC)
  while (isxdigit(*result))    /* Use this routine if it is available */
#else
  while ( isdigit(*result) || ('a' <= *result && *result <= 'f') ||
          ('A' <= *result && *result <= 'F'))
#endif
   { 
     switch (*result)
      { case '0' : case '1' : case '2' : case '3' : case '4' :
        case '5' : case '6' : case '7' : case '8' : case '9' :
                      temp = (WORD) (*result - '0'); break;

        case 'a' : case 'b' : case 'c' : case 'd' : case 'e' :
        case 'f' :    temp = 10L + (WORD) (*result - 'a'); break;

        case 'A' : case 'B' : case 'C' : case 'D' : case 'E' :
        case 'F' :    temp = 10L + (WORD) (*result - 'A'); break;
      }

     if (temp >= base) break;
     value = (base * value) + temp;
     result++;
   }

 return(value * sign);
}

/*------------------------------------------------------------------------
--
--  Server - main routine.
--
------------------------------------------------------------------------*/

/**
***   Routine Server() is called by main() when the user is in Server mode
***  and when the transputer has just been booted up. It returns zero if
***  the system should exit, or non-zero if the system should reboot or 
***  enter debugging mode.
**/

PRIVATE int Server()
{ int i=0;                /* If I fail to initialise I do not try to reboot */

  if (Init())             /* Init() sets up lists and basic coroutines */
    i = MainLoop();
  else
    ServerDebug("Server unable to initialise.");

  TidyUp();               /* clean up the Server system. */

  return(i);
}


/*------------------------------------------------------------------------
--
--  Initialise the I/O system
--
------------------------------------------------------------------------*/

/**
*** This routine is responsible for initialising the Server once the
*** transputer has been bootstrapped.
***
*** First I call InitColib() to set up the basic coroutine library.
***
*** Next I work out the current system time as a Unix time stamp, i.e.
*** seconds since 1970, using one of the machine-dependant routines.
***
*** I cannot escape any longer - time to explain the coroutine mechanism
*** used in the Server....
***
*** At any time, there will be a number of different devices available such
*** as clock, console, disk drives etc. There will also be a large number of
*** streams open. Each server and each stream has a Helios message port,
*** and programs running on the transputer send requests to these ports. 
*** All the servers and all the streams act independently of each other.
***
*** Now, programming each server and each stream so that they do not
*** interfere with each other, e.g. zapping each other's variables, is not
*** trivial. The solution is coroutines : each server and each stream runs
*** as a separate coroutine in the system, and hence it has its own stack
*** and its own local variables which are safe from the rest of the system.
*** I do not care how sceptical you are - they work for me and they have
*** saved me a lot of trouble.
***
*** The next problem is how to keep track of all these servers and streams.
*** Each coroutine (i.e. server or stream) can be in one of three states :
*** a) running, which means that it has got the CPU to itself, it can do
***    what it want, and there will not be another message coming in at the
***    wrong moment. Coroutines only stop running when they explicitly want
***    to.
*** b) waiting, which means that the coroutine has nothing to do until
***    another message arrives from the transputer.
*** c) polling, which means that the coroutine is handling some existing
***    request from the transputer but has to wait for an event to happen
***    on the host machine, e.g. for the user to press a key. In this state
***    the coroutine cannot accept new messages from the transputer, but
***    the transputer may be trying to send messages to other coroutines and
***    there is no reason why these messages cannot be handled. The polling
***    coroutine has to suspend itself regularly, the Server's main loop will
***    wake it up again, and the coroutine will check whether its event has
***    occurred yet and if not suspend itself again.
***
*** So, I have two linked lists WaitingCo and PollingCo, and a coroutine is
*** either running or it is held in one of these two linked lists. If it
*** is running it can move itself freely between the two lists, but when
*** it suspends itself it must be in one of the two lists.
***
***  A problem with coroutines is that they share all global values, and in
*** particular the message buffer used to communicate with the transputer is
*** global. It is guaranteed that only one coroutine runs at the same time,
*** and that coroutines keep running until they do an explicit Suspend()
*** (barring interrupts ofcourse). Hence coroutines can access global data
*** safely whilst they are running, but before they do a Suspend() they must
*** save global data they want to preserve on the stack or in private bits of
*** the heap. This includes things like reply ports. Also, the main message
*** buffer cannot be used to hold data collected by polling.
***
*** The main message buffer is allocated by a single call to init_main_message(),
*** and all modules have access to a pointer mcb. Having large message
*** buffers, the upper limit being 64K, tends to reduce the time taken to
*** load large files at the cost of reducing the responsiveness of keyboard,
*** mouse etc. The main buffer is freed in TidyUp().
***
*** Back to the code. I create the two lists by calls to MakeHeader() in
*** module cofuns.c, where all the list and coroutine functions are kept,
*** Then I create a server coroutine for each device supported by the Server
*** and for each drive connected to the machine. See header files server.h
*** and fundefs.h for the declarations of all these strange arrays of
*** functions etc. Each coroutine starts running in routine General_Server()
*** later in this module, and General_Server() implements a general purpose
*** Helios server. When the coroutine gets say an Open request
*** General_Server() calls the appropriate functions held in its array of
*** request handlers, e.g. Console_Open(). All the coroutines have a unique
*** integer identifier, which is also the coroutine's Helios message port,
*** and they all have a name to facilitate handling distributes searches
*** described later. Finally all the servers are started so that they can
*** initialise themselves.
***
*** The first device is the IOprocessor, and it is important that this is the
*** first coroutine in the linked list. For details of this server, see module
*** devices.c
***
*** There is a special server called Helios, which is a pseudo-drive. This
*** allows the transputer to access files /Helios/lib/ram etc. without
*** having to know anything about the local filing system. The directory
*** which corresponds to drive Helios, e.g. c:\helios, is part of the
*** configuration file but I provide a default.
***
*** The extra field in each coroutine node specifies the type of the device.
*** This information is used by the IOPROC device when it starts up to
*** work out its linked list of server directory entries.
***
*** The optimal order for the coroutines in the linked list WaitingCo is as
*** follows : ioproc server, helios server, active streams, other servers.
*** This order is important, because each time the Server gets a request from
*** the transputer it has to search through the list to find the coroutine
*** whose id matches the destination port. Obviously when a file is being read
*** its stream coroutine is likely to get more requests than say the mouse
*** server. To keep the coroutines in the right order, stream coroutines are
*** postinserted after the Helios node, and I keep a pointer Heliosnode which
*** is initialised by Helios_InitServer() in module files.c
**/

PRIVATE WORD StreamTimeout;

PRIVATE WORD Init()
{ int i;
  Conode *tempco;

  StreamTimeout = get_int_config("StreamTimeout");
  if (StreamTimeout eq Invalid_config)
   StreamTimeout = DefaultStreamTimeout;
  StreamTimeout  = divlong( StreamTimeout, time_unit);

  Startup_Time = get_unix_time();
  initial_stamp = clock();
  Device_count = 0;

  unless(InitColib())           /* start up the coroutine library */
     return(FALSE);

  WaitingCo = MakeHeader();    /* set up the list structure */
  PollingCo = MakeHeader();
  if((WaitingCo eq NULL) || (PollingCo eq NULL))
      return(FALSE);

                                    /* Find out where the Helios directory is */
  Heliosdir = get_config("Helios_directory");
  if (Heliosdir eq NULL)       /* the default varies from machine to machine */
#if (ST || PC)
      Heliosdir = "c:\\helios";
#endif
#if AMIGA
      Heliosdir = "helios:";
#endif
#if (SUN)
      Heliosdir = "/usr/Helios";
#endif
                                /* create a server coroutine for each device */
  for (i=0; devices[i].name ne NULL; i++) 
    {  
       if (devices[i].handlers[Testfun_off] ne Nullfn)
        { bool x;
          (*devices[i].handlers[Testfun_off])(&x);
          unless(x) continue;
        }
       tempco = NewCo(General_Server);
       unless(tempco) return(FALSE);
       Device_count += 1;
       AddTail(&(tempco->node), WaitingCo); /* Add new server to tail of list */
       tempco->id         = CoCount++;
       tempco->timelimit  = MAXINT;       /* servers never time out */
       strcpy(tempco->name, devices[i].name);
       tempco->handlers   = devices[i].handlers;
       tempco->extra      = (ptr) devices[i].type;
       tempco->flags      = 0L;
    }

#if drives_are_special
                /* And another server coroutine for each drive           */
#if (PC || ST)   
  { WORD   drivemap, mask, floppies;
                /* get_drives() tells me what drives are connected, and  */
                /* afterwards floppies will tell me which of these are   */
                /* floppies */
                /* This area needs a lot of rewriting on other machines. */
                /* Possibly make up a linked list of mounted drives each */
                /* node having the name ?                                */
    drivemap = get_drives(&floppies);

    mask     = 0x00000001L;
    for (i=0; i < 32; i++)
     { if (drivemap & mask)
         { 
           tempco            = NewCo(General_Server);
           unless(tempco) return(FALSE);
           Device_count     += 1;
           AddTail(&(tempco->node), WaitingCo);
           tempco->id        = CoCount++;
           tempco->timelimit = MAXINT;          /* servers never time out */
           tempco->name[0]   = (char) ( (int) 'a' + i);
           tempco->name[1]   = '\0';
           tempco->handlers  = Drive_Handlers;
           tempco->extra     = (ptr) Type_Directory;
           if (floppies & mask)
            tempco->flags    = CoFlag_Floppy;
           else
            tempco->flags    = 0L;
         }
       mask <<= 1;
     }
  }
#endif   /* MSdos_compatible */


#if AMIGA
                /* And another server coroutine for each drive           */
  {
    extern WORD create_drive_list(); /* returns number of drives found */
    extern char * get_drive_name(); /* get next drive name from list */
    extern void free_drive_list();
    WORD nr_of_drives;
    char * tempname;
    
    nr_of_drives = create_drive_list();
    for(i=0;i<nr_of_drives;i++) {
       tempco            = NewCo(General_Server);
       unless(tempco) {
          free_drive_list();
          return(FALSE);
       }
       tempname = get_drive_name();
       Device_count     += 1;
       AddTail((Node *)tempco, WaitingCo);
       tempco->id        = CoCount++;
       tempco->timelimit = MAXINT;          /* servers never time out */
       strcpy(tempco->name,tempname);
       tempco->handlers  = Drive_Handlers;
       tempco->extra     = (ptr) Type_Directory;
       tempco->flags     =  0L;
     }
  }

#endif /* AMIGA */

#endif   /* drives_are_special */

#ifdef SYNERGY
#include "synergy\\synergy.h"
#endif

#if AMIGA
#include "x.h"
#endif

   for (tempco = (Conode *) WaitingCo->head;
        tempco->node.next ne NULL;
        tempco = (Conode *) tempco->node.next)
    { if (debugflags & Init_Flag)
       ServerDebug("Initialising device %s.", &(tempco->name[0]));
      StartCo(tempco); 
    }

#if multi_tasking
   InitMultiwait();             /* Start up multitasking support */
#endif

   if (debugflags & Init_Flag)
    ServerDebug("Server successfully initialised.");

   return(TRUE);
}


/*------------------------------------------------------------------------
--
-- Tidy up the Server
--
------------------------------------------------------------------------*/

/**
*** When the user wishes to reboot the server or enter the debugger, it
*** is important that I do not leave tens of coroutines each with large
*** unfreed stacks. The following little routine should persuade all
*** the coroutines to tidy up and kill themselves, by setting the
*** CoSuicide flag in the coroutine node. Each coroutine has to be programmed
*** fairly carefully so that whenever it suspends itself it checks this
*** suicide bit afterwards, and takes appropriate action. It is distinctly
*** possible that I have not been entirely successful at this, and that you
*** lose a couple of hundred bytes each time you reboot, but this is probably
*** acceptable for the time being. Once I start getting complaints I shall
*** sort it out.
***
*** Having killed off all the server and stream coroutines, I can release the
*** entire coroutine library and the large shared message buffer.
**/

PRIVATE void fn( TerminateCo, (Conode *));

PRIVATE void TidyUp()
{
  if (debugflags & Quit_Flag)
   ServerDebug("Shutting down streams and servers.");

  WalkList(WaitingCo, func(TerminateCo));   /* get rid of the coroutines */
  WalkList(PollingCo, func(TerminateCo));

  free(WaitingCo);   /* and free all memory that has been allocated */
  free(PollingCo);

#if multi_tasking
  TidyMultiwait();
#endif

  if (debugflags & Quit_Flag)
   ServerDebug("Streams and servers shut down.");
}
                      /* try to persuade a coroutine to die off */
PRIVATE void TerminateCo(conode)
Conode *conode;
{ conode->type = CoSuicide;
  StartCo(conode);
}

/*------------------------------------------------------------------------
-- The main loop for the Server. 
------------------------------------------------------------------------*/

/**
*** This is the main loop of the Server. It is an infinite loop, which exits
*** only when the transputer sends a special message or when the user
*** wants to exit, reboot, or enter debugging mode. It returns 0 if the user
*** wants to exit or upon receipt of a quit message from the transputer,
*** non-zero if the system should continue to run.
***
*** First the system clock is updated. An approximation to the current time
*** is needed quite frequently, and it might be expensive to call the
*** appropriate clock routines each time. Hence I use a single variable Now
*** to hold a value roughly corresponding to the current time.
***
*** Next I check some special flags which are usually altered by keyboard
*** interrupt routines or similar low level routines, and take appropriate
*** action.
***
*** At this point I reactivate all the coroutines which are waiting for some
*** event to happen on the host side of the world, e.g. for a key to be
*** pressed. These coroutines are all held in PollingCo, and a simple
*** WalkList does the trick nicely. I must also check all the coroutines
*** waiting for messages to see if they have timed out : Helios streams
*** time out if they receive no messages for ??? minutes. Since I have
*** already activated all the PollingCo routines I do not bother to check
*** these for timeouts - they can do it themselves.
***
*** Poll_the_devices() in module devices.c does exactly that : it performs
*** whatever checks of the systems are desirable, checks which may be
*** inappropriate for a server or stream to perform. For example, console
*** coroutines should not access the keyboard directly because there may
*** be multiple streams trying to do that, so poll_the_devices() does this
*** and stores the data somewhere where it can be accessed safely by the
*** coroutines.
***
*** This has finished all the work I MUST do each time around the loop.
*** Now the question is : is the transputer trying to send me
*** some more work to do ? If not, I can just go around my main loop again,
*** updating the time stamp, reactivating polling coroutines, and so on.
*** However, if there is a message I had better get hold of it and process it.
*** And this documentation is becoming rather silly.
***
*** If the message debugging flags is set, I should print out details of the
*** message for the user to admire. Then I check for a couple of special
*** messages. ReqDie is a quit request, a message from the transputer to tell
*** the server to exit.  The message may be an error message,
*** indicating that e.g. a message to a transputer at the other end of a long
*** pipeline was lost due to congestion. I have no idea what to do with such
*** messages, so I ignore them.
***
*** If a message has got this far, it must be a message for an existing
*** server or stream. This means it has been sent to a message port, which
*** is really a coroutine identifier. Therefore I check all my coroutines
*** to see which one is meant to be activated. It may be a message for a
*** coroutine which is waiting for one, in which case there is no problem and
*** I activate the coroutine. It may be a message for a coroutine that is still
*** handling an old request, in which case I cannot accept the new message.
*** It may be a message for a coroutine that has already died, e.g. because
*** of a five-minute timeout, and this can be detected using the coroutine
*** counter. Finally it may be a message for a completely wrong port.
*** Whatever the case, I send back an appropriate error code. Routine
*** Request_Return() in module tload.c takes care of sending messages
*** back to the transputer.
**/

PRIVATE   void fn( timeout_check, (Conode *));
PRIVATE   void fn( polling_check, (Conode *));
PRIVATE   Conode *fn(CheckPkt, (Conode *, WORD));

PRIVATE int MainLoop()
{ Conode    *result;
  word      tempport;

#if multi_tasking
  AddMultiwait(NULL, Multi_LinkMessage);
#endif

  for (;;)          /* infinite loop to fetch requests */
   { 
     Now = clock();

#if (memwatch)
#if (PC)
     { PRIVATE int memcount = 0;
       if ((memcount eq 0) && (memory_buffer ne (WORD *) NULL))
        { WORD *ptr1, *ptr2; uint i;
          ptr1 = (WORD *) 0; ptr2 = memory_buffer;
          for (i = 0; i < memsize; i+= 4, ptr1++, ptr2++)
           if (*ptr1 ne *ptr2)
            { ServerDebug("Memwatch alert : %x) was %lx, is %lx", i, *ptr2, *ptr1);
              *ptr2 = *ptr1;
            }
	    }
       memcount = (memcount + 1) & 0x3ff;
     }
#endif /* PC */
#endif /* memwatch */

     if (Special_Exit)                  /* check for exit key pressed */
        return(0);
     if (Special_Status)                /* or for status request */
      { output("Server alive and well. Logging goes to ");
        if (log_dest eq Log_to_screen)
         output("screen");
        elif (log_dest eq Log_to_file)
         output("file");
        elif (log_dest eq Log_to_both)
         output("screen and file");
        else
         { log_dest = Log_to_screen; output("screen"); }
        output(".\r\n");
        Special_Status = FALSE;
      }
     if (Special_Reboot)               /* or for reboot condition */
      { Special_Reboot = FALSE;
        output("\014");                /* Clear server window */
        return(1);
      }
#if debugger_incorporated
     if (DebugMode) return(1);
#endif

     if ( debugflags & Log_Flag)
      { debugflags ^= Log_Flag;
        if (log_dest eq Log_to_screen)
         { log_dest = Log_to_file;
           output("Switching logging to file only.\r\n");
         }
        elif (log_dest eq Log_to_file)
         { log_dest = Log_to_both;
           output("Switching logging to file and screen.\r\n");
         }
        else
         { log_dest = Log_to_screen;
           output("Switching logging to screen only.\r\n");
         }
      }

     if ( debugflags & Memory_Flag)
      { int    total, streams;
        Conode *mynode;
#if use_own_memory_management
        memory_map();
#endif
        total = 0; streams = 0;
        ServerDebug("Open streams are : %q");

        for (mynode = (Conode *) WaitingCo->head;
             mynode->node.next ne (Node *) NULL;
             mynode = (Conode *) mynode->node.next)
         { total++;
           if (mynode->name[0] eq 0x7F)
            { if (streams++) ServerDebug(", %q");
              ServerDebug("%s%q", &(mynode->name[1]));
            }
         }

        for (mynode = (Conode *) PollingCo->head;
             mynode->node.next ne (Node *) NULL;
             mynode = (Conode *) mynode->node.next)
         { total++;
           if (mynode->name[0] eq 0x7F)
            { if (streams++) ServerDebug(", %q");
              ServerDebug("%s%q",&(mynode->name[1]));
            }
         }

        ServerDebug(
       "\r\nThere are %d devices, %d open streams, and %d other coroutines.",
                Device_count, streams, total - (Device_count + streams) );
        debugflags &= ~Memory_Flag;
      }

                                    /* wake up any coroutines that may */
     WalkList(PollingCo, func(polling_check)); /* be waiting */

                                    /* allow stream coroutines to die if */
                                    /* no message has arrived in time    */
     WalkList(WaitingCo, func(timeout_check));
     WalkList(PollingCo, func(timeout_check));

#if multi_tasking
    if (!Multiwait(1)) continue;
#endif /* multi_tasking */

     poll_the_devices();           /* in devices.c */

     unless(Request_Stat())         /* is there a new request ? */
       continue;                    /* No, so more polling. */

     tempport = (mcb->MsgHdr).Dest;
     (mcb->MsgHdr).Dest = NullPort;

     if (debugflags & Message_Flag)
       ServerDebug("Request: fn %lx, for %lx from %lx, csize %d, dsize %ld",
                  (mcb->MsgHdr).FnRc, tempport, (mcb->MsgHdr).Reply,
                  (mcb->MsgHdr).ContSize, (WORD) (mcb->MsgHdr).DataSize ); 

     if ((mcb->MsgHdr).FnRc eq ReqDie)
       return(0);                        /* Transputer can kill server */

                      /* I want to ignore error messages for the time being */
     if (((mcb->MsgHdr).FnRc & 0x80000000L) ne 0L)
       continue;
     
                                                /* who is the request for ? */
     result = (Conode *)Wander(WaitingCo, (WordFnPtr) func(CheckPkt), tempport);
     if (result)
       StartCo(result);    /* for a coroutine that is waiting for a message */
     else                         
       { result = (Conode *) Wander(PollingCo, (WordFnPtr) func(CheckPkt), tempport);
         if (result)                          /* coroutine is already busy */
          { if ((mcb->MsgHdr.FnRc & FG_Mask) eq FG_Close)
             result->type = CoSuicide;  /* Kill it off, even though it is busy */
            Request_Return(EC_Recover + SS_IOProc + EG_InUse + EO_Port,
                          0L, 0L);
          }
         else                                 /* coroutine does not exist */
           Request_Return(EC_Warn + SS_IOProc + EG_Unknown + EO_Port,
                          0L, 0L);
        /* I do not distinguish anymore between ports that never existed and  */
        /* ports that have died off. */
       }
   }
}

PRIVATE void timeout_check(conode)
Conode *conode;
{ if (conode->timelimit < Now)
    { conode->type = CoTimeout;
      StartCo(conode);
    }
}

PRIVATE void polling_check(conode)
Conode *conode;
{ /* if (conode->type eq CoReady) future extension */
   StartCo(conode);
}

             /* used to see whether a message is for an existing coroutine */
Conode *CheckPkt(conode, id)
Conode *conode;
word id;
{ if (conode->id eq id)
    return(conode);
  else
    return((Conode *) FALSE);
}

/*------------------------------------------------------------------------
--
--  A general purpose server for handling GSP requests.
--
------------------------------------------------------------------------*/

/**
*** The following code implements a general-purpose server suitable for all
*** devices and drives. This function always runs as a separate coroutine,
*** with its own stack, etc. When the transputer sends a message to this
*** server, the message is received from the link by the main Server coroutine
*** in routine main_loop(), and the server coroutine is reactivated.
***
*** The coroutine is started up by NewCo() in module cofuns.c when the
*** coroutine is created. Myco is the coroutine list node containing all
*** the required information, including a table of handler routines,
*** and I can call the specific initialisation routine for server.
*** After this the server is ready to receive messages from the transputer,
*** so the coroutine suspends itself to wait for one.
***
*** When the server is woken up again, either the system is supposed to die
*** or a message has been received. It is impossible for servers to time out,
*** so I do not need to check for that. If the server got a message, it may
*** be a private protocol message, or a General Server Protocol or GSP message,
*** which can be checked by looking at some of the bits in the function code,
*** and the GSP function code has to be in the right range for a server, to
*** make sure that the server does not try to handle stream requests.
***
*** All GSP requests to a server involve a name of some sort, and it is
*** necessary to convert the Helios naming scheme to a local name. Ofcourse the
*** name supplied may be complete junk, so I have to allow for failure here.
***
*** Having passed all these hurdles, I can now call the handling routine
*** appropriate for the request, by indirecting through the table of handlers.
**/

PRIVATE void General_Server(myco)
Conode *myco;
{ 
  int    handler_off;

  (*(myco->handlers)[InitServer_off])(myco);

  forever
   { Suspend();              /* wait for the next request        */
     if (myco->type eq CoSuicide)    /* which may be an order to die        */
       { (*(myco->handlers)[TidyServer_off])(myco);
         Seppuku();
       }
                             /* The message may be part of a private protocol */
     if (((mcb->MsgHdr).FnRc & FC_Mask) ne FC_GSP)
       { (*((myco->handlers)[Private_off]))(myco);
         continue;
       }
                             /* Check the range of the GSP request */ 
     unless((FG_Open <= getfnrc(mcb)) && (getfnrc(mcb) <= FG_CloseObj))
      { Request_Return(EC_Error + SS_IOProc + EG_FnCode + EO_Message, 0L, 0L);
        continue;
      }

                                       /* and protect us from Nick Garnett */
     mcb->Data[mcb->MsgHdr.DataSize] = '\0';

                                           /* work out the local name IOname  */
     unless(convert_name())
      { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
        continue;
      } 
/**
*** At this stage the function code is known to be a GSP request between 0x10
*** for Open and 0xC0 for CloseObj, so I mask off the useless bits, and use the
*** remainder as an array index.
**/
    handler_off = ((int) (mcb->MsgHdr.FnRc & 0x000000F0L)) >> 4;
    (*(myco->handlers)[Open_off - 1 + handler_off])(myco);
   }
}

/*------------------------------------------------------------------------
--
--  A similar routine for handling streams.
--
------------------------------------------------------------------------*/

/**
*** The code below is very similar to the above General_Server(), but it
*** is for handling stream coroutines rather than servers. Whenever a
*** server receives a valid Open request, e.g. to open a file, it creates
*** a new coroutine for the corresponding Helios stream, which has its own
*** message port/coroutine identifier.
***
*** There are some differences. The main difference is that opening a stream
*** is not always possible, e.g. the server may have received a request to
*** open a non-existant file in read-only mode. Routine General_Stream() is
*** responsible for this, because it knows all about its coroutine nodes etc.
*** unlike the server. This works by calling an init_stream handler which
*** forms part of the usual array of functions. Ofcourse the program on the
*** transputer which attempted to Open the stream needs to be informed, so
*** it is the responsibility of General_Stream() to reply to the Open
*** request and it has to know the address of the Open message. Most of
*** the reply's data structure has been set up already by the server, on 
*** the assumption that the request was valid.
***
*** After initialisation the coroutine is ready to handle GSP stream
*** requests, so there is the usual Suspend() to suspend the coroutine.
*** Streams can time out, so I have to check the timeout flag as well
*** as the suicide flag. Also, tidying up the world is rather more
*** complicated for a stream than for a server as there are files to be
*** closed etc.
**/

PRIVATE void General_Stream(myco)
Conode *myco;
{ 
  WORD   temp;
  int    handler_off;

                              /* start up the stream using the Init routine */
  temp = ( (*( (WordFnPtr)myco->handlers[InitStream_off])) (myco) );
  if (temp)
    { Request_Return(temp, 0L, 0L);     /* The Open failed for some reason */
      Seppuku();
    }
                             /* now the stream is ready for stream requests */
                             /* so I reply to the Open request              */
  Request_Return(ReplyOK, open_reply,
                 (WORD) strlen(mcb->Data) + 1L);

  forever
   { myco->timelimit   = Now + StreamTimeout;
     myco->type        = 0L; 
     Suspend();                         /* wait for the next request          */
     if (myco->type eq CoSuicide)       /* which may be an order to die       */
       {                                /* do a tidy up     */
         (*myco->handlers[TidyStream_off])(myco);
         Seppuku();
       }
     elif (myco->type eq CoTimeout)     /* or a timeout may have occurred     */
       { 
         if ((*((WordFnPtr)myco->handlers[TidyStream_off]))(myco))
             continue;             /* Do not die off for some reason or other */
                                   /* e.g. because there is an event handler  */
         Seppuku();
       }

                             /* The message may be part of a private protocol */
     if (((mcb->MsgHdr).FnRc & FC_Mask) ne FC_GSP)
       { (*((myco->handlers)[StreamPrivate_off]))(myco);
         continue;
       }

     unless((FG_Read <= getfnrc(mcb)) && (getfnrc(mcb) <= FG_NegAcknowledge))
      { Request_Return(EC_Error + SS_IOProc + EG_FnCode + EO_Message, 0L, 0L);
        continue;
      }

/**
*** At this stage the function code is known to be a GSP request between 0x1010
*** for Read and 0x10B0 for NegAck, so I mask off the useless bits, and use the
*** remainder as an array index.
**/
    handler_off = ((int) (mcb->MsgHdr.FnRc & 0x000000F0L)) >> 4;
    (*(myco->handlers)[Read_off - 1 + handler_off])(myco);
   }
}


/*------------------------------------------------------------------------
--------------------------------------------------------------------------
--
-- The IO processor device
--
--------------------------------------------------------------------------
------------------------------------------------------------------------*/

/**
*** The IOprocessor is a node in the Helios network. This means that it must
*** have a name, e.g. /IO. All devices are servers inside this network node.
*** Hence the network name of the console device starts off as /IO/console.
***
*** It is possible to perform a number of operations on the entire network node,
*** i.e. on the device IOproc. For example, you should be able to list the
*** node to find out all the servers supported by that node, and you can
*** rename the node when the network changes. Note that it is possible to
*** access a network node in two different ways: via its network name, or
*** via the links. Thus listing /IO and listing /00/Link.0 should have the
*** same effect. This complicates matters slightly for the Server, because
*** if you access /00/Link.0 there are no bits of the name left when the
*** message gets to me, whereas if you access /IO the name I get is IO. This
*** means I have to do a number of special checks.
***
*** The network node may receive distributed search requests for servers that
*** the rest of the network does not know about yet. These are sent only to
*** IOproc, not to any servers inside the network node. However, it is possible
*** for Helios to send server requests to IO when they are meant for some other
*** server inside IO, on the assumption that IOproc will forward them to the
*** appropriate server. 
**/
PRIVATE void fn( Forward, (void));
PRIVATE char network_name[IOCDataMax];

/**
*** Forward_handler() is a default handler for many IOproc requests. The request
*** is inappropriate if sent directly to IOproc, but it may be fine for another
*** server. Hence I need to check that it is for another server and forward it.
***
*** The routine always executes in the IOproc coroutine, so myco->name holds the
*** current local name.
**/

void Forward_handler(myco)
Conode *myco;
{
                 /* Messages arriving at the IOproc device can come either by */
                 /* accessing the device or by accessing a link. If the latter*/
                 /* the local name is nothing. This has to be treated as a    */
                 /* special case.                                             */
  if (strcmp(IOname, myco->name) && IOname[0] ne '\0' )
    Forward();
  else
    Request_Return(EC_Error + SS_IOProc + EG_WrongFn + EO_Server, 0L, 0L);

use(myco)
}

/**
*** If a message arrives at the IOproc device that is intended for another
*** device it should be forwarded to that device. This only affects server
*** requests so the local name is something like IO/d/helios/bin/shell.
*** I need to extract the appropriate server, i.e. the server for drive D,
*** and work out which is the corresponding coroutine. This is handled by
*** do_search() which actually forms part of the distributed search mechanism.
*** Forwarding a message involves reactivating the server coroutine, since the
*** single global message buffer is shared. The coroutine will start
*** up inside General_Server() because it is waiting for a message. It will
*** perform a new name conversion automatically which will cause IOname to
*** become d/helios/bin/shell, and everything is fine. When the other
*** coroutine suspends itself this one is reactived, and the routine returns
*** into General_Server() where the coroutine suspends itself, reactivating
*** the main coroutine which goes around MainLoop again.
***
*** NOTE : it may be desirable to extend this, so that internal Servers can
*** forward messages e.g. to allow console/../helios/x as a valid name, or
*** even to forward messages to other bits of the network.
**/

PRIVATE WORD fn( do_search, (Conode *, char *));
PRIVATE void Forward()
{ char *server_name, *server_end;
  Conode *tempco;

  for (server_name = IOname; *server_name ne '/'; server_name++);
  server_name++;

  for (server_end = server_name; *server_end ne '/' && *server_end ne '\0';
        server_end++ );

  *server_end = '\0';

  if ((tempco = (Conode *) Wander(WaitingCo, func(do_search), server_name) )
      ne (Conode *) FALSE)
   if (tempco->id ne 1L)                   /* Cannot forward to IOPROC device */
    { 
      StartCo(tempco);
      return;
    }

   Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Message, 0L, 0L);
}

/**
*** The main thing that can be done with the IOproc device is listing it to
*** find out what servers are supported by the Server. This involves Open
*** requests and Reads of DirEntry. The servers supported are generally fixed,
*** so I can set up my directory structure when the IOproc device is given a
*** chance to initialise. This directory structure is the same as that used
*** for real directories in module files.c, which means that the code for
*** reading the IOproc directory is shared with the code for the reading of
*** file directories. All that is handled by the declarations in fundefs.h
***
*** Working out what servers exist is very easy, because the list WaitingCo
*** contains all the server coroutines and they are named in the Conode
*** structure. The IOproc device is the first to initialise itself, and via
*** some hacking in Init() above the extra field in each conode contains the
*** type of the device, i.e. file, directory or special.
***
*** InitServer() is also responsible for setting up the initial network name,
*** which is always /IO. The IOproc device may receive Rename requests from
*** the network server, which will affect this network name.
***
*** It is essential that the IOproc device has a constant message port even
*** across reboots, and this port is 1L. The configuration info sent in module
*** tload.c includes this port id.
**/

DirHeader IO_extra;

void IOPROC_InitServer(myco)
Conode *myco;
{ Conode *curnode  = (Conode *) myco->node.next;
  char *proc_name  = get_config("io_processor");

  if (proc_name ne (char *) NULL)         /* change the ioprocessors name */
   { strcpy(myco->name, &(proc_name[1]));
     strcpy(network_name, proc_name);
   }
  else
   strcpy(network_name, slashDefaultServerName);  /* Initial network name */
 
  myco->id         = 1L;
  myco->extra      = (ptr) &IO_extra;
  InitList(&IO_extra.list);
  IO_extra.entries = 0L;

  for ( ; curnode->node.next ne (Conode *) NULL;
          curnode = (Conode *) curnode->node.next)
   { DirEntryNode *newnode = (DirEntryNode *) malloc(sizeof(DirEntryNode));
     WORD type = (WORD) curnode->extra;

     if (newnode eq (DirEntryNode *) NULL)
       { ServerDebug("Insufficient memory to initialise IOProc device.");
         longjmp(exit_jmpbuf, 1);
       }

     clear_bytes((char *) newnode, sizeof(DirEntryNode));
     (newnode->direntry).Type     = swap(type);
     (newnode->direntry).Flags    = swap(0L);
     (newnode->direntry).Matrix   = swap( (type eq Type_Directory) ? 
                                       DefDirMatrix : DefFileMatrix  );
     strcpy(newnode->direntry.Name, curnode->name);

     AddTail(&(newnode->node), &IO_extra.list);
     IO_extra.entries++;
   }
}

void IOPROC_TidyServer(myco)
Conode *myco;
{ FreeList(&IO_extra.list);
  use(myco)
}

/*------------------------------------------------------------------------
--
--  code to handle distributed search requests.
--
------------------------------------------------------------------------*/

/**
*** All distributed search requests are sent to the IOproc device, using a
*** private protocol. Distributed search is the only private protocol
*** understood by the IOproc device, so that makes life slightly easier.
*** Essentially, when I get a search request the data vector contains a
*** C string identifying the server Helios is trying to get hold off.
*** Distributed searches are always for servers.
***
*** Needless to say there are a couple of complications. First of all, the
*** string may be rather complicated and I need to extract the bit I am
*** interested in. Unfortunately nobody has told me exactly what the naming
*** scheme is, so I have done some guess work on what I have to do. Secondly,
*** there is a debugging option to handle searches, so I may have to display
*** a message. Note that I only need to check WaitingCo, and not PollingCo,
*** because servers will never suspend themselves in a polling loop...
***
*** As for what I return to the transputer, if I know about the name I send
*** back a message port for the server so that the server can get messages
*** from the transputer, and this port is the unique coroutine identifier.
*** Otherwise I return a suitable error code. I am also meant to return a
*** full network name for the server.
**/

PRIVATE void fn( handle_search, (void));
void IOPROC_Private(myco)
Conode *myco;
{ if (getfnrc(mcb) eq FG_Search)      /* check for distributed search */
    handle_search();          /* request */
  else
    Request_Return(EC_Error + SS_IOProc + EG_FnCode + EO_Message, 0L, 0L);

use(myco)
}

PRIVATE word do_search(conode, str)      /* function passed to list walking */
Conode *conode;
char   *str;
{ 
  if (!strcmp(conode->name, str))
    return((word) conode);
  else
    return(FALSE);
}

                   /* the data field of the MCB contains a string such as : */
                   /* /Cluster/Console. I need to extract the last bit, and */
                   /* compare it with known servers.                        */
PRIVATE void handle_search()
{ Conode *tempco;
  char *ptr1=mcb->Data, *ptr2=mcb->Data;

  for ( ; *ptr1 ne '\0'; ptr1++)
     if (*ptr1 eq '/') ptr2 = ptr1 + 1;

  tempco = (Conode *) Wander(WaitingCo, func(do_search), ptr2);
  if (!tempco)
    { if (debugflags & Search_Flag)
       ServerDebug("Distributed search for %s : not found", mcb->Data);
      Request_Return(EC_Error + SS_IOProc + EG_Unknown + EO_Server, 0L, 0L);
    }
  else
   { if (debugflags & Search_Flag)
       ServerDebug("Distributed search for %s : found, server is %lx",
                 mcb->Data, (WORD) tempco->id);
     mcb->MsgHdr.Dest = tempco->id;
     strcpy(mcb->Data, network_name);
     if (tempco->id ne 1L)
      { strcat(mcb->Data, "/");            /* Send back the full network name */
        strcat(mcb->Data, tempco->name);
      }
     Request_Return(ReplyOK, 0L, (WORD) (strlen(mcb->Data) + 1) );
   }
}

/*------------------------------------------------------------------------------
--
-- Other handlers for messages sent to the IOproc device
--
------------------------------------------------------------------------------*/

/**
*** This code is for both Locate and Create. If the message is for the IOproc
*** device I return the appropriate reply, otherwise I forward the message
*** to the appropriate server.
***
*** Note to Wolf and Chris : if you want to mount new devices dynamically,
*** they should be created inside IOproc. Remember to update the name list or
*** else you will not be able to access the new device. Exactly how you are
*** going to do it is upto you.
**/

void IOPROC_Locate(myco)
Conode *myco;
{ WORD temp;

  if (strcmp(IOname, myco->name) && IOname[0] ne '\0')
    Forward();
  else
   { IOname[0] = '\0';
     temp = FormOpenReply(Type_Directory, 0L, -1L, -1L);
     Request_Return(ReplyOK, open_reply, temp);
   }

 use(myco)
}

/**
*** Open requests for devices other than IOproc are forwarded, as usual, and
*** opening the IOproc device itself involves creating a new Stream coroutine
*** to handle Directory reads. Most of the IOPROC_handlers are the same as
*** the corresponding Dir_handlers. The extra field in the new stream's conode
*** structure is initialised to the directory list structure, so there is no
*** need for a separate InitStream.
**/

void IOPROC_Open(myco)
Conode *myco;
{ if (strcmp(IOname, myco->name) && IOname[0] ne '\0')
    Forward();
  else
   { if ( ((mcb->Control)[OpenMode_off] & 0x0F) eq O_WriteOnly)
      { Request_Return(EC_Error + SS_IOProc + EG_WrongFn + EO_Server, 0L, 0L);
        return;
      }

     NewStream(Type_Directory, Flags_Closeable, (WORD) &IO_extra,
               IOPROC_Handlers);
   } 
}

/**
*** ObjectInfo's intended for sub-servers are forwarded, and an ObjectInfo
*** on the IOproc device is easy to support. This means that e.g. the server
*** for drive d has to be able to handle ObjectInfo requests on drive d,
*** but it is more likely to succeed at that than the IOproc device.
**/

void IOPROC_ObjectInfo(myco)
Conode *myco;
{ ObjInfo *Heliosinfo = (ObjInfo *) mcb->Data;

  if (strcmp(IOname, myco->name) && IOname[0] ne '\0')
   { Forward();
     return;
   }

             /* The ObjectInfo information is in the Data vector so it does */
             /* not get swapped automatically by the message passing        */
             /* routines. All integers must be swapped explicitly. On the   */
             /* PC swap() is defined to be a no-op.                         */
  Heliosinfo->DirEntry.Type   = swap(Type_Directory);
  Heliosinfo->DirEntry.Flags  = swap(0L);
  Heliosinfo->DirEntry.Matrix = swap(DefDirMatrix);
  Heliosinfo->Size            = swap(0L);
  Heliosinfo->Account         = swap(0L);
  Heliosinfo->Creation        = swap(Startup_Time);
  Heliosinfo->Access          = swap(Startup_Time);
  Heliosinfo->Modified        = swap(Startup_Time);
  strcpy(Heliosinfo->DirEntry.Name, DefaultServerName);

  Request_Return(ReplyOK, 0L, (word) sizeof(ObjInfo));
  use(myco)
}

/**
*** IOPROC_Rename() is one of the most important operations on the IOPROC
*** device because that is how the network server renames my node in the
*** network. There are two stages. First, my local name i.e. IO may have
*** changed. Secondly my global network name may have changed.
***
**/

void IOPROC_Rename(myco)
Conode *myco;
{ char *newname = (char *) &(misc_buffer1[0]);
  char *data = mcb->Data;
  int  context = (int) mcb->Control[Context_off];
  int  dest    = (int) mcb->Control[RenameToname_off];

  if (strcmp(IOname, myco->name) && IOname[0] ne '\0')
   { Forward(); return; }

  if (context ne -1 && data[dest] ne '/')
   { char *temp;
     strcpy(newname, &(data[context]));
     for (temp = newname; *temp ne '\0'; temp++);
     if (*(--temp) ne '/')
       *(++temp) = '/';
     *(++temp)   = '\0';  
     strcat(newname, &(data[dest]));
   }
  else
   strcpy(newname, &(data[dest]));

  unless(flatten(newname))
    { Request_Return(EC_Error + SS_IOProc + EG_Name + EO_Server, 0L, 0L);
      return;
    }

  strcpy(network_name, newname);

  data = newname + strlen(newname) - 1;
 
  while (*data ne '/') data--;

  strcpy(myco->name, ++data);

  Request_Return(ReplyOK, 0L, 0L);

use(myco)
}

/**
*** I cannot use Dir_Close() here because that frees all the space taken up
*** by the linked list used to hold the directories.
**/

void IOPROC_Close(myco)
Conode *myco;
{ Request_Return(ReplyOK, 0L, 0L);
  Seppuku();
  use(myco)
}

/*----------------------------------------------------------------------
--
-- Some general purpose handlers
--
----------------------------------------------------------------------*/

/**
*** The following routines are used as default request handlers with
*** particular servers, and are declared in header file fundefs.h as
*** the handler for that request for the server in question. For example,
*** the Rename request is illegal for the console server so the handler
*** Console_Rename, which is stored in the array of handlers for the
*** Console device in the Console server coroutine node, is hash-defined
*** to InvalidFn.
**/ 
                /* for requests that are not defined on a particular object */
void Invalidfn_handler(myco)
Conode *myco;
{ Request_Return(EC_Error + SS_IOProc + EG_WrongFn + EO_Server, 0L, 0L);
  use(myco)
}

                      /* for requests to devices that can be ignored */
void Dummy_handler(myco)
Conode *myco;
{ Request_Return(ReplyOK, 0L, 0L);
  use(myco)
}

WORD Ignore()                /* dummy initialisation routine */
{ return(0L);
}

void IgnoreVoid()
{ return;
}

                               /* for create and locate requests on devices */
void Create_handler(myco)
Conode *myco;
{ WORD temp;

  temp = FormOpenReply(Type_File, 0L, -1L, -1L);
  Request_Return(ReplyOK, open_reply, temp);
  use(myco)
}

              /* ObjectInfo is straightforward for all devices except Clock */
              /* so the devices can share this ObjectInfo handler.          */
void Device_ObjectInfo_handler(myco)
Conode *myco;
{                       /* put info in data vector straightaway, nothing in */
                        /* there that we need any more */
  ObjInfo *Heliosinfo = (ObjInfo *) mcb->Data;

             /* The ObjectInfo information is in the Data vector so it does */
             /* not get swapped automatically by the message passing        */
             /* routines. All integers must be swapped explicitly. On the   */
             /* PC swap() is defined to be a no-op.                         */
  Heliosinfo->DirEntry.Type   = swap(Type_File);
  Heliosinfo->DirEntry.Flags  = swap(0L);
  Heliosinfo->DirEntry.Matrix = swap(DefFileMatrix);
  Heliosinfo->Size            = swap(0L);
  Heliosinfo->Account         = swap(0L);
  Heliosinfo->Creation        = swap(Startup_Time);
  Heliosinfo->Access          = swap(Startup_Time);
  Heliosinfo->Modified        = swap(Startup_Time);
  strcpy(Heliosinfo->DirEntry.Name, IOname);

  Request_Return(ReplyOK, 0L, (word) sizeof(ObjInfo));
  use(myco)
}

void GetDefaultAttr(myco)
Conode *myco;
{ InitAttributes((Attributes *) mcb->Data);
  Request_Return(ReplyOK, 0L, (word) sizeof(Attributes));
  use(myco)
}

void Device_GetSize(myco)
Conode *myco;
{ (mcb->Control)[Reply1_off] = MAXINT;
  Request_Return(ReplyOK, 1L, 0L);
  use(myco)
}

void NullFn(myco)
Conode *myco;
{ Request_Return(ReplyOK, 0L, 0L);
  use(myco)
}

/*----------------------------------------------------------------------------
--
--  Name conversion
--
----------------------------------------------------------------------------*/

/**
*** Name conversion is non-trivial. The technical manual describes how the
*** Helios name is packed inside my message structure, and the code below
*** extracts the local name, i.e. the bit relevant to me, from all the junk.
*** It starts by extracting the necessary bits from the control vector and
*** giving a debugging message.
***
*** The first bit of code goes through the names, starting at pointer rest which
*** may be in either the context or the name fields, and puts all the data into
*** the array IOname. I try to update the next field in the message in case the
*** message has to be passed from one server to another, i.e. if another name
*** conversion may be required. This bit of code is rather dubious. It may be
*** necessary to extract bits of the name from the name field as well, in case
*** the rest pointer was still somewhere inside context. Having got hold of
*** the whole name I put in a terminator.
***
*** Sadly IOname may still contain elements . and .. which have to be filtered
*** out. Hence I must go through my entire local name again looking for these
*** special cases, and this is done by routine flatten() which is also called
*** by Rename handlers. Only now am I finished, and I can put in another
*** terminator just in case and produce another debugging option. It is
*** possible for the name conversion to fail if there is an attempt to
*** backtrack to a server outside the Server, as I refuse to forward messages
*** to other bits of the network.
**/

word convert_name()
{ byte *data     = mcb->Data;
  word *control  = mcb->Control;         /* the control vector contains the */
                                         /* offsets needed for the conversion */
  int  context   = (int) control[Context_off];
  int  name      = (int) control[Pathname_off];
  int  next      = (int) control[Nextname_off];
  char *dest     = IOname;

  if ((context eq -1) && (name eq -1))
   { if (debugflags & Name_Flag)
      ServerDebug("System library error : NULL context and NULL name");
     return(FALSE);
   }

  if (debugflags & Name_Flag)
   ServerDebug("Context %s, name %s, rest %s", 
             (context ne -1) ? &(data[context]) : "NULL", 
             (name ne -1) ? &(data[name]) : "NULL",
             (next ne -1) ? &(data[next]) : "NULL" );

  for ( ; data[next] ne '/' && data[next] ne '\0'; next++)
    *dest++ = data[next];

  if (data[next] eq '/')
    { control[Nextname_off] = next+1;
      for ( ; data[next] ne '\0'; next++)
        *dest++ = data[next];
    }
  elif (name ne -1)
    if ( ((next < name) && (context < name)) ||
         ((next > name) && (context > name)) )
      control[Nextname_off] = name;

  if (name eq -1) goto finished;

  if ( ((next < name) && (context < name)) ||
       ((next > name) && (context > name)) )
   { *dest++ = '/';
     for ( ; data[name] ne '\0'; name++)
       *dest++ = data[name];
   }

finished:

  *dest = '\0';

  if (!flatten(IOname))
   { if (debugflags & Name_Flag)
       ServerDebug("Name conversion failed.");
     return(FALSE);
   }

  if (debugflags & Name_Flag)
    ServerDebug("IOname is %s", IOname);

  return(TRUE);
}

WORD flatten(name)
char *name;
{ char *source = name, *dest = name;
  int  entries = 0;

  while(*source ne '\0')
   { if (*source eq '.')
       { source++;
         if   (*source eq '/') { source++; continue; }
         elif (*source eq '\0')
           { if (entries < 1) return(FALSE);
             dest--; break;
           }
         elif (*source eq '.')
           { source++;
             if (*source eq '/' || *source eq '\0')
               { 
                 if (entries <= 1) return(FALSE);
                 dest--; dest--; while (*dest ne '/')  dest--;
                 if (*source ne '\0') 
                  { dest++; source++; }
                 entries--; continue;
               }
             else
               { *dest++ = '.'; *dest++ = '.'; }
           }
         else *dest++ = '.';
       }

     while (*source ne '/' && *source ne '\0') *dest++ = *source++;
     if (*source ne '\0') { *dest++ = '/'; source++; entries++; }
   }

  *dest = '\0';

  return(TRUE);
}

/*------------------------------------------------------------------------
--
-- Here are some other odds and ends needed by the server
--
------------------------------------------------------------------------*/

/**
*** The following routines provide miscellaneous utilities.
***
*** FormOpenReply() is sets up the reply to Open, Create, and Locate requests,
*** which is a non-trivial job.
***
*** mystrcmp() does a string comparison which ignores differences in case.
*** The server is now case-sensitive, so it is not used as often, but the
*** configuration info in host.con is still case-independent.
*** It returns 0 if the strings are the same, non-zero otherwise, as does
*** C strcmp().
***
*** The swap routine is used for any machine with a different byte-ordering
*** from the transputer - if the byte ordering is the same swap is #define'd
*** as a no-op. The current swap routine is suitable for the 68000.
***
*** Clear_bytes() clears an area of memory, as I do not trust memclr on all
*** the hardware I use.
***
*** NewStream() is used by the device servers to create new streams.
**/



/*----------------------------------------------------------------------------
--
-- FormOpenReply()
--
----------------------------------------------------------------------------*/

/**
*** This utility routine sets up a reply to Open, Locate and Create requests.
*** All the information comes as arguments, or is available in network_name
*** The routine returns the size of the data vector.
***
*** The capability fields passed to FormOpenReply() are usually 0, since
*** I cannot support the protection system anyway. I have no idea what
*** the last field is for, but NHG has not yet complained about permanent
*** zero so I hope it is satisfactory.
***
*** See header file iogsp.h for details of the returned data structure.
**/

word FormOpenReply(type, flags, cap1, cap2)
word type, flags, cap1, cap2;
{ word *control = mcb->Control;
  char *data    = mcb->Data;

  control[Reply1_off] = type;
  control[Reply2_off] = flags;
  control[Reply3_off] = cap1;
  control[Reply4_off] = cap2;
  control[Reply5_off] = 0;     /* offset for pathname */
  control[Reply6_off] = 0;     /* object value if no reply port ? */

  strcpy(data, network_name);
  if (IOname[0] ne '\0')
   { strcat(data, "/"); strcat(data, IOname); }

  return((word) (strlen(data) + 1) );
}


word mystrcmp(s1, s2)
char *s1, *s2;
{ for (;;)
   { if (*s1 eq '\0')
       return((*s2 eq '\0') ? 0L : -1L);
     elif (*s2 eq '\0')
       return(1L);
     elif(ToUpper(*s1) < ToUpper(*s2))
       return(-1L);
     elif(ToUpper(*s1) > ToUpper(*s2))
       return(1L);
     else
       { s1++; s2++; }
   }
}

#if swapping_needed
#if (ST || TRIPOS || SUN)
                  /* for the 68000's byte ordering */
                  /* on the Amiga it is written in assembler */
word swap(a)
word a;
{ word b = 0;
  int i;

  if (a eq 0L) return(0L);

  for (i=0; i<4; i++)
    { b <<= 8; b |= (a & 0xFF); a >>= 8; }

  return(b);
}
#endif     /* 68000 */
#endif

#if (!SUN)
void clear_bytes(addr, no)
char *addr;
int no;
{ int i;

  for (i=0; i < no; i++)
    addr[i] = '\0';
}
#endif

/**
*** This is called by device servers e.g. Console or Mouse to open a new stream,
*** which involves creating a stream coroutine. The routine is given the message
*** containing the open request, which must be filled in correctly so that it
*** can be returned by General_Stream(). The necessary information is provided
*** as arguments. Creating the coroutine is straightforward and independent of
*** of the device, apart from the handlers for Read etc. requests which are
*** provided as an argument.
***
*** For certain devices the conode extra field must be filled in, so this is
*** another argument.
***
*** The new stream needs a new message port / unique coroutine id, which must
*** be sent back to the transputer. This is done by filling in the dest field
*** of the message header. Request_Return() is responsible for swapping the
*** source and destination ports, so it has to be in dest.
**/

void NewStream(type, flags, extra, handlers)
WORD extra;
WORD type, flags;
VoidFnPtr *handlers;
{ WORD   temp;
  Conode *newco;

  temp = FormOpenReply(type, flags, -1L, -1L);

  newco = NewCo(General_Stream);
  unless(newco)
    { Request_Return(EC_Warn + SS_IOProc + EG_NoMemory + EO_Server, 0L, 0L);
      return;
    }

  PostInsert(&(newco->node), Heliosnode);
  newco->id        = CoCount++;
  newco->timelimit = Now + StreamTimeout;
  newco->name[0]   = 0x7f;
  strcpy(&(newco->name[1]), IOname);
  newco->handlers  = handlers;  
  newco->extra     = (ptr) extra;
  mcb->MsgHdr.Dest = newco->id;

  StartCo(newco);
}

/*------------------------------------------------------------------------
--
-- Routines to manipulate attributes
--
------------------------------------------------------------------------*/

/**
*** The various routines below are used to examine and change device
*** stream attributes, and are the same as in syslib.c apart from problems
*** with 16/32 bit integers.
**/
                /* In an Attribute, the bottom two bits indicate which word */
                /* in the Attributes structure we are interested in, and    */
                /* the rest of the word specifies a single bit.             */
WORD IsAnAttribute(attr, item)
Attributes *attr;
Attribute item;
{ WORD *a_ptr = (WORD *) attr + (item & 0x00000003L);
  item     &= 0xFFFFFFFCL;
  return( *a_ptr & item);
}

void AddAttribute(attr, item)
Attributes *attr;
Attribute item;
{ WORD *a_ptr = (WORD *) attr + (item & 0x00000003L);
  item     &= 0xFFFFFFFCL;
  *a_ptr     |= item;
}

void RemoveAttribute(attr, item)
Attributes *attr;
Attribute item;
{ WORD *a_ptr = (WORD *) attr + (item & 0x00000003L);
  item     &= 0xFFFFFFFCL;
  *a_ptr     &= ~item;
}


                     /* The input and output speeds of a stream are kept in */
                     /* the bottom byte of the relevant Attributes field.   */
                     /* thus allowing upto 256 different speeds.            */
WORD GetInputSpeed(attr)
Attributes *attr;
{ return((attr->Input) & 0x000000FFL);
}

WORD GetOutputSpeed(attr)
Attributes *attr;
{ return((attr->Output) & 0x000000FFL);
}

void SetInputSpeed(attr, speed)
Attributes *attr;
WORD speed;
{ 
  attr->Input  &= 0xFFFFFF00L;  /* clear bottom byte */
  speed        &= 0x000000FFL;  /* prevent nasty crash if speed invalid */
  attr->Input  |= speed;
}

void SetOutputSpeed(attr, speed)
Attributes *attr;
WORD speed;
{ 
  attr->Output &= 0xFFFFFF00L;  /* clear bottom byte */
  speed        &= 0x000000FFL;  /* prevent nasty crash if speed invalid */
  attr->Output |= speed;
}

void InitAttributes(pointer)
Attributes *pointer;
{ int i;
  WORD *wpointer = (WORD *) pointer;
  for (i=0; i < 5; i++)
    *wpointer++ = 0L;
}

void CopyAttributes(to_ptr, from_ptr)
Attributes *to_ptr, *from_ptr;
{ int i;
  WORD *wto_ptr, *wfrom_ptr;
  wto_ptr = (WORD *) to_ptr; wfrom_ptr = (WORD *) from_ptr;
  
  for (i=0; i < 5; i++)
    *wto_ptr++ = swap(*wfrom_ptr++);
}



