/*------------------------------------------------------------------------
--                                                                      --
--          H E L I O S   I N P U T / O U T P U T   S E R V E R         --
--          ---------------------------------------------------         --
--                                                                      --
--             Copyright (C) 1988, Perihelion Software Ltd.             --
--                        All Rights Reserved.                          --
--                                                                      --
--      structs.h                                                       --
--                                                                      --
--         Declarations of the server's manifests and data structures   --
--                                                                      --
--      Author:  BLV 9/6/87                                             --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89       Copyright (C) 1987, Perihelion Software Ltd.   */


/**
***   The Server makes extensive use of arrays of functions to handle all
***  the Helios requests in a general way, and the following manifests
***  declare offsets within the arrays for these requests.
**/
#define InitServer_off      0
#define TidyServer_off      1
#define Private_off         2
#define Testfun_off         3
#define Open_off            4
#define Create_off          5 
#define Locate_off          6
#define ObjectInfo_off      7
#define ServerInfo_off      8
#define Delete_off          9
#define Rename_off         10
#define Link_off           11
#define Protect_off        12
#define SetDate_off        13
#define Refine_off         14
#define CloseObj_off       15
#define handler_max        16

#define InitStream_off      0
#define TidyStream_off      1
#define StreamPrivate_off   2
#define Read_off            3
#define Write_off           4
#define GetSize_off         5
#define SetSize_off         6
#define Close_off           7
#define Seek_off            8
#define GetAttr_off         9
#define SetAttr_off        10
#define EnableEvents_off   11
#define Acknowledge_off    12
#define NegAcknowledge_off 13
#define Stream_max         14

/**
*** This is returned by get_int_config() in module server.c to indicate failure
**/
#define Invalid_config     0x64928AF2L

/**
*** The main data structure in the Server is the linked list.
**/
typedef struct Node { struct Node *next;       /* next node in list */
                      struct Node *prev;       /* previous node in list */
} Node;

typedef struct List { Node *head;       /* first node in list  */
                      Node *earth;      /* always NULL pointer */
                      Node *tail;       /* last node in list   */
} List;

typedef struct GenData {
                 Node    node;
                 uint    size;
                 int     junk;     /* To guarantee word alignment of data */
                 BYTE    data[1];
} GenData;

/**														
*** The Server depends heavily on coroutines. All the coroutines are held in
*** linked lists, using the following structure.
**/
typedef struct { Node node;           /* conodes held in linked list  */
                 word id;             /* unique integer identifier    */
                 word type;           /* usually zero, may be suicide */
                                      /* or timedout                  */
                 word flags;
                 word timelimit;      /* when stream should die       */
                 ptr  cobase;         /* from CreateCo                */
                 void (**handlers)(); /* request handlers             */
                 ptr  extra;          /* coroutines global data       */
                 char name[128];      /* a name field if required     */
} Conode;

/**
*** Various odds and ends for use with coroutines
**/
#define CoSuicide (654321L)
#define CoTimeout (666666L)
#define CoReady   (777777L)

#define CoFlag_Floppy       0x0001L
#define CoFlag_CtrlCed      0x0002L
#define CoFlag_FileStream   0x0004L

/**
*** This structure is used to declare my devices. The type is File, Directory
*** or Private, the name gives the device name e.g. Console, and the handlers
*** field gives the Helios request handlers. I.E. there is an entry in this
*** array to deal with Open requests, another for Locate requests, etc.
**/

typedef struct device_declaration {
       WORD type;
       char *name;
       VoidFnPtr handlers[handler_max];
} device_declaration;

/**
*** The Server contains various directories. In addition to disk directories,
*** the IO processor itself is a directory of servers, and some of these servers
*** such as the communication port servers are directories of ports. Hence I
*** share code to deal with these various directories, possible by keeping the
*** directory in a linked list of DirEntryNode's.
**/
   
typedef struct { Node       node;
                 DirEntry   direntry;
} DirEntryNode;

typedef struct { List list;
                 WORD entries;
} DirHeader;

/**
*** Communication Port_node's are slightly more complicated because I need
*** additional information. However, the first fields in the structure are
*** compatible with DirEntryNode, so that I can use the usual directory handling
*** routines.
**/

typedef struct { Node       node;
                 DirEntry   direntry;
                 WORD       id;
                 VoidFnPtr  error_fn;
                 VoidFnPtr  done_fn;
                 VoidFnPtr  configure_fn;
                 WordFnPtr  send_fn;
                 WordFnPtr  pollwrite_fn;
                 WordFnPtr  abortwrite_fn;
                 WordFnPtr  receive_fn;
                 WordFnPtr  pollread_fn;
                 WordFnPtr  abortread_fn;
                 VoidFnPtr  *handlers;
                 Attributes *attr;
} Port_node;

/**
*** Initially the server boots up a simple network consisting of the root
*** processor and the IO processor. These must be named during the booting
*** processes, so the following manifests define the default names.
**/
#define DefaultServerName "IO"
#define DefaultRootName   "00"
#define slashDefaultServerName "/IO"
#define slashDefaultRootName   "/00"

/**
*** The following structures are used for handling windows and the console
*** device. Microwave is used to handle cooked input processing (pun
*** definitely intended), Screen is used by the ANSI emulator, and Window is
*** used to point at a window structure.
**/
#define       Console_limit 256

typedef struct Microwave {
        UBYTE  buffer[Console_limit];  /* where data is processed    */
        int    count;                  /* current position in buffer */
        int    outstanding;
        WORD   EOFco;             /* Coroutine id for EOF */
        Conode *readerco;         /* coroutine currently reading from window */
} Microwave;

#define Cooked_EOF  1               /* ctrl-D detected         */
#define Cooked_EOD  2               /* no more data in buffer  */   
#define Cooked_Done 3               /* Read has been satisfied */

#if use_ANSI_emulator
typedef struct Screen { BYTE          **map;
                        int           Rows;
                        int           Cols;
                        int           Cur_x;
                        int           Cur_y;
                        int           mode;
                        int           in_escape;
                        int           escape_found;
                        int           args[5];
                        int           *args_ptr;
                        int           gotdigits;
                        int           firstdigit; 
} Screen;
#endif

typedef struct Window { DirEntryNode  node;
                        Attributes    attr;
                        event_handler break_handler;
                        Microwave     cooker;
#if use_ANSI_emulator
                        Screen        screen;
#endif
                        UBYTE         Table[Console_limit];
                        WORD          handle;
                        int           head, tail;
                        int           XOFF;
                        int           stream_count;
} Window;

/**
*** Multi-tasking support. To prevent the server from hogging all the
*** CPU time, the main sources inform the local module whenever it is
*** waiting for particular input, e.g. for a key to be pressed or for
*** the mouse to be moved. Then at a strategic point in the server's main
*** loop I call a multiwait function. This can suspend the CPU for upto
*** half a second or until one of the devices specified is ready.
*** The following constants define the various forms of IO which the
*** server can be waiting on.
**/
#if multi_tasking
#define Multi_LinkMessage          1L
#define Multi_ConsoleInput         2L
#define Multi_FileRead             3L
#define Multi_GemInput             4L
#define Multi_MouseInput           5L
#define Multi_KeyboardInput        6L
#define Multi_RS232Event           7L
#define Multi_PortIO               8L
#endif

