/*------------------------------------------------------------------------
--                                                                      --
--          H E L I O S   I N P U T / O U T P U T   S E R V E R         --
--          ---------------------------------------------------         --
--                                                                      --
--             Copyright (C) 1987, Perihelion Software Ltd.             --
--                        All Rights Reserved.                          --
--                                                                      --
--      tload.c                                                         --
--                                                                      --
--              This module is responsible for bootstrapping the        --
--                                                                      --
--      root transputer on start-up, and interacting with the           --
--                                                                      --
--      transputer network.                                             --
--                                                                      --
--  Author:  BLV 8/10/87                                                --
--                                                                      --
------------------------------------------------------------------------*/
/* SccsId: 3.7 14/4/89  Copyright (C) 1987, Perihelion Software Ltd. */

#define Tload_Module

#include "helios.h"

/**
*** The bits below are pretty horrible. The code for booting up the transputer
*** used to be a separate program tload, but it is necessary for the server
*** and the debugger to be able to do this themselves. Hence the code from
*** tload.c was inserted here and fiddled about to make it work. Things become
*** very nasty because the Server and the debugger have rather different
*** requirements.
***
*** The following declarations are all rather boring.
**/

#define MAXREAD  0xffffL
#define MAXCHUNK 0x1000L

PRIVATE List  system_list;
List  *Image_List;
WORD  isize;

FILE  *infd;

PRIVATE void fn( timeout,       (char *));
PRIVATE void fn( send_boot,     (void));
PRIVATE void fn( send_conf,     (void));
PRIVATE void fn( handle_byteF0, (void));
PRIVATE void fn( server_helios, (void));

/**
*** When the Server or the debugger wants to boot the transputer it calls
*** this routine in either mode debugboot or serverboot. The routine loads
*** the bootstrap code and the system image, and sends them off to the
*** transputer. Then the space take up by the bootstrap code is freed, and
*** if the system is in Server mode the space taken up by the system image
*** is freed as well because this lets me have a couple more coroutines before
*** running out of memory, and that is rather important on a 640K PC.
**/

static jmp_buf failure_buf;

void boot_processor(mode)
int mode;
{ if (setjmp(failure_buf) ne 0) goto failed;

  infd = (FILE *)NULL;
  if (!loadimage()) goto failed;

  send_boot();

  if ((mode ne debugboot) && TstList(Image_List))
    { FreeList(Image_List);
      server_helios();
    }

  return;

failed:                      /* to bootstrap the transputer for some reason */
  if (mode eq debugboot)
   return;                   /* continue debugging */
  else
     longjmp(exit_jmpbuf, 1);  /* exit Server */
}

/**
*** The following bits of code deal with loading items off disk. The file name
*** corresponding to the system image is known because the system cannot
*** start up without it - either it is in the configuration file or it is on
*** the command line. Also, the system image may be loaded already if in
*** debugging mode so I check for that. The system image is loaded in chunks
*** because some systems object to loading vast quantities of data in one go.
***
*** loadboot() is called once only, from main() in module server.c . A bootfile
*** may be specified in the configuration file, with the default being nboot.i
*** The space for the bootstrap code is freed at the end of main().
**/
 
int loadimage()
{ WORD l;
  int  cur;
  GenData *cur_data;

  if (TstList(Image_List)) return(1);               /* already loaded ? */

  if (debugflags & Boot_Flag)
    ServerDebug("Loading system image %s.", system_image);

#if (SUN)
  infd = fopen(system_image,"r");
#else
  infd = fopen(system_image,"rb");
#endif
  if( infd eq (FILE *) NULL ) 
   { ServerDebug("Cannot open image %s for input.",system_image);
     return(0);
   }

  if (fread((BYTE *) &isize,1,4,infd) ne 4)
   { ServerDebug("Cannot read image header.");
     goto done;
   }

  isize = swap(isize);                   /* extract size of system image */

  if (fseek(infd, 0L, SEEK_SET) ne 0)    /* and reset file to beginning */
   { ServerDebug("Failed to read image.");
     goto done;
   }

  for (l = 0L; l < isize; l += MAXCHUNK)
   { cur = ((isize - l) > MAXCHUNK) ? (int) MAXCHUNK : (int) (isize - l);
     cur_data = (GenData *) malloc(sizeof(GenData) + cur);
     if (cur_data eq (GenData *) NULL)
      { ServerDebug("Insufficient memory to store system image.");
        goto done;
      }
     if (fread(&(cur_data->data[0]), 1, cur, infd) eq 0)
      { ServerDebug("Failed to read system image.");
        goto done;
      }
     cur_data->size = cur;
     AddTail(&(cur_data->node), Image_List);
   }

  fclose( infd );                           /* finished successfully */
  infd = (FILE *) NULL;
  return(1);

done:
  fclose( infd );
  infd = (FILE *) NULL;
  FreeList(Image_List);
  return(0);  /* return error */
}


void init_boot()
{ int s;
  WORD ihdr[3];
  char *name = get_config("bootfile");    /* get boot name from configuration */

  InitList(&system_list);
  Image_List = &system_list;

  if (name eq (char *) NULL)              /* use default bootfile */     
    name = "nboot.i";

  if (debugflags & Boot_Flag)
   ServerDebug("Loading bootstrap program %s", name);

#if (SUN)
  infd = fopen(name,"r");
#else
  infd = fopen(name,"rb");
#endif
  if( infd eq (FILE *) NULL ) 
   { ServerDebug("Cannot open boot program %s for input.",name);
     goto fail;
   }

  if( (s=fread((BYTE *) &(ihdr[0]),1,12,infd)) ne 12)
   { ServerDebug("Cannot read boot header %d %d.",s,ferror(infd));
     goto fail;
   }

  if( swap(ihdr[0]) ne 0x12345678L )
   { ServerDebug("First word of file not magic number.");
     goto fail;
   }

  bootsize = swap(ihdr[2]);

  bootstrap = (UBYTE *) malloc((int) bootsize);
  if( bootstrap eq NULL )
   { ServerDebug("Cannot get image buffer."); goto fail; }

  if( (s=fread(bootstrap,1,(int)bootsize,infd)) ne (int) bootsize)
   { ServerDebug("Image too small %d, error %d.",s,ferror(infd)); goto fail; }

  fclose( infd );
  infd = (FILE *) NULL; 
  return;

fail:
  if (infd ne (FILE *) NULL)
   { fclose( infd );
     infd = (FILE *) NULL;
   }
 longjmp(exit_jmpbuf, 1);  /* exit Server */
}

void tidy_boot()
{ if (bootstrap ne NULL)
   { free(bootstrap); bootstrap = NULL; }
  if (TstList(Image_List))
   FreeList(Image_List);
}

/**
*** Timeout() is called when any of the boot writes to the transputer fails.
*** It longjmps back to boot_transputer() above, which either recovers or
*** exits depending on what the user is doing.
**/

PRIVATE void timeout(when)
char *when;
{ ServerDebug("Timed out when sending %s.", when);
  longjmp(failure_buf, 1);
}

/**
*** Send_boot() is responsible for sending all the bootstrap code to the
*** transputer. First I attempt to reset all the transputer hardware, which is
*** a very troublesome piece of code. Then I send the size of the bootstrap
*** program, followed by the bootstrap program itself. This program starts to
*** run on the transputer and on Meiko hardware the first thing it does is
*** clear all of the parity memory - Yukk. Then the program does an IN
*** instruction to receive the system image, and I send that. When the system
*** image has been loaded the kernel starts to run and expects some
*** configuration to come down its link, so I send that.
***
*** Next there is some low-level message interaction between the transputer and
*** the host, involving a byte-F0 message, which only NHG understands. In
*** server mode this involves calling server_helios(), and in debugging mode
*** it involves calling debug_helios() in module debug.c .
**/

PRIVATE WORD send_bit(data)
GenData *data;
{ poll_the_devices();
  if (Special_Exit) longjmp(exit_jmpbuf, 1);
  if (debugflags & Boot_Flag)
    ServerDebug("Sending %d bytes at %p.", data->size, &(data->data[0]));

  if (!xpwrdata(&(data->data[0]), (WORD) data->size))
   return(1L);
  else
   return(0L);
}

PRIVATE void send_boot()
{ char *temp;

  if (debugflags & Boot_Flag)
   ServerDebug("Resetting link and transputer.");
  resetlnk();
  xpreset();
   

  if (debugflags & Boot_Flag)
   ServerDebug("Sending bootstrap, size is %ld", bootsize);

  if (!xpwrbyte(bootsize))              /* bootstrap size */
   timeout("bootstrap size");
  if (!xpwrdata(bootstrap,bootsize))    /* bootstrap */
   timeout("bootstrap code");

  temp = get_config("box");                   /* This entry is known to exist */
  if (!mystrcmp(temp, "MK026"))        /* or anything else with parity memory */
    { WORD size = get_int_config("transputer_memory");
      if (size eq Invalid_config) size = 0x300000L;
      xpwrbyte(5L);
      xpwrint(size);
      while (!xprdrdy());           /* wait for the clear to finish */
      (void) xprdbyte();
    }

  if (debugflags & Boot_Flag)
   ServerDebug("Sending System Image (%ld bytes) ...",isize);

  xpwrbyte(4L);

  if (Wander(Image_List, send_bit))
   timeout("system image");

  send_conf();
}

/**
*** The configuration vector
**/

#if !(PC)
PRIVATE Config config;
#else
PRIVATE Config far config;
#endif

PRIVATE WORD build_config()
{ char *procname = slashDefaultRootName, *myname = slashDefaultServerName;
  char *tempname = get_config("root_processor");
  WORD memory_size = get_int_config("transputer_memory");

  if (tempname ne (char *) NULL) procname = tempname;
  tempname = get_config("io_processor");
  if (tempname ne (char *) NULL) myname = tempname;
  if (memory_size eq Invalid_config) memory_size = 0L;

  config.PortTabSize = swap(1024L);
  config.Incarnation = swap(1L);
  config.loadbase    = swap(0x80001000L);
  config.ImageSize   = swap(isize);
  config.Date        = swap(get_unix_time());
  config.FirstProg   = swap(6L);
  config.Memory      = swap(memory_size);
  config.Flags       = swap(Config_flags_rootnode);
  config.Spare       = swap(0L);
  config.NLinks      = swap(4L);
  config.LinkConf[0] = swap(0x00060200L);
  config.LinkConf[1] = swap(0x01060200L);
  config.LinkConf[2] = swap(0x02060200L);
  config.LinkConf[3] = swap(0x03060200L);
  config.MyName      = swap(((WORD) &(config.names[0])) -
                            ((WORD) &(config.MyName  )) );
  config.ParentName  = swap(((WORD) &(config.names[strlen(procname)+1])) -
                            ((WORD) &(config.ParentName)) );
  strcpy(config.names, procname);
  strcpy(&(config.names[strlen(procname) + 1]), myname);
  return((WORD) (sizeof(Config) + strlen(procname) + strlen(myname) - 48));
}

PRIVATE void send_conf()
{ WORD size = build_config();

  if (debugflags & Boot_Flag)
     ServerDebug("Sending configuration ...");

  if (!xpwrint(size))
   timeout("system configuration size");
  if (!xpwrdata((BYTE *) &config, size))
    timeout("system configuration");
}

/**
*** The main job of this routine is to wait for the kernel to start up
*** sufficiently for it to need to know a port on my side of the link. This
*** involves the transputer sending a byte F0, followed by some more junk,
*** and I reply with junk plus a pseudo port for the IOproc device.
**/

PRIVATE void server_helios()
{ for(;;)
   {
     poll_the_devices();
     if (Special_Exit) longjmp(exit_jmpbuf, 1);
     if (Special_Status)                /* or for status request */
      { ServerDebug("Server alive and well.");
        Special_Status = FALSE;
      }
     if (Special_Reboot || DebugMode)            /* or for reboot condition */
        return;

     if( xprdrdy() )
      { WORD b=0L;

        b = xprdbyte();

        switch( (int) b )
         { case 0xf0:
                      handle_byteF0();
                      output("Booted...\r\n");
                      return; /* the transputer is now ready */
           default:
                      ServerDebug("Unexpected byte %lx.",b);
                      break;
         }
      }
   }
}

PRIVATE void handle_byteF0()
{ WORD a, iocport;

  if (debugflags & Boot_Flag)
   ServerDebug("Received a byte F0.");
  xprdbyte();         /* three bytes of junk to fill up a long word */
  xprdbyte(); 
  xprdbyte(); 

  a=xprdint();
  iocport=xprdint();

  if ( ( a & 0x00ff0000L ) ne 0L )
   { xpwrint (0xf0f0f0f0L);
     xpwrint (0x00000100L);
     xpwrint (1L);     /* The IOprocessor is always at port 1 */
   }
}

/*------------------------------------------------------------------------
-- The following routines provide an interface to message passing       --
-- with the Transputer.                                                 --
------------------------------------------------------------------------*/

/**
*** It is desirable that all message interaction between transputer and host
*** go via a small set of routines, and for historical reasons these are the
*** ones. Messages are always received by the root coroutine inside the main
*** loop, by calls to Request_Stat() and Request_Get(). The messages are
*** buffered in mcbstruct, which is set up by init_main_message() and the
*** space obtained is freed again by free_main_message(). It is important
*** that coroutines do not do their own GetMsg's, because the local GetMsg
*** does not know anything about ports etc. and I do not want to complicate it.
*** 
*** Request_Stat() is the routine that gets hold of any messages the transputer
*** is trying to send. Some of these messages may be very low-level debugging
*** ones, identified by a FnRc of 0x22222222L, and these are treated separately.
*** I have to worry a little bit to make sure that all debugging messages
*** appear on separate lines, assuming they are not too interleaved. If another
*** message is received the routine returns TRUE, and main_loop calls
*** Request_Get() to get hold of it. In fact the value returned is always the
*** same, the address of the main message, but this call allows me to clear
*** the flags field which is desirable according to NHG.
***
*** When a reply is to be sent to the transputer this goes via Request_Return(),
*** which plays about a bit with the message to make sure that each coroutine
*** does not have to worry about swapping ports around etc. There is also a
*** user debugging facility, and a test for the special case of sending a
*** message to the invalid port zero which is strongly indicative of a bug
*** in the server. If for some reason Request_Return() is unable to send its
*** message this is indicative of the transputer having crashed, so I output
*** a suitable message.
***
*** GetMsg() can return three values. It can return TRUE if the machine has
*** received a complete Helios message, FALSE if it has not received anything
*** or 0xF0 if it has received the first byte of a raw message. GetMsg() itself
*** is responsible for handling all peek and poke requests coming from the
*** transputer.
**/

PRIVATE MCB  mcbstruct;
PRIVATE word control[256];
PRIVATE BYTE *real_buff;

void init_main_message()
{ uint amount = (uint) (maxdata + 1000L);
  BYTE *temp;

  temp = (BYTE *) malloc(amount);
  if (temp eq NULL)
    { ServerDebug("Insufficient memory on the server machine for message buffer.");
      longjmp(exit_jmpbuf, 1);
    }
  real_buff          = temp;
  mcbstruct.Data     = (BYTE *) (( ((WORD)temp) + 511L) & 0xFFFFFE00L);
  mcbstruct.Control  = &(control[0]);
  mcb = &mcbstruct;
}

void free_main_message()
{ free(real_buff);
}

int newdebug = TRUE;

word Request_Stat()
{ word j = 0L;

  mcbstruct.Timeout = 0xFFL;
#if AMIGA
  j = GetHeliosMsg(&mcbstructs);
#else  
  j = GetMsg(&mcbstruct);
#endif

  if (j eq 0x00F0L)
    handle_byteF0();
  elif (j) 
     { if (mcbstruct.MsgHdr.FnRc eq 0x22222222L)
         { int x;
           for (x = 0; x < mcbstruct.MsgHdr.DataSize; x++)
            { if (newdebug)
               { ServerDebug("+++ %q");
                 newdebug = FALSE;
               }
              if (mcbstruct.Data[x] eq '\n')
               { ServerDebug("");
                 newdebug = TRUE;
                 continue;
               }
              ServerDebug("%c%q", mcbstruct.Data[x]);
            }
           return(FALSE);
         }
       else
        { mcbstruct.MsgHdr.Flags = 0;        /* on NickG's orders */
          return(TRUE);
        }
     }
  else
   return(FALSE);
}

void Request_Return(code, contlen, datalen)
word code, contlen, datalen;
{ word temp;

  mcb->MsgHdr.FnRc      = code;
  mcb->MsgHdr.DataSize  = (USHORT) datalen;
  mcb->MsgHdr.ContSize  = (BYTE) contlen;
  temp                  = mcb->MsgHdr.Dest;
  mcb->MsgHdr.Dest      = mcb->MsgHdr.Reply;
  mcb->MsgHdr.Reply     = temp;
  mcb->Timeout          = 0x77FFFL;

  if (debugflags & Message_Flag)
   ServerDebug("Reply  : fn %lx, port %lx, csize %ld, dsize %ld",
             code, mcb->MsgHdr.Dest, contlen, datalen);

  if (mcb->MsgHdr.Dest eq 0L)
      return;

#if AMIGA
  if (!PutHeliosMsg(mcb))
#else  
  if (!PutMsg(mcb))
#endif
#if AMIGA
    { output("***\r\n");
      output("*** Serious : the root transputer failed to receive a message.\r\n");
      output("***           To reboot the system, select the ServerWindow\r\n");
      output("***           and use the Reboot option in the Control menu.\r\n");
      output("***\r\n");
    }
#else
    { output("***\r\n");
      output("*** Serious : the root transputer failed to receive a message.\r\n");
      output("***           To reboot the system, use the Control, Shift and\r\n");
      output("***           F10 keys at the same time.\r\n");
      output("***\r\n");
    }
#endif

  mcb->MsgHdr.Flags = 0;     /* Clean out these entries */
  mcb->MsgHdr.Dest  = 0L;
  mcb->MsgHdr.Reply = 0L;
}


