#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <codes.h>

#define MYDEBUG

typedef struct Point {
	short	x ;
	short	y ;
} Point ;

typedef struct PolyNode {
	struct PolyNode	*NextNew ;
	struct PolyNode	*Left ;
	struct PolyNode *Right ;
	struct PolyNode *Parent ;
	struct PolyNode *Link[2] ;
	int		X ;
	int		Y ;
} PolyNode ;

#define PREV	0
#define NEXT	1

#define X_FRAC_SHIFT		11	/* All X co-ordinates in fractions */
#define HALF_X_FRAC		1024	/* Half of the fraction amount */

#ifndef WindingRule
#	define WindingRule	1
#	define EvenOddRule	2
#	define Complex		1
#	define Nonconvex	2
#	define Convex		4
#endif

/* -------------------------------------------------------------------------------- */
 
void PolyInsert( PolyNode *root, PolyNode *node )
{
	while( 1 )
	{
		if( root->Y < node->Y )
		{
			if( ! root->Right )
			{
				root->Right = node ;
				break ;
			}
			else	root = root->Right ;
		}
		else if( root->Y == node->Y  &&  root->X <= node->X )
 		{
			if( ! root->Right )
			{
				root->Right = node ;
				break ;
			}
			else	root = root->Right ;
		}
		else if( ! root->Left )
		{
			root->Left = node ;
			break ;
		}
		else	root = root->Left ;
	}

	node->Parent = root ;	
}

/* ................................................................................ */

int PolyRemove( PolyNode *lowX, PolyNode *highX, PolyNode **free, int n )
{
	PolyNode *next ;
	
	do
	{
		/* Find the next node and remove lowX */
		if( lowX->Right )
		{
			next = lowX->Right ;
			while( next->Left )	next = next->Left ;
			
			/* Set the left to be the old left */
			next->Left = lowX->Left ;
			
			/* And the parent left to be the old right */
			lowX->Parent->Left = lowX->Right ;
		}
		else 	
		{
			next = lowX->Parent ;
			lowX->Parent->Left = lowX->Left ;
		}
		
		/* Add it to the free list */
		lowX->Right = *free ;
		*free = lowX ;
		
		/* And set lowX to next and decrement count */
		lowX = next ;
		n-- ;
	}
	while( lowX->Y == highX->Y  &&  lowX->X < highX->X ) ;

	return( free ) ;
}

/* -------------------------------------------------------------------------------- */

void FillBlock( tlx, tly, trx, try, blx, bly, brx, bry, col )
int tlx, tly, trx, try, blx, bly, brx, bry, col ;
{
	tlx = ( tlx + HALF_X_FRAC ) >> X_FRAC_SHIFT ;
	trx = ( trx + HALF_X_FRAC ) >> X_FRAC_SHIFT ;
	blx = ( blx + HALF_X_FRAC ) >> X_FRAC_SHIFT ;
	brx = ( brx + HALF_X_FRAC ) >> X_FRAC_SHIFT ;
	
	printf("FillBlock( %d, %d, %d, %d, %d, %d, %d, %d, %d )\n",
			tlx, tly, trx, try, blx, bly, brx, bry, col ) ;
}

/* -------------------------------------------------------------------------------- */

int FillConvex( PolyNode *tl, int n, int cw, int fg )
{
	int		acw = !cw, tp = 1 ;
	PolyNode	*tr, *bl, *br, t[2] ;
	int		dx, dy1, dy2 ;

	/* Work out top right */
	if( tl->Link[cw]->Y != tl->Y )	tr = tl->Link[cw] ;
	else				tr = tl ;

	while( n > 2 )
	{
		/* Work out bottom two */
		bl = tl->Link[acw] ;
		br = tr->Link[cw] ;
		
		/* Switch over temporary pointers */
		tp = !tp ;

		/* If the bottom do not lie on the same y, create a temporary */
		if( bl->Y < br->Y )
		{
			dx 	= br->X - tr->X ;
			dy1 	= br->Y - tr->Y ;
			dy2 	= bl->Y - tl->Y ;
			
			dx	= ( dy2 * dx ) / dy1 ;
			
			t[tp].Y 	= bl->Y ;
			t[tp].X 	= tr->X + dx ;
			t[tp].Link[cw] 	= br ;
			br 		= &t[tp] ;
		}
		else if( bl->Y > br->Y )
		{
			dx	= bl->X - tl->X ;
			dy1	= bl->Y - tl->Y ;
			dy2	= br->Y - tr->Y ;
			
			dx	= ( dy2 * dx ) / dy1 ;

			t[tp].Y		= br->Y ;
			t[tp].X		= tl->X + dx ;
			t[tp].Link[acw]	= bl ;
			bl		= &t[tp] ;
		}

		/* Now fill this partial block */		
		FillBlock(	tl->X, tl->Y, tr->X, tr->Y,
				bl->X, bl->Y, br->X, br->Y,
				fg 	) ;
		
		/* Decrement count and Set the new tops to be the old bottoms */
		tl = bl ;	tr = br ;	
		n -= 2 ;
	}
	
	return( 0 ) ;
}

/* ................................................................................ */

int FillNonConv( PolyNode *root, PolyNode *tl, int cw, int fg )
{
	int		acw = !cw ;
	PolyNode	*tr, *bl, *br, *free, *new, *next_y ;
	int		dx, dy1, dy2, y ;

	/* Initialise free and new */
	free = new = NULL ;
	
	/* Work out top right */
	if( tl->Link[cw]->Y == tl->Y )	tr = tl->Link[cw] ;
	else				tr = tl ;

	while( n > 2 )
	{
		/* Work out bottom two */
		bl = tl->Link[acw] ;
		br = tr->Link[cw] ;

		/* See if there is concave interference in the y plane */		
		if( tl->Right )	
		{
			next_y = tl->Right ;
			while( next_y->Left )	next_y = next_y->Left ;
		}
		else	next_y = tl->Parent ;
		
		if( next_y )
		{
			/* See if there is with respect to the left side */
			if( next->y < bl->Y )
		/* If the bottom do not lie on the same y, create a temporary */
		if( bl->Y < br->Y )
		{
			if( !free )
			{
				if( ( free = malloc( sizeof( PolyNode ) ) ) != NULL )
				{
					if( new )	free->NextNew = new ;
					else		free->NextNew = NULL ;
					
					free->Right = NULL ;
					new = free ;
				}
				else	break ;	
			}
						
			dx 	= br->X - tr->X ;
			dy1 	= br->Y - tr->Y ;
			dy2 	= bl->Y - tl->Y ;
			
			dx	= ( dy2 * dx ) / dy1 ;
			
			free->Y 	= bl->Y ;
			free->X 	= tr->X + dx ;
			free->Link[cw] 	= br ;
			br 		= free ;
			free		= free->Right ;

			PolyInsert( root, br ) ;
		}
		else if( bl->Y > br->Y )
		{
			if( !free )
			{
				if( ( free = malloc( sizeof( PolyNode ) ) ) != NULL )
				{
					if( new )	free->NextNew = new ;
					else		free->NextNew = NULL ;
					
					free->Right = NULL ;
					new = free ;
				}
				else	break ;
			}

			dx	= bl->X - tl->X ;
			dy1	= bl->Y - tl->Y ;
			dy2	= br->Y - tr->Y ;
			
			dx	= ( dy2 * dx ) / dy1 ;

			free->Y 	= br->Y ;
			free->X 	= tl->X + dx ;
			free->Link[acw] = bl ;
			bl 		= free ;
			free		= free->Right ;

			PolyInsert( root, bl ) ;
		}

		/* Now fill this partial block */		
		FillBlock(	tl->X, tl->Y, tr->X, tr->Y,
				bl->X, bl->Y, br->X, br->Y,
				fg 	) ;

		/* Add the old tops to the free list (and remove from tree) */
		free = PolyRemove( tl, tr, free ) ;
				
		/* Get the first top from the list */
		tl = bl ;	tr = br ;
	}

	/* Free off any nodes allocated in this routine */
	while( new )
	{
		tl = new->
		
	if( ! free )	return( EC_Error | EG_NoMemory ) ;		
	else		return( 0 ) ;
}

/* ................................................................................ */

int FillNonConv( PolyNode *tl, int cw, int fg )
{
	tl = tl ; cw = cw ; fg = fg ;
	
	return( 0 ) ;
}

/* ................................................................................ */

int FillEvenOdd( PolyNode *tl, int cw, int fg )
{
	tl = tl ; cw = cw ; fg = fg ;
	
	return( 0 ) ;
}

/* ................................................................................ */

int FillWinding( PolyNode *tl, int cw, int fg )
{
	tl = tl ; cw = cw ; fg = fg ;
	
	return( 0 ) ;
}

/* -------------------------------------------------------------------------------- */

int PolyDraw( Point *xy, int count, int rule, int shape, int fg )
{
	PolyNode	*curr = NULL, *head ;
	Point		*pxy = xy + count - 1 ;
	int		i, cw, error, points = 0 ;

	/* Allocate memory for PolyNodes for each point */
	if( ( head = malloc( sizeof( PolyNode ) * count ) ) == NULL )
		return( EC_Error | EG_NoMemory ) ;
		
	/* Build up the ordered lists and ring of points (getting rid of adjacent duplicates) */
	for( i = 0 ; i < count ; i++, xy++ )
	{
		if( xy->x != pxy->x  ||  xy->y != pxy->y )
		{
			if( ! curr )
			{
				head->X = xy->x << X_FRAC_SHIFT ;
				head->Y = xy->y ;
				
				head->Parent = NULL ;
				curr = head ;
			}
			else
			{
				PolyNode *prev = curr++ ;

				curr->Link[PREV] = prev ;
				prev->Link[NEXT] = curr ;
				curr->X = xy->x << X_FRAC_SHIFT ;
				curr->Y = xy->y ;
				PolyInsert( head, curr ) ;
			}
			
			curr->Left = curr->Right = NULL ;
			pxy = xy ;
		}
		else
		{
			i-- ;
			count-- ;
		}
	}

	if( count > 2 )
	{
		curr->Link[NEXT] = head ;
		head->Link[PREV] = curr ;
		
		/* Now find the point of lowest y (and then lowest x if more than one y) */
		curr = head ;
		while( curr->Left )	curr = curr->Left ;
	
		/* Find out which edge runs clockwise */
		{
			int dx1, dy1, dx2, dy2 ;
			
			dx1 = curr->Link[PREV]->X - curr->X ;
			dy1 = curr->Link[PREV]->Y - curr->Y ;
			dx2 = curr->Link[NEXT]->X - curr->X ;
			dy2 = curr->Link[NEXT]->Y - curr->Y ;
			
			if( dy1 < dy2 )
			{
				if( dy1 )	dx1 = ( dy2 * dx1 ) / dy1 ;
				else		dx1 = dx2 + 1 ;
			}
			else if( dy1 > dy2 )	
			{
				if( dy2 )	dx2 = ( dy1 * dx2 ) / dy2 ;
				else		dx2 = dx1 + 1 ;
			}
			
			if( dx1 < dx2 )	cw = NEXT ;
			else		cw = PREV ;
		}
	
		/* Now we have initialised everything call the actual filling routine */
		if( shape == Complex )
		{
			if( rule == WindingRule )
				error = FillWinding( curr, cw, fg ) ;
			else				
				error = FillEvenOdd( curr, cw, fg ) ;
		}
		else if( shape == Nonconvex )
		{	
			error = FillNonConv( head, curr, count, cw, fg ) ;
		}
		else	error = FillConvex( curr, count, cw, fg ) ;
	}
	else
	{
		printf("DrawLine()\n") ;
	}
	
	/* Free off allocated memory */
	free( head ) ;
	
	return( error ) ;
}

/* ================================================================================ */

#define		MAX_POINTS	20

static Point	data[MAX_POINTS] ;
static char	buf[BUFSIZ] = "Dummy" ;

int main( void )
{
	while( *buf )
	{
		fputs( "> ", stdout ) ;	fflush( stdout ) ;
		gets( buf ) ;
		
		if( *buf )
		{
			int 	i = 0 ;
			char 	*s = buf ;
			
			while( i < MAX_POINTS )
			{
				data[i].x = strtol( s, &s, 0 ) ;
				data[i].y = strtol( s, &s, 0 ) ;
				
				if( data[i].y == 0 )	break ;
				else			i++ ;
			} 
				
			PolyDraw( data, i, WindingRule, Convex, 3 ) ;			
		}
	}
	
	return( 0 ) ;
}

/* -------------------------------------------------------------------------------- */

#ifdef BLITSCRN
int blitscrn( void )
{
	clock_t t ;
	graphics_init() ;

	Point xy[] = {  {  25,  1 }, 
			{  75,  1 }, 
			{ 100, 50 }, 
			{  75, 40 },
			{ 100,100 }, 
			{   1,100 }, 
			{  25, 40 }, 
			{   1, 50 }, 
			{  25,  1 } 	} ;
	int	points = sizeof( xy ) / sizeof( Point ) ;
	
	t = clock() ;
	{
		int x, y, i, c = 2 ;
		
		puts( "Start" ) ;
		
		for( i = 1 ; i <= 100 ; i++ )
		{
			byt_blkfill( (char*)cmode->addr, 1<<cmode->y_shft, 1,1, 100, 100, c ) ;
/*			byt_polyfill( (char*)cmode->addr, 1<<cmode->y_shft,  */
		}
	}
}

#endif
