#include <stdio.h>
#include <stdlib.h>
#include <codes.h>

#include <blossom.h>

/* Structure used to send in data */
typedef struct CoOrd {
	short	x ;
	short	y ;
} CoOrd ;

/* Structure used to store the active edges */
typedef struct Edge {
	struct Edge	*NextX ;	/* Next X Node */
	short		y ;		/* Value of y */
	short		dy ;		/* Delta y */
	int		x ;		/* Start X co-ordinate (in Nibble Fractions) */
	int		m ;		/* dx/dy */
} Edge ;

/* Structure used to store Y Buckets */
typedef struct YBucket {
	struct YBucket	*NextY ;	/* Next Y Bucket */
	struct YBucket	*PrevY ;	/* Prev Y Bucket */
	Edge		*NextX ;	/* First X Entry */
	int		y ;		/* Value of y */
} YBucket ;

#define SOAK_TEST

extern void bytblt( char *, char *, int,int, int,int, int,int, int ) ;

static const int	Convex		= 1,
			Nonconvex	= 2;

void			(*PolyBlit)() ;

/* The following variables are extracted from the byte based knowledge of the system */
static const int	BitsPerNibble		= 4 ;
static const int	NibblesPerByteShift	= 1 ;
static const int	BitsPerWord		= 32 ;
static const int	NibbleFractions		= 2048 ;	/* Accuracy of nibble calcs */
static const int	BlitFractions		= 4096 ;
static BlitterPtr	blitter 		= BLITTER_BASE ;

/* The following variables describe the mode characteristics */
int			ScreenBase,		/* Byte address of base of screen */
			ScreenStride,		/* Screen stride in pixels */
			ScreenHeight,		/* Screen height in pixels */
			ScreenWidth,		/* Screen width in pixels */
			ScreenDepth ;		/* Number of bits per pixel */

/* The following global variables are set up by InitBlossomGraphics */
int			NibblesPerPixel,	/* Number of nibbles per pixel */
			NibblesPerPixelShift,	/* Amount to shift to convert */
			NibbleFractionShift,	/* Shift to use to get Nibbles from fraction */
			NibbleFractionMask,	/* Mask to use to extract pixel fraction */
			StrideinNibbles,	/* The stride of the screen in nibbles */
			StrideinNibblesShift,
			StrideinNibblesMask,
			ScreenBaseinNibbles,	/* Base address in nibbles */
			PixelFractionShift,	/* Amount to shift to convert */
			HalfPixelinFractions,	/* Half a pixel in fractions */
			BlitFractionMask ;	/* Mask to use to extract blit fraction */

/* The following are used where graphics are done by the T8 */
int			StrideinBytes ,		/* Stride of screen in bytes */
			StrideinBytesShift,	/* Stride in bytes */
			StrideinBytesMask,
			ByteFractionMask,	/* Extract bytes from fractions */
			ByteFractionShift ;	/* Amount to shift to get to bytes */
			
static int		BCforTrapeziumFill ;	/* Value of BC to use for trap fill */


/* The following macros are used to alter between formats */
#define PixelstoFractions( x )	( (x) << PixelFractionShift	)
#define FractionstoPixels( x )	( (x) >> PixelFractionShift	)
#define NibbleFractions( x )	( (x) &  NibbleFractionMask	)
#define WholeNibbles( x )	( (x) >> NibbleFractionShift	)

#define YtoNibbAddr( x )	( (x) << StrideinNibblesShift	)
#define NibbAddrtoY( x )	( (x) >> StrideinNibblesShift	)
#define NibbAddrtoX( x )	( (x) &  StrideinNibblesMask	)

#define YtoByteAddr( x )	( (x) << StrideinBytesShift	)
#define ByteAddrtoY( x )	( (x) >> StrideinBytesShift	)
#define ByteAddrtoX( x )	( (x) &  StrideinBytesMask	)
#define WholeBytes( x )		( (x) >> ( NibbleFractionShift + NibblesPerByteShift ) )

#define NibblestoBytes( x )	( (x) >> NibblesPerByteShift	)
#define BytestoNibbles( x )	( (x) << NibblesPerByteShift	)

#define GetBlitFractions( x )	( (x) &  BlitFractionMask	)

#define ExpandColour( x )	{ int n = ScreenDepth ; \
				  while( n < BitsPerWord )  (x) |= ((x) << n), n <<= 1 ; }

/* The following macros are used to ease blitter use */
#define InitFillBlit( b,d ) 	( (b)->PM = 0xFFFFFFFF, \
				  (b)->BC = BCssp + BCssh, \
				  (b)->SD = (d), \
				  (b)->BF = BFcopy, \
				  (b)->TC = 0x1 )

/* -------------------------------------------------------------------------------- */

void PrintStaticGlobalVars( void )
{
	printf( "BitsPerNibble         = %d\n", BitsPerNibble ) ;
	printf( "NibblesPerByteShift   = %d\n", NibblesPerByteShift ) ;
	printf( "BitsPerWord           = %d\n", BitsPerWord ) ;
	printf( "NibbleFractions       = %d\n", NibbleFractions ) ;
	printf( "BlitFractions         = %d\n", BlitFractions ) ;
	putchar( '\n' ) ;
	printf( "ScreenBase            = %#x\n", ScreenBase ) ;
	printf( "ScreenStride          = %d\n", ScreenStride ) ;
	printf( "ScreenHeight          = %d\n", ScreenHeight ) ;
	printf( "ScreenWidth           = %d\n", ScreenWidth ) ;
	printf( "ScreenDepth           = %d\n", ScreenDepth ) ;
	putchar( '\n' ) ;
	printf( "NibblesPerPixel       = %d\n", NibblesPerPixel ) ;
	printf( "NibblesPerPixelShift  = %d\n", NibblesPerPixelShift ) ;
	printf( "NibbleFractionMask    = %d\n", NibbleFractionMask ) ;
	printf( "NibbleFractionShift   = %d\n", NibbleFractionShift ) ;
	printf( "StrideinNibbles       = %d\n", StrideinNibbles ) ;
	printf( "StrideinNibblesShift  = %d\n", StrideinNibblesShift ) ;
	printf( "StrideinNibblesMask   = %d\n", StrideinNibblesMask ) ;
	printf( "ScreenBaseinNibbles   = %#x\n", ScreenBaseinNibbles ) ;
	printf( "PixelFractionShift    = %d\n", PixelFractionShift ) ;
	printf( "HalfPixelinFractions  = %d\n", HalfPixelinFractions ) ;
	printf( "BlitFractionMask      = %d\n", BlitFractionMask ) ;
	putchar( '\n' ) ;
	printf( "StrideinBytes         = %d\n", StrideinBytes ) ;
	printf( "StrideinBytesShift    = %d\n", StrideinBytesShift ) ;
	printf( "StrideinBytesMask     = %d\n", StrideinBytesMask ) ;
	printf( "ByteFractionMask      = %d\n", ByteFractionMask ) ;
	printf( "ByteFractionShift     = %d\n", ByteFractionShift ) ;
}

/* -------------------------------------------------------------------------------- */

void InitBlossomGraphics( int mode )
{
	int i ;
	
	switch( mode )
	{
		case	0 :	/* Mode 0, etc */
		{
			/* Fill in for mode 0 when ready */
		}
		
		default	:	/* 1024 by 768, 8-bit pixels, 256 Colours */
		{
			ScreenBase	= 0x00200000 ;
			ScreenStride	= 2048 ;
			ScreenHeight	= 768 ;
			ScreenWidth	= 1024 ;
			ScreenDepth	= 8 ;
		}
	}

	NibblesPerPixel = 1 ;
	for( i = BitsPerNibble ; i < ScreenDepth ; i += BitsPerNibble )
		NibblesPerPixel++ ;
		
	NibblesPerPixelShift = 0 ;
	for( i = NibblesPerPixel ; i > 1 ; i >>= 1 )
		NibblesPerPixelShift++ ;
	
	NibbleFractionMask = NibbleFractions - 1 ;

	NibbleFractionShift = 0 ;
	for( i = NibbleFractions ; i > 1 ; i >>= 1 )
		NibbleFractionShift++ ;
		
	PixelFractionShift 	= NibbleFractionShift + NibblesPerPixelShift ;
	HalfPixelinFractions	= ( 1 << ( PixelFractionShift - 1 ) ) - 1 ;

	StrideinNibbles		= ScreenStride << NibblesPerPixelShift ;
	StrideinNibblesMask	= StrideinNibbles - 1 ;
	
	StrideinNibblesShift	= 0 ;
	for( i = StrideinNibbles ; i > 1 ; i >>= 1 )
		StrideinNibblesShift++ ;
		

	ScreenBaseinNibbles	= ScreenBase << NibblesPerByteShift ;

	BlitFractionMask	= BlitFractions - 1 ;
	
	StrideinBytes		= StrideinNibbles >> NibblesPerByteShift ;
	StrideinBytesMask	= StrideinBytes - 1 ;
	StrideinBytesShift	= StrideinNibblesShift - NibblesPerByteShift ;
	ByteFractionMask	= ( NibbleFractions << NibblesPerByteShift ) - 1 ;
	ByteFractionShift	= NibbleFractionShift + NibblesPerByteShift ;
	
	BCforTrapeziumFill 	= BCgo + BCfcen*1 + BCeols*0 
				+ ( BCdps + BCsps ) * NibblesPerPixelShift + BCdxc*1
				+ BCdyc*0 + BCsxc*1 + BCsyc*0 + BCdst*1 ; 
	
/*	PrintStaticGlobalVars() ;	*/
}

/* -------------------------------------------------------------------------------- */

void bytblt2( char *s, char *d, int os, int od, int ss, int ds, int w, int h, int op )
{
	ss = (int)s - ScreenBase ;
	os = ByteAddrtoY( ss ) ;
	od = ByteAddrtoX( ss ) ;
	d  = d ;
	ds = ds ;
	h  = h ;
	op = op ;
	 	
	printf("bytblt @ ( %d,%d ) for %d to %d\n", od, os, w, od + w ) ;
}

/* -------------------------------------------------------------------------------- */

void LineBlit( int x1, int y1, int dy, int m, int fg )
{
	int x2 ;
	int y2 ; 
	
	if( !dy )	x2 = x1 + m ;
	else		x2 = x1 + m * dy ;

	y2 	= y1 + dy ;
	
	x1 	= FractionstoPixels( x1 ) ;
	x2 	= FractionstoPixels( x2 ) ;

	printf("Line : dy = %3d, m = %6d, (%3d,%3d), (%3d,%3d), fg = %d\n", dy, m, x1, y1, x2, y2, fg ) ;
}

/* -------------------------------------------------------------------------------- */

/* Code to emulate the blitter using the T8 */

void EmulateB( lx, lf, ls, li, rx, rf, rs, ri, y, dy, fg )
int lx, lf, ls, li, rx, rf, rs, ri, y, dy, fg ;
{
	char 	*addr ;
	int	dx ;

	dx = YtoNibbAddr( y ) ;
 	lx += dx ;
	rx += dx ;

/*	printf("PolyBlit( %7d, %7d ) dy = %4d, f(%d,%d), i(%d,%d)\n", 
						lx, rx, dy, lf, rf, li, ri ) ;
*/
	while( dy > 0 )
	{
		/* Plot points */
		y	= lx - NibbAddrtoX( lx ) ;		/* y in nibbles */
		dx	= NibblestoBytes( lx - y ) ;		/* x in bytes */
		addr 	= (char*)( ScreenBase + YtoByteAddr( NibblestoBytes( y ) ) + dx ) ;
		dx	= NibblestoBytes( rx - y ) - dx  ; 	/* dx in bytes */
		*addr	= fg ;
			
		bytblt( addr,addr+1, 0,0, 0,1, dx,1, 0 ) ;

		if( lx > rx  ||  ( lx == rx  &&  lf >= rf ) )
		{
/*			bytblt2( addr,addr+1, 0,0, 0,1, dx,1, 0 ) ;	*/
			printf("Overlap( %7d, %7d ) dy = %4d, f(%4d,%4d), s(%5d,%5d), i(%4d,%4d)\n", 
					lx, rx, dy, lf, rf, 
					ls-StrideinNibbles, rs-StrideinNibbles, li, ri ) ;
		}
		
		lx += ls ;
		lf += li ;
		if( lf >= BlitFractions )
		{
			lf -= BlitFractions ;
			lx++ ;
		}
		
		rx += rs ;
		rf += ri ;
		if( rf >= BlitFractions )
		{
			rf -= BlitFractions ;
			rx++ ;
		}
		
		dy-- ;
	}
}

/* -------------------------------------------------------------------------------- */

void PolyBlitB( int tlx, int blx, int trx, int brx, int y, int dy, int fg )
{
	int	li, ri ;	/* Right and left hand side increment (fracs) */
	int	ls, rs ;	/* Left and right hand stride in nibbles */
	int	lx, rx ;	/* Start x for left and right sides */
	int	dx ;		/* Difference in x */
	

	/* If the top two are the same increment the rhs by 1 */
	if( trx == tlx ) trx++ ;
	
	/* Get the difference at the bottom */
	dx = brx - blx ;
	
	/* If there is no margin for rounding add the margin */
	if( dx < dy )
	{
		brx = ( brx + blx + dy ) >> 1 ;
		blx = brx - dy ;
	}

	/* Now calculate out the start positions */
	lx = WholeNibbles( tlx ) ;
	rx = WholeNibbles( trx ) ;

	dx = ScreenBaseinNibbles + YtoNibbAddr( y ) ;
	lx += dx ;
	rx += dx ;

	/* Initialise those parts of the blitter which we need to */
	InitFillBlit( blitter, fg ) ;
	blitter->DL = dy ;
	blitter->DS = lx ;
	blitter->DF = GetBlitFractions( tlx ) ;
	blitter->SS = rx ;
	blitter->SF = GetBlitFractions( trx ) ;
	
	/* Now do the lhs */
	if( blx >= tlx )
	{
		dx = blx - tlx ;
		li = dx / dy ;
		ls = WholeNibbles( li ) ;
		blitter->DT = StrideinNibbles + ls ;
		blitter->DI = GetBlitFractions( li ) ;
	}
	else
	{
		dx = tlx - blx ;
		li = dx / dy ;
		ls = WholeNibbles( li ) ;
		li = GetBlitFractions( li ) ;
		
		if( li )
		{
			blitter->DT = StrideinNibbles - ls - 1 ;
			blitter->DI = BlitFractions - li ;
		}
		else	
		{
			blitter->DT = StrideinNibbles - ls ;
			blitter->DI = li ;
		}
	}
	
	/* And similarly for the rhs ( forcing right ) */
	if( brx >= trx )
	{
		dx = brx - trx ;
		ri = dx / dy ;
		rs = WholeNibbles( ri ) ;
		blitter->ST = StrideinNibbles + rs ;
		blitter->SI = GetBlitFractions( ri ) ;
	}
	else
	{
		dx = trx - brx ;
		ri = dx / dy ;
		rs = WholeNibbles( ri ) ;
		ri = GetBlitFractions( ri ) ;

		if( ri )
		{
			blitter->ST = StrideinNibbles - rs - 1 ;
			blitter->SI = BlitFractions - ri ;
		}
		else	
		{
			blitter->ST = StrideinNibbles - rs ;
			blitter->SI = ri ;
		}
	}

	/* Now tell it to start */
	blitter->BC = BCforTrapeziumFill ;

/*	EmulateB( lx, lf, ls, li, rx, rf, rs, ri, y, dy, fg ) ; */
}

/* -------------------------------------------------------------------------------- */

void PolyBlitT( int tlx, int blx, int trx, int brx, int y, int dy, int fg )
{
	int	lx, rx ;	/* Right and left start positions */
	int	li, ri ;	/* Right and left hand side increment (fracs) */
	int	ls, rs ;	/* Left and right hand stride in nibbles */
	int	lf, rf ;	/* Left and right Fractions */
	int	dx ;		/* Difference in x */
	char	*addr ;
	
	/* Lift dy up to extra line */
	dy++ ;

	/* Modify co-ordinates to remove any overlaps */
	if( trx == tlx ) trx++ ;/* Increment trx to prevent overlaps when trx == tlx */
	dx = brx - blx ;	/* Get the difference of the bottoms */
	if( dx < dy )		/* Add the two, add the effective error term dy, and average */
	{
		brx = ( brx + blx + dy ) >> 1 ;
		blx = brx - dy ;
	}
	
	/* Calculate lhs */
	if( blx >= tlx )
	{
		dx = blx - tlx ;
		li = dx / dy ;
		ls = WholeBytes( li ) ;
		li = GetBlitFractions( li ) ;
		ls += StrideinBytes ;
	}
	else
	{
		dx = tlx - blx ;
		li = dx / dy ;
		ls = WholeBytes( li ) ;
		if( li )	tlx -= ( li - 1 ) ;
		li = GetBlitFractions( li ) ;
		
		if( li )
		{
			ls = StrideinBytes - ls - 1 ;
			li = BlitFractions - li ;
		}
		else	ls = StrideinBytes - ls ;
	}
	
	/* And similarly for the rhs */
	if( brx >= trx )
	{
		dx = brx - trx ;
		ri = dx / dy ;
		rs = WholeBytes( ri ) ;
		if( ri )	trx += ( ri - 1 ) ;
		ri = GetBlitFractions( ri ) ;
		rs += StrideinBytes ;
	}
	else
	{
		dx = trx - brx ;
		ri = dx / dy ;
		rs = WholeBytes( ri ) ;
		ri = GetBlitFractions( ri ) ;
		
		if( ri )
		{
			rs = StrideinBytes - rs - 1;
			ri = BlitFractions - ri ;
		}
		else	rs = StrideinBytes - rs ;
	}
		
	/* And finally calculate out the start positions */
	lx = WholeBytes( tlx ) ;
	rx = WholeBytes( trx ) ;
	lf = GetBlitFractions( tlx ) ;
	rf = GetBlitFractions( trx ) ;

	dx = ScreenBase + YtoByteAddr( y ) ;
 	lx += dx ;
	rx += dx ;

/*	printf("PolyBlit( %7d, %7d ) dy = %4d, f(%d,%d), i(%d,%d)\n", 
						lx, rx, dy, lf, rf, li, ri ) ;
*/
	while( dy > 0 )
	{
		/* Plot points */
		addr	= (char*)lx ;
		*addr	= fg ;
			
		bytblt( addr,addr+1, 0,0, 0,1, rx-lx,1, 0 ) ;

		lx += ls ;
		lf += li ;
		if( lf >= BlitFractions )
		{
			lf -= BlitFractions ;
			lx++ ;
		}

		rx += rs ;
		rf += ri ;
		if( rf >= BlitFractions )
		{
			rf -= BlitFractions ;
			rx++ ;
		}

		dy-- ;
	}
}

/* -------------------------------------------------------------------------------- */

void PolyPrint( char *txt , YBucket *bucket )
{
	Edge 	*edge ;
	char	type ;
	
	while( bucket )
	{
		edge = bucket->NextX ;
		type = 'Y' ;
					
		while( edge )
		{
			printf("%-10s %c : ", txt, type ) ;
			LineBlit( edge->x, edge->y, edge->dy, edge->m, 0 ) ;
			edge = edge->NextX ;			
			type = 'X' ;
		}
		
		bucket = bucket->NextY ;
	}
}

/* -------------------------------------------------------------------------------- */

/* RemoveX() removes the given X node from the list (and removes the y bucket if appropriate */

YBucket *RemoveX( Edge *edge, YBucket *bucket )
{
	Edge	*prev ;
	YBucket *curr = NULL ;

	if( bucket->y == edge->y  &&  bucket->NextX )
	{
		/* Special case if edge is the first one */
		if( bucket->NextX != edge )
		{
			prev = bucket->NextX ;
			while( prev->NextX )
			{
				if( prev->NextX == edge )
				{
					prev->NextX = edge->NextX ;
					break ;
				}
				else	prev = prev->NextX ;
			}
		}
		else	bucket->NextX = edge->NextX ;
	
		edge->NextX = NULL ;
		
		/* If the bucket is empty also remove that */
		if( ! bucket->NextX )
		{
			if( bucket->NextY )
			{
				curr		= bucket->NextY ;
				curr->PrevY	= bucket->PrevY ;
			}
		
			if( bucket->PrevY )
			{
				curr		= bucket->PrevY ;
				curr->NextY	= bucket->NextY ;
			}
		}
		else	curr = bucket ;
	}
	else	curr = bucket ;

	return( curr ) ;
}

/* -------------------------------------------------------------------------------- */

/* InsertX() inserts the given x edge into the given y bucket. The edges are sorted in
 * x order, then slope
 */
 
void InsertX( Edge *edge, YBucket *bucket )
{
	Edge	*curr = bucket->NextX ;

	/* Reject if the two do not agree */
	if( bucket->y != edge->y )	return ;

	/* See if the edge should be inserted at the head of the list */
	if( edge->x < curr->x )
	{
		edge->NextX 	= bucket->NextX ;
		bucket->NextX 	= edge ;
	}
	else if( edge->x == curr->x  &&  edge->m < curr->m )
	{
		edge->NextX	= bucket->NextX ;
		bucket->NextX	= edge ;
	}
	else
	{
		while( curr->NextX  &&  curr->NextX->x < edge->x )
			curr = curr->NextX ;
			
		while( curr->NextX  &&  curr->NextX->x == edge->x )
		{
			if( curr->NextX->m > edge->m )	break ;
			else				curr = curr->NextX ;
		}
			
		edge->NextX 	= curr->NextX ;
		curr->NextX	= edge ;
	}
}

/* -------------------------------------------------------------------------------- */

/* DrawNonconv() is slightly more complex than a simple convex but still straight forward */

void DrawNonConv( YBucket *bucket, int fg )
{
	Edge	*lhs, *rhs, *edge ;
	YBucket	*curr ;
	int	dy, ymin, lx, rx ;
	 
	while( 1 )
	{
		/* Set up left hand side and right hand side */
		lhs = bucket->NextX ;
		
		if( !lhs  ||  !lhs->NextX )	break ;
		else				rhs = lhs->NextX ;

		/* Set up initial dy */
		if( lhs->dy < rhs->dy )	dy = lhs->dy ;
		else			dy = rhs->dy ;

		/* See if there is any interference */
		curr 	= bucket ;
		ymin 	= lhs->y + dy ;
		
		if( curr->NextY ) while( curr->NextY  &&  curr->NextY->y <= ymin )
		{
			curr 	= curr->NextY ;
			dy	= curr->y - lhs->y ;
			lx 	= lhs->x + dy * lhs->m ;
			rx	= rhs->x + dy * rhs->m ;
			edge 	= curr->NextX ;

			/* Note the error term dy (in case rounded) */
			while( edge )
			{
				if( edge->x > lx  &&  edge->x < rx + dy ) break ;
				else	edge = edge->NextX ;
			}
			
			if( edge )	break ;
		}
		else
		{
			lx	= lhs->x + dy * lhs->m ;
			rx	= rhs->x + dy * rhs->m ;
		}

		/* Draw the polygon */
		PolyBlit( lhs->x, lx, rhs->x, rx, lhs->y, dy, fg ) ;
		
		/* This time remove the lhs and right hand side from the list, remember next */
		bucket = RemoveX( lhs, bucket ) ;
		bucket = RemoveX( rhs, bucket ) ;

		/* Check if still some left on lhs and/or rhs */
		if( bucket )
		{
			if( dy < lhs->dy )
			{
				lhs->x 	= lx ;
				lhs->y	+= dy ;
				lhs->dy	-= dy ;
				InsertX( lhs, curr ) ;
			}
			
			if( dy < rhs->dy )
			{
				rhs->x 	= rx ;
				rhs->y	+= dy ;
				rhs->dy	-= dy ;
				InsertX( rhs, curr ) ;
			}
		}
		else	break ;
	}
} 	

/* -------------------------------------------------------------------------------- */

/* DrawConvex() is the simplest routine for drawing a completely convex polygon */

void DrawConvex( YBucket *bucket, int fg )
{
	Edge	*lhs, *rhs ;
	int	dy, lx, rx ;
	 
	while( 1 )
	{
		/* Set up left hand side and right hand side */
		lhs = bucket->NextX ;
		rhs = lhs->NextX ;

		/* Set up initial dy */
		if( lhs->dy < rhs->dy )	dy = lhs->dy ;
		else			dy = rhs->dy ;

		lx  = lhs->x + lhs->m * dy ;
		rx  = rhs->x + rhs->m * dy ;
		
		/* Draw the polygon */
		PolyBlit( lhs->x, lx, rhs->x, rx, lhs->y, dy, fg ) ;
		
		/* Increment bucket */
		if( bucket->NextY )
		{
			bucket = bucket->NextY ;		
			
			/* Check if still some left on lhs and/or rhs (remember slope) */
			if( dy  &&  dy < lhs->dy )
			{
				lhs->x 	= lx ;
				lhs->y	+= dy ;
				lhs->dy	-= dy ;
				InsertX( lhs, bucket ) ;
			}
			else if( dy  &&  dy < rhs->dy )
			{
				rhs->x 	= rx ;
				rhs->y	+= dy ;
				rhs->dy	-= dy ;
				InsertX( rhs, bucket ) ;
			}
		}
		else	break ;
	}
} 	
			
/* -------------------------------------------------------------------------------- */

int PolyDraw( CoOrd *xy, int points, int shape, int fg )
{
	YBucket *ybucket,		/* First y bucket allocated */
		*ynext,			/* Next free y bucket */
		*ycurr ;		/* Current y bucket */
	Edge	*xcurr ;		/* Current edge */
	int	i, edges, added,	/* Loop counter, number of edges, error term */
		x1, y1, x2, y2 ;	/* Co-ordinates */
	CoOrd	*xy1, *xy2 ;		/* Temporary xy co-ordinates */

	/* Allocate memory for edges */
	ybucket	= malloc( ( sizeof( YBucket ) + sizeof( Edge ) ) * points) ;
	
	/* If OK Initialise data pointers */
	if( ybucket )	
	{
		edges = points ;
		ynext = ybucket ;
		xcurr = (Edge*)( ybucket + edges) ;
		added = 0 ;
		ExpandColour( fg ) ;
	}
	else	return( EC_Error | EG_NoMemory ) ;

	/* See if the last point is the same as the first */
	xy1	= xy + edges- 1 ;

	if( xy1->x == xy->x  &&  xy1->y == xy->y )
	{
		edges-- ;
		xy1-- ;
	}

	/* Initialise appropriate x and y */
	x2 	= PixelstoFractions( xy1->x ) + HalfPixelinFractions ;
	y2 	= xy1->y ;
	xy2	= xy ;
	
	/* For each point calculate out the edge details */
	for( i = edges ; i > 0 ; i-- )
	{
		/* Set up x1, y1, and x2, y2 */
		x1	= x2 ;
		y1	= y2 ;
		x2	= PixelstoFractions( xy2->x ) + HalfPixelinFractions ;
		y2	= xy2->y ;
		xy2++ ;

		/* Find out greatest y, ignoring horizontal edges, and set values */
		if( y1 == y2 )
		{
			edges-- ;
			continue ;
		}

		/* Note that the slope is set to be the nearest (by adding dy/2) */
		if( y1 < y2 )
		{
			xcurr->x	= x1 ;
			xcurr->y	= y1 ;
			xcurr->dy 	= y2 - y1 ;
			xcurr->m	= ( x2 - x1 ) / xcurr->dy ;
		}
		else
		{
			xcurr->x	= x2 ;
			xcurr->y	= y2 ;
			xcurr->dy	= y1 - y2 ;
			xcurr->m	= ( x1 - x2 ) / xcurr->dy ;
		}

		xcurr->NextX	= NULL ;

		/* If this is the first one just set the structures */
		if( ! added ) 
		{
			ycurr		= ynext++ ;
			ycurr->NextY	= ycurr->PrevY	= NULL ;
			ycurr->NextX	= xcurr ;
			ycurr->y	= xcurr->y ;
			added		= 1 ;
			xcurr++ ;
		}
		else 
		{
			/* Find the nearest edge with y lower than or equal to xcurr->y */
			if( ycurr->y > xcurr->y )
			{
			 	while( ycurr->y > xcurr->y  &&  ycurr->PrevY )
					ycurr = ycurr->PrevY ;
			}
			else	while( ycurr->y < xcurr->y  &&  ycurr->NextY )
					ycurr = ycurr->NextY ;
			
			/* If one of same y not found insert a new one */
			if( ycurr->y < xcurr->y )
			{
				ynext->NextX	= xcurr ;
				ynext->y	= xcurr->y ;

				ynext->PrevY	= ycurr ;

				if( ycurr->NextY )
				{
					ynext->NextY		= ycurr->NextY ;
					ycurr->NextY->PrevY	= ynext ;
				}
				else	ynext->NextY = NULL ;
				
				ycurr->NextY	= ynext++ ;	/* Remember to increment next */
				xcurr++ ;			/* ditto */
			}
			else if( ycurr->y > xcurr->y )
			{
				ynext->NextX	= xcurr ;
				ynext->y	= xcurr->y ;
				
				ynext->NextY	= ycurr ;
				
				if( ycurr->PrevY )
				{
					ynext->PrevY		= ycurr->PrevY ;
					ycurr->PrevY->NextY	= ynext ;
				}
				else	ynext->PrevY = NULL ;
				
				ycurr->PrevY	= ynext++ ;
				xcurr++ ;
			}
			else
			{
				InsertX( xcurr, ycurr ) ;
				xcurr++ ;
			}
		}
	}

	if( edges > 1 )
	{
		/* Find lowest y */
		while( ycurr->PrevY )	ycurr = ycurr->PrevY ;
	
		/* Send it on to the drawing routine */
		if( shape == Convex )		DrawConvex( ycurr, fg ) ;
		else if( shape == Nonconvex )	DrawNonConv( ycurr, fg ) ;
		else				PolyPrint( "End", ycurr ) ;
	}
	else if( ! edges )	/* A Horizontal line or dot */
	{
		x1 = x2 = xy->x ;
		xy++ ;
		
		for( i = 1 ; i < points ; i++, xy++ )
		{
			if	( xy->x < x1 )	x1 = xy->x ;
			else if	( xy->x > x2 )	x2 = xy->x ;
		}
		
		x1 	= PixelstoFractions( x1 ) + HalfPixelinFractions ;
		x2 	= PixelstoFractions( x2 ) + HalfPixelinFractions ;

		LineBlit( x1, y1, 0, x2 - x1, fg ) ;
	}
	else			/* A Single edge */
	{
		xcurr = ycurr->NextX ;
		LineBlit( xcurr->x, xcurr->y, xcurr->dy, xcurr->m, fg ) ;
	}
	
	free( ybucket ) ;
	return( 0 ) ;
}

/* ================================================================================ */

#include <string.h>

#define	 MAX_POINTS	50

static CoOrd	data[MAX_POINTS] ;
static char	buf[BUFSIZ] = "Dummy" ;

int main( void )
{
	int	error ;
	
	InitBlossomGraphics( 1 ) ;

	while( *buf )
	{
		fputs( "> ", stdout ) ;	fflush( stdout ) ;
		gets( buf ) ;
		
		if( *buf )
		{
			int 	i = 0 ;
			char 	*s = buf ;
			int	fg ;

			PolyBlit = PolyBlitB ;
			
			while( i < MAX_POINTS )
			{
				data[i].x = strtol( s, &s, 0 ) ;
				data[i].y = strtol( s, &s, 0 ) ;
				
				if( data[i].y == 0 )	break ;
				else			i++ ;
			}
			
#ifdef SOAK_TEST
			{
				int	min_x, min_y, dx, dy, count ;
			
				min_x = data[0].x ;
				min_y = data[0].y ;
				dx    = data[1].x - min_x ;
				dy    = data[1].y - min_y ;
				count = data[2].x ;
				
				while( 1 )
				{
					srand( 1 ) ;
				
					for( i = count ; i > 0 ; i-- )
					{
						data[0].x = min_x + rand() % dx ;
						data[0].y = min_y + rand() % dy ;
						data[1].x = min_x + rand() % dx ;
						data[1].y = min_y + rand() % dy ;
						data[2].x = min_x + rand() % dx ;
						data[2].y = min_y + rand() % dy ;
						data[3].x = min_x + rand() % dx ;
						data[3].y = min_y + rand() % dy ;
						fg	  = rand() % 60 ;

						error = PolyDraw( data, 4, Nonconvex, fg ) ;
						if( error )	break ;
					}
					
					if( PolyBlit == PolyBlitB )	
					{
						printf( "Hit <Return> to Compare with T8 > " ) ;
						fflush( stdout ) ;
						gets( buf ) ;
						if( ! *buf )
						{
							PolyBlit = PolyBlitT ;
							*buf = 'Y' ;
						}
						else	break ;
					}
					else	break ;
				}
					
			}
#else
			error = PolyDraw( data, i, Nonconvex, rand() % 60 ) ;
#endif
		}
		
		if( error )	break ;
	}
	
	return( error ) ;
}
