/********************************************************
demos.c

Various routines to demonstrate Blossom
********************************************************/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "blomstr.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	 960
#define WIDTH2	(WIDTH/2)
#define HEIGHT2	(HEIGHT/2)

static int pfill0[16] = {
	0x00000000, 0x11111111, 0x22222222, 0x33333333,
	0x44444444, 0x55555555, 0x66666666, 0x77777777,
	0x88888888, 0x99999999, 0xAAAAAAAA, 0xBBBBBBBB,
	0xCCCCCCCC, 0xDDDDDDDD, 0xEEEEEEEE, 0xFFFFFFFF
};
#define PFILL0(c)	pfill0[(c)&15]			/* mode 0 */

static int pfill1L[16] = {
	0x00000000, 0x01010101, 0x02020202, 0x03030303,
	0x04040404, 0x05050505, 0x06060606, 0x07070707,
	0x08080808, 0x09090909, 0x0A0A0A0A, 0x0B0B0B0B,
	0x0C0C0C0C, 0x0D0D0D0D, 0x0E0E0E0E, 0x0F0F0F0F
};
static int pfill1H[16] = {
	0x00000000, 0x10101010, 0x20202020, 0x30303030,
	0x40404040, 0x50505050, 0x60606060, 0x70707070,
	0x80808080, 0x90909090, 0xA0A0A0A0, 0xB0B0B0B0,
	0xC0C0C0C0, 0xD0D0D0D0, 0xE0E0E0E0, 0xF0F0F0F0
};
#define PFILL1(c)	pfill1L[(c)&0x0F]+pfill1H[(c)&0xF0] /* mode 1 */

static int topmask[4] = {
	0xf000, 0xe000, 0xc000, 0x8000
};
static int botmask[4] = {
	0x0f00, 0x0100, 0x0300, 0x0700
};

/* low-level routines */

static int pDstBase;
static int delay;
static int ppw;			/* Pixels per word */
static int PixelSize;

static void
wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

/**********************************************************
General Blossom init program
**********************************************************/
static void
InitBlitter()
{
	BlossomPtr blossomPtr = BLOSSOM_BASE;
	BlitterPtr bPtr = BLITTER_BASE;

	blossomPtr->MC = 0x4A;	/* enable blitter */

	bPtr->MB = 0xff00;	/* top/bot masks both true */
	bPtr->BF = 10;		/* Bit function, = destination */
	bPtr->TC = 1;		/* FWR = 1. Ignore linear tests */
	bPtr->PM = 0xFFFFFFFF;	/* Planar mask */
	bPtr->MD = 0xFFFFFFFF;	/* Bit mask data */
	bPtr->BC = 0x00000000;	/* Reset everything */
}

static void
Stipple(int x, int y, int w, int h, int col)
{
	
WORD	start = ((y>>2)*2048*4)+x;	/* 4 line chunks */
	if (w<=0 || h<=0) return;
	{
	BlitterPtr bPtr = BLITTER_BASE;

	bPtr->DS = (MAP1+start)*2;	/* Nibble address */
	bPtr->DT = 2048*4*2;		/* Stride is 4 lines, in nibbles */
	bPtr->DW = (((w/ppw)-1)<<3);	/* width in words */
	bPtr->DL = ((h+(y&3)+3)/4)-1;		/* Length in lines */

	bPtr->SD = 0xffffffff;		/* Stipple pattern */
	bPtr->DD = PFILL1(col);		/* Colour (mode 1)*/

	bPtr->MB = (topmask[y&3]
		   +botmask[(y+h)&3]);		/* top/bot masks */

	bPtr->BC =	BCgo + BCstp + BCeols + /* Blitter control register */
			BCsrc*0 + BCdst*0 +
	        	BCdxc*1 + BCdyc*0 +
	        	BCsxc*1 + BCsyc*0;
	}
}
	


static void
FillTrap(int x, int y, int w, int h, int dx1, int dx2)
{
	int lStart = (pDstBase + y*2048)*2 + x;
	BlitterPtr bPtr = BLITTER_BASE;

 
	bPtr->DS = lStart;
	bPtr->SS = lStart + w - 1;
	bPtr->DT = 4096 + dx1;
	bPtr->ST = 4096 + dx2;
	bPtr->DL = h - 1;
	bPtr->SD = 0x12345678;
	bPtr->BC = (BCgo + BCfcen*0 + BCeols*0 +
		BCdps*PixelSize + BCsps*PixelSize +
		BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
}

static void DoTraps()
{
	int o, i, c;

	pDstBase = MAP1;
	printf("Test one, lstride = 0, rstide = 1\n");

	for (o=0; o<8; o++) {
		printf("offset %d\n", o);
		for (i=1; i<32; i++) {
			printf("Width %d\n", i);
			if (delay>1) wait(NULL);
			for (c=0 ; c<1000; c++) {
				if (delay>0) printf("\r%d", c);
				FillTrap( 128+o, 128+o*40, 1, i, 0, 1);
			}
		}
	}
}

int
main(int argc, char *argv[])
{
	ppw = 4;		/* Pixels per word */
	PixelSize = 1;		/* As required by Blossom */
	
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

	if (argc!=6) printf("x,y,width,height,colour\n");

	InitBlitter();
	Stipple(atoi(argv[1]),
		atoi(argv[2]),
		atoi(argv[3]),
		atoi(argv[4]),
		atoi(argv[5]));
	
/*	DoTraps(); */
}
