/* Initial test program for Blossom */

#include <stdio.h>
#include <math.h>
#include "blomstr.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	1024

/* low-level routines */

static int pDstBase;

static void
Wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

/**********************************************************
Initialise the graphics registers to default values

Assumes that the blitter has been enabled.
**********************************************************/
static void bl_init()
{

	BlossomPtr blossomPtr = BLOSSOM_BASE;
	BlitterPtr bPtr = BLITTER_BASE;

	blossomPtr->MC = 0x4A;	/* enable blitter */

	bPtr->MB = 0xff00;	/* top/bot masks both true */
	bPtr->BF = 10;		/* Bit function, = destination */
	bPtr->TC = 1;		/* FWR = 1. Ignore linear tests */
	bPtr->PM = 0xFFFFFFFF;	/* Planar mask */
	bPtr->MD = 0xFFFFFFFF;	/* Bit mask data */
	bPtr->BC = 0x00000000;	/* Reset everything */

}


#define NibblesPerPixel		1
#define PixelSize		0
#define DstStride		(2048*2)

static void
InitDrawLines(void)
{
	pDstBase = MAP1;
}

static void
DrawLine(int x1, int y1, int x2, int y2, int width) 
{
	int nibDstL, subDstL,  nibDstR, subDstR;
	int dx, dy, stride, nibStride, subStride;

	bl_init();

	if (y2==y1) return;

	if (y2<y1) {
		int tmp;
		tmp = x1; x1 = x2; x2 = tmp;
		tmp = y1; y1 = y2; y2 = tmp;
	}

	dx = x2 - x1 + 1;
	dy = y2 - y1 + 1;

	printf("Draw Line (%d,%d)->(%d,%d) dx=%d, dy=%d\n",
					x1, y1, x2, y2, dx, dy);

	nibDstL = (pDstBase + y1*2048)*2 + x1;
	subDstL = 0;

	if (dx>=0) {
		stride = dx*4096 / dy;
		subStride = stride & 4095;
		nibStride = DstStride + stride / 4096;
		stride = (stride+4096)*width;
		nibDstR = nibDstL + stride / 4096;
		subDstR = stride & 4095;
	} else {
		stride = -dx*4096 / dy;
		subStride = 4096 - (stride & 4095);
		nibStride =  DstStride - stride / 4096 - 1;
		stride = stride*width;
		nibDstR = nibDstL + stride / 4096;
		subDstR = stride & 4095;
	}

	nibDstR = nibDstL + width / 4096;
	subDstR = subDstL + ( width &4095);

	printf("L %x.%04x R %x.%04x Stride %x.%04x Height %d\n",
		nibDstL, subDstL, nibDstR, subDstR, nibStride, subStride, dy);

	Wait("About to draw");
	{
		BlitterPtr bPtr = BLITTER_BASE;

		bPtr->DL = dy - 1;

		bPtr->LS = nibDstL;
		bPtr->LF = subDstL;
		bPtr->LT = nibStride;
		bPtr->LI = subStride;

		bPtr->SS = nibDstR;
		bPtr->SF = subDstR;
		bPtr->DD = 0xFFFFFFFF;	/* colour of line */

		bPtr->BC = 	(BCgo + BCfcen + BCeols*0 +
		        	BCdps*PixelSize + BCsps*PixelSize +
				BCdxc*1 + BCdyc*0 + 
				BCsxc*1 + BCsyc*0 );

	}
}


#define PI		3.14159265
#define RAD(ang)	((ang)*(PI/180.0))
#define DEG(ang)	((ang)*(180.0/PI))

static void
Lines( int width )
{
	int ang;
	int xcen = WIDTH/2;
	int ycen = HEIGHT/2;

	InitDrawLines();

/*
	for (xcen = WIDTH/2 - 100; xcen<WIDTH/2 + 100; xcen += 5) {
*/
		for (ang = 0; ang < 360; ang += 10)
		{
			float dx = cos(RAD((float)ang));
			float dy = sin(RAD((float)ang));
			unsigned int x1 = xcen + dx*(WIDTH/10.0),
			     y1 = ycen + dy*(HEIGHT/10.0);
			unsigned int x2 = xcen + dx*(WIDTH/2.5),
			     y2 = ycen + dy*(HEIGHT/2.5);

			DrawLine(x1, y1, x2, y2, width);
		}
/*
	}
*/
}


int
main(int argc, char *argv[])
{
	int width = 1;
	if (argc>1) width = atoi(argv[1]);
	if (width<1) width = 1;

	Wait("About to initialise video");
	bl_init();
	Wait("About to draw lines");
	Lines(width);
}
