/* Initial test program for Blossom: block functions */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <syslib.h>
#include "blossom.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	1024

static int pfill[ 16 ] =
{
	0x00000000, 0x11111111, 0x22222222, 0x33333333,
	0x44444444, 0x55555555, 0x66666666, 0x77777777,
	0x88888888, 0x99999999, 0xAAAAAAAA, 0xBBBBBBBB,
	0xCCCCCCCC, 0xDDDDDDDD, 0xEEEEEEEE, 0xFFFFFFFF
};

#define PFILL(c)	pfill[ (c) & 15 ]

/* low-level routines */

static int pDstBase;

static void
wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

static void
InitBlitter(int alu, int planemask, int bitmask)
{
	WORD	*px;

	px = (WORD *)MC_REG;	/* enable blitter */
	*px = 0x4A;

	px = (WORD *)MB_REG;
	*px = 0xff00;		/* top/bot masks both true */

	px = (WORD *)BF_REG;
	*px = alu;		/* Bit function */

	px = (WORD *)TC_REG;
	*px = 1;		/* FWR = 1. Ignore linear tests */

	px = (WORD *)PM_REG;
	*px = PFILL(planemask);	/* Planar mask */

	px = (WORD *)MD_REG;
	*px = bitmask;		/* Bit mask data */

	px = (WORD *)BC_REG;
	*px = 0x00000000;	/* Reset everything */
}

#define NibblesPerPixel		1
#define PixelSize		0
#define DstStride		(2048*2)

static void
DrawBox(int x, int y, int w, int h, int col)
{
	int DstL, DstR, dw;

	if ( w<=0 || h<=0 ) return;

	DstL = (pDstBase + y*2048)*2 + x;
	DstR = DstL + w - 1;

	dw = ((DstR&0xFF8)-(DstL&0xFF8)) + (DstR & 0x007);

	wait("About to draw");
	{
		WORD	*px;

		px = (WORD *)LS_REG;
		*px++ = DstL;
		*px++ = 0;
		*px++ = 4096;

		px = (WORD *)DW_REG;
		*px++ = dw;
		px = (WORD *)DL_REG;
		*px++ = h - 1;

		px = (WORD *)SD_REG;
		*px = PFILL(col);

		px = (WORD *)BC_REG;
		*px = (BCgo + BCfcen*0 + BCeols*1 + BCdst +
			(BCsps + BCdps + BCmps)*PixelSize + BCmpr*8 +
			BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
		Delay(OneSec/16);
	}
}

static void
Boxes( int alu, int pmsk, int bmsk)
{
	int i;
	int width = 1280/16;
	int height = 960/16;

	pDstBase = MAP1;

	InitBlitter(12, 15, 0xFFFFFFFF);
	for (i = 0 ; i < 16 ; i++)
		DrawBox(0, i*height, 1279, height, i);

	InitBlitter(alu, pmsk, bmsk);
	for (i = 0 ; i < 16 ; i++)
		DrawBox(i*width, 0, width, 1023, i);
}

int
main(int argc, char *argv[])
{
	int alu = 12, pmsk = 0xF, bmsk = 0xFFFFFFFF;

	if (argc>1) alu = atoi(argv[1]);
	if (argc>2) pmsk = atoi(argv[2]);
	if (argc>3) bmsk = atoi(argv[3]);

	Boxes(alu, pmsk, bmsk);
}
