/* Initial test program for Blossom: copy functions */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <syslib.h>
#include "blossom.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	1024

static int pfill[16] = {
	0x00000000, 0x11111111, 0x22222222, 0x33333333,
	0x44444444, 0x55555555, 0x66666666, 0x77777777,
	0x88888888, 0x99999999, 0xAAAAAAAA, 0xBBBBBBBB,
	0xCCCCCCCC, 0xDDDDDDDD, 0xEEEEEEEE, 0xFFFFFFFF
};
#define PFILL(c)	pfill[(c)&15]

/* low-level routines */

static int pDstBase;
static int delay = TRUE, basew, baseh;

static void
wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

static void
InitBlitter(int alu, int planemask, int bitmask)
{
	WORD	*px;

	px = (WORD *)MC_REG;	/* enable blitter */
	*px = 0x4A;

	px = (WORD *)MB_REG;
	*px = 0xff00;		/* top/bot masks both true */

	px = (WORD *)BF_REG;
	*px = alu;		/* Bit function */

	px = (WORD *)TC_REG;
	*px = 1;		/* FWR = 1. Ignore linear tests */

	px = (WORD *)PM_REG;
	*px = PFILL(planemask);	/* Planar mask */

	px = (WORD *)MD_REG;
	*px = bitmask;		/* Bit mask data */

	px = (WORD *)BC_REG;
	*px = 0x00000000;	/* Reset everything */
}

#define NibblesPerPixel		1
#define PixelSize		0
#define DstStride		(2048*2)

static void
DrawBox(int x, int y, int w, int h, int col)
{
	int DstL, DstR, dw;

	if ( w<=0 || h<=0 ) return;

	DstL = (pDstBase + y*2048)*2 + x;
	DstR = DstL + w - 1;

	dw = ((DstR&0xFF8)-(DstL&0xFF8)) + (DstR & 0x007);

	{
		WORD	*px;

		px = (WORD *)LS_REG;
		*px++ = DstL;
		*px++ = 0;
		*px++ = 4096;

		px = (WORD *)DW_REG;
		*px++ = dw;
		px = (WORD *)DL_REG;
		*px++ = h - 1;

		px = (WORD *)SD_REG;
		*px = PFILL(col);

		px = (WORD *)BC_REG;
		*px = (BCgo + BCfcen*0 + BCeols*1 +
			(BCsps + BCdps + BCmps)*PixelSize +
			BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
	}
}

static void
CopyBox(int x1, int y1, int x2, int y2, int w, int h)
{
	int SrcL, SrcR;
	int DstL, DstR, dw, dy;

	if (h<0) dy = -h - 1;
	else	 dy =  h - 1;

	if ( w<=0 ) {
		if ( h<=0 ) {
			SrcR = (pDstBase + (y1+dy)*2048)*2 + x1;
			DstR = (pDstBase + (y2+dy)*2048)*2 + x2;
		} else {
			SrcR = (pDstBase + y1*2048)*2 + x1;
			DstR = (pDstBase + y2*2048)*2 + x2;
		}
		SrcL = SrcR - w - 1;
		DstL = DstR - w - 1;
		dw = (SrcL & 0xFF8) - (SrcR & 0xFF8) + (SrcL & 7);
	} else {
		if ( h<=0 ) {
			SrcL = (pDstBase + (y1+dy)*2048)*2 + x1;
			DstL = (pDstBase + (y2+dy)*2048)*2 + x2;
		} else {
			SrcL = (pDstBase + y1*2048)*2 + x1;
			DstL = (pDstBase + y2*2048)*2 + x2;
		}
		SrcR = SrcL + w - 1;
		DstR = DstL + w - 1;
		dw = (SrcR & 0xFF8) - (SrcL & 0xFF8) + (SrcR & 7);
	}

	printf("Width %d Height %d DstL %x DstR %x dw %x\n",
			w, h, DstL, DstR, dw);

	if (delay) wait("About to copy");
	{
		WORD	*px;
		int bc;

		px = (WORD *)SS_REG;
		*px++ = SrcL;
		*px++ = 0;
		*px++ = 4096;

		px = (WORD *)DS_REG;
		*px++ = DstL;
		*px++ = 0;
		*px++ = 4096;

		px = (WORD *)DW_REG;
		*px++ = dw;
		px = (WORD *)DL_REG;
		*px++ = dy;

		bc = BCgo + BCfcen*0 + BCeols*1 + BCsrc + BCdst +
			(BCsps + BCdps)*PixelSize;
		if (w<0)
			if (h<0)
				bc += BCdxc*3 + BCdyc*2 + BCsxc*3 + BCsyc*2;
			else
				bc += BCdxc*3 + BCdyc*0 + BCsxc*3 + BCsyc*0;
		else
			if (h<0)
				bc += BCdxc*1 + BCdyc*2 + BCsxc*1 + BCsyc*2;
			else
				bc += BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0;

		px = (WORD *)BC_REG;
		*px = bc;
	}
}

static void
Boxes( int alu, int pmsk, int bmsk)
{
	int i;
	int width = 1280/32;
	int height = 960/32;

	pDstBase = MAP1;

	InitBlitter(12, 15, 0xFFFFFFFF);
	for (i = 0 ; i < 16 ; i++) {
		DrawBox(0, i*height, 1280/2, height, i);
		DrawBox(1280/2 + i*width, 960/2 + 0, width, 960/2, i);
	}

	delay = FALSE;
	InitBlitter(alu, pmsk, bmsk);
	for (i = 0 ; i<8; i++) {
		int j;
		for (j=0 ; j<8; j++) {
			int dx = j*64 + i;
			int dy = i*64;
			CopyBox(1280/8 + dx, 960/8 + dy,
				1280/2 + dx, 960/2 + dy,
					basew+j, baseh);
		}
	}
	delay = TRUE;

	InitBlitter(alu, pmsk, bmsk);
	CopyBox(1280/4, 960/4, 1280*2/3, 960*2/3,  1280/4,  960/4);

	wait("About to reset");
	InitBlitter(12, 15, 0xFFFFFFFF);
	for (i = 0 ; i < 16 ; i++) {
		DrawBox(0, i*height, 1280/2, height, i);
		DrawBox(1280/2 + i*width, 960/2 + 0, width, 960/2, i);
	}

	InitBlitter(alu, pmsk, bmsk);
	CopyBox(1280/4, 960/4, 1280*2/3, 960*2/3,  1280/4, -960/4);

	wait("About to reset");
	InitBlitter(12, 15, 0xFFFFFFFF);
	for (i = 0 ; i < 16 ; i++) {
		DrawBox(0, i*height, 1280/2, height, i);
		DrawBox(1280/2 + i*width, 960/2 + 0, width, 960/2, i);
	}

	InitBlitter(alu, pmsk, bmsk);
	CopyBox(1280/4, 960/4, 1280*2/3, 960*2/3,  -1280/4,  960/4);

	wait("About to reset");
	InitBlitter(12, 15, 0xFFFFFFFF);
	for (i = 0 ; i < 16 ; i++) {
		DrawBox(0, i*height, 1280/2, height, i);
		DrawBox(1280/2 + i*width, 960/2 + 0, width, 960/2, i);
	}

	InitBlitter(alu, pmsk, bmsk);
	CopyBox(1280/4, 960/4, 1280*2/3, 960*2/3,  -1280/4,  -960/4);
}

int
main(int argc, char *argv[])
{
	int alu = 12, pmsk = 0xF, bmsk = 0xFFFFFFFF;

	basew = 12;
	baseh = 32;

	if (argc>1) alu = atoi(argv[1]);
	if (argc>2) basew = atoi(argv[2]);
	if (argc>3) baseh = atoi(argv[3]);
/*
	if (argc>2) pmsk = atoi(argv[2]);
	if (argc>3) bmsk = atoi(argv[3]);
*/
	Boxes(alu, pmsk, bmsk);
}
