/* Initial test program for Blossom */

#include <stdio.h>
#include <math.h>
#include "blossom.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	1024

/* low-level routines */

static int pDstBase;

static void
Wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

/**********************************************************
Initialise the graphics registers to default values

Assumes that the blitter has been enabled.
**********************************************************/
static void bl_init()
{
WORD	*px;

px = (WORD *)MC_REG;
*px = 0x4A;

px = (WORD *)MB_REG;
*px = 0xff00;		/* top/bot masks both true */

px = (WORD *)BF_REG;
*px = 5;		/* Bit function = ~destination */

px = (WORD *)TC_REG;
*px = 1;		/* FWR = 1. Ignore linear tests */

px = (WORD *)PM_REG;
*px = 0xffffffff;	/* Planar mask, all bits true */

px = (WORD *)MD_REG;
*px = 0xffffffff;	/* Bit mask data, all true */

px = (WORD *)SD_REG;
*px = 0xFFFFFFFF;	/* source data */

px = (WORD *)DD_REG;
*px = 0xFFFFFFFF;	/* destination data */

px = (WORD *)BC_REG;
*px = 0x00000000;	/* Reset everything */

}

static void
InitBlitter(int op)
{
	WORD *px;

	bl_init();

	px = (WORD *)BF_REG;
	*px = op;		/* Bit function */
}

#define NibblesPerPixel		1
#define PixelSize		0
#define DstStride		(2048*2)

static void
InitDrawLines(void)
{
	pDstBase = MAP1;
}

static void
DrawLine(x1, y1, x2, y2, width)
int x1, y1, x2, y2;
int width;
{
	double tdx, tdy, dx, dy, hyp, proj, x, y;
	double gradL, gradR;
	int nibDstL, subDstL,  nibDstR, subDstR;
	int nibStrideL, subStrideL, nibStrideR, subStrideR;

	if (x1==x2 || y1==y2) return;

	tdx = (double)(x2 - x1);
	tdy = (double)(y2 - y1);
	hyp = sqrt(tdx*tdx + tdy*tdy);
	proj = (((float)width)/2.0)/hyp;

	printf("tdx %f tdy %f hyp %f proj %f\n", tdx, tdy, hyp, proj);

	dx = tdy*proj;
	dy = tdx*proj;

	x = ((float)x1) + dx;
	y = ((float)y1) - dy;

	printf("dx %f dy %f x %f y %f\n", dx, dy, x, y);

/* Section one, top triangle */

	nibDstL = (pDstBase + ((int)y)*2048)*2 + ((int)x);
	subDstL = (int)(x*4096.0);

	gradL = -tdy/tdx;
	if (gradL>0.0) {
		nibStrideL = 4096 + ((int)gradL);
		subStrideL = ((int)(gradL*4096.0)) & 4095;
	} else {
		nibStrideL = 4096 - ((int)(-gradL)) - 1;
		subStrideL = 4096 - (((int)((-gradL)*4096.0)) & 4095);
	}

	nibDstR = nibDstL;
	subDstR = subDstL;

	gradR = tdx/tdy;
	if (gradR>0.0) {
		nibStrideR = 4096 + ((int)gradR);
		subStrideR = ((int)(gradR*4096.0)) & 4095;
	} else {
		nibStrideR = 4096 + ((int)gradR) - 1;
		subStrideR = 4096 - (((int)(gradR*(-4096.0))) & 4095);
	}

	printf("gradR %f * gradL %f = %f\n", gradR, gradL, gradL*gradR);

	printf("Left: %x.%04x %x.%04x Right: %x.%04x %x.%04x Height %d\n",
		nibDstL, subDstL, nibStrideL, subStrideL,
		nibDstR, subDstR, nibStrideR, subStrideR,
		((int)(2.0*dy)) - 1);


	Wait("About to draw");
	{
		WORD	*px;

		px = (WORD *)DL_REG;
		*px = ((int)2.0*dy) - 1;

		px = (WORD *)DS_REG;

		*px++ = nibDstL;
		*px++ = subDstL;
		*px++ = nibStrideL;
		*px++ = subStrideL;

		px = (WORD *)SS_REG;

		*px++ = nibDstR;
		*px++ = subDstR;
		*px++ = nibStrideR;
		*px++ = subStrideR;

		px = (WORD *)BC_REG;
		*px = (BCgo + BCfcen + BCeols*0 + BCdst +
		        BCdps*PixelSize + BCsps*PixelSize
			+ BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
	}

/* Section two, main body */

	printf("height %d", ((int)(tdy - 2.0*dy)));

	Wait("About to draw");
	{
		WORD	*px;

		px = (WORD *)DL_REG;
		*px = ((int)(tdy - 2.0*dy)) - 1;

		px = (WORD *)DT_REG;

		*px++ = nibStrideR;
		*px++ = subStrideR;

		px = (WORD *)BC_REG;
		*px = (BCgo + BCfcen + BCeols*0 + BCdst +
		        BCdps*PixelSize + BCsps*PixelSize
			+ BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
	}
/* Section three, bottom triangle */

	printf("height %d", ((int)(2.0*dy))-1);

	Wait("About to draw");
	{
		WORD	*px;

		px = (WORD *)DL_REG;
		*px = ((int)(2.0*dy)) - 1 - 1;

		px = (WORD *)ST_REG;

		*px++ = nibStrideL;
		*px++ = subStrideL;

		px = (WORD *)BC_REG;
		*px = (BCgo + BCfcen + BCeols*0 + BCdst +
		        BCdps*PixelSize + BCsps*PixelSize
			+ BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
	}
}

#define PI		3.14159265
#define RAD(ang)	((ang)*(PI/180.0))
#define DEG(ang)	((ang)*(180.0/PI))

static void
Lines( int width)
{
	int ang;
	int xcen = WIDTH/2;
	int ycen = HEIGHT/2;

	InitDrawLines();

/*
	for (xcen = WIDTH/2 - 100; xcen<WIDTH/2 + 100; xcen += 5) {
*/
		for (ang = 0; ang < 360; ang += 10)
		{
			float dx = cos(RAD((float)ang));
			float dy = sin(RAD((float)ang));
			unsigned int x1 = xcen + dx*(WIDTH/10.0),
			     y1 = ycen + dy*(HEIGHT/10.0);
			unsigned int x2 = xcen + dx*(WIDTH/2.5),
			     y2 = ycen + dy*(HEIGHT/2.5);

			DrawLine(x1, y1, x2, y2, width);
		}
/*
	}
*/
}

int
main(int argc, char *argv[])
{
	int width = 10, op = 12;
	if (argc>1) width = atoi(argv[1]);
	if (width<1) width = 1;
	if (argc>2) op = atoi(argv[2]);
	if (op<0 || op>15) op = 12;

	InitBlitter(op);
	Lines(width);
}

