/* Initial test program for Blossom */

#include <stdio.h>
#include <math.h>
#include "blossom.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	1024

/* low-level routines */

static int pDstBase;

static void
Wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

/**********************************************************
Initialise the graphics registers to default values

Assumes that the blitter has been enabled.
**********************************************************/
static void bl_init()
{
WORD	*px;

px = (WORD *)MC_REG;
*px = 0x4A;

px = (WORD *)MB_REG;
*px = 0xff00;		/* top/bot masks both true */

px = (WORD *)BF_REG;
*px = 10;		/* Bit function = destination */

px = (WORD *)TC_REG;
*px = 1;		/* FWR = 1. Ignore linear tests */

px = (WORD *)PM_REG;
*px = 0xffffffff;	/* Planar mask, all bits true */

px = (WORD *)MD_REG;
*px = 0xffffffff;	/* Bit mask data, all true */

px = (WORD *)DD_REG;
*px = 0xFFFFFFFF;	/* destination data */

px = (WORD *)BC_REG;
*px = 0x00000000;	/* Reset everything */

}

#ifdef NOTDEF
WORD	bl_line_lines;		/* height */
WORD	bl_line_colour;
WORD	bl_line_psz;		/* pixel size (0,1,2,3) */

/**********************************************************
bl_line

Draws arbitrary lines
**********************************************************/

static void bl_line(starti,startf,stridei,lstridef,rstridei,rstridef,width)
WORD	starti,startf,stridei,lstridef;
WORD	rstridei,rstridef,width;
{
WORD	*px;

px = LS_REG;

*px++ = starti;
*px++ = startf;
*px++ = lstridei;
*px++ = lstridef;

if(width == 0) {
}

px = BC_REG
*px = (BCgo + BCfcen + BCdps*bl_line_psz + BCsps*bl_line_psz
       + BCdxc*1 + BCdyx*0 + BCsxc*1 + BCsyc*0);
}
/* get mode, get magic numbers */
#endif

static void
InitBlitter(void)
{
	bl_init();
}

#define NibblesPerPixel		1
#define PixelSize		0
#define DstStride		(2048*2)

static void
InitDrawLines(void)
{
	pDstBase = MAP1;
}

static void
DrawLine(int x1, int y1, int x2, int y2, int width) 
{
	int nibDstL, subDstL,  nibDstR, subDstR;
	int dx, dy, stride, nibStride, subStride;

	bl_init();

	if (y2==y1) return;

	if (y2<y1) {
		int tmp;
		tmp = x1; x1 = x2; x2 = tmp;
		tmp = y1; y1 = y2; y2 = tmp;
	}

	dx = x2 - x1 + 1;
	dy = y2 - y1 + 1;

	printf("Draw Line (%d,%d)->(%d,%d) dx=%d, dy=%d\n",
					x1, y1, x2, y2, dx, dy);

	nibDstL = (pDstBase + y1*2048)*2 + x1;
	subDstL = 0;

	if (dx>=0) {
		stride = dx*4096 / dy;
		subStride = stride & 4095;
		nibStride = DstStride + stride / 4096;
		stride = (stride+4096)*width;
		nibDstR = nibDstL + stride / 4096;
		subDstR = stride & 4095;
	} else {
		stride = -dx*4096 / dy;
		subStride = 4096 - (stride & 4095);
		nibStride =  DstStride - stride / 4096 - 1;
		stride = stride*width;
		nibDstR = nibDstL + stride / 4096;
		subDstR = stride & 4095;
	}

	nibDstR = nibDstL + width / 4096;
	subDstR = subDstL + ( width &4095);

	printf("L %x.%04x R %x.%04x Stride %x.%04x Height %d\n",
		nibDstL, subDstL, nibDstR, subDstR, nibStride, subStride, dy);

	Wait("About to draw");
	{
		WORD	*px;

		px = (WORD *)DL_REG;
		*px = dy - 1;

		px = (WORD *)LS_REG;

		*px++ = nibDstL;
		*px++ = subDstL;
		*px++ = nibStride;
		*px++ = subStride;

		px = (WORD *)SS_REG;

		*px++ = nibDstR;
		*px++ = subDstR;

		px = (WORD *)BC_REG;
		*px = (BCgo + BCfcen + BCeols*0 +
		        BCdps*PixelSize + BCsps*PixelSize
			+ BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );

		px = (WORD *)T1_REG;
		printf("T1 = %ld\n", ((*px)>>16)&0xFFF);
	}
}

static void
InitFillRectangles(void)
{
}

static void
FillRectangle(int x1, int y1, int width, int height)
{
	x1 = x1;
	y1 = y1;
	width = width;
	height = height;
}

#define PI		3.14159265
#define RAD(ang)	((ang)*(PI/180.0))
#define DEG(ang)	((ang)*(180.0/PI))

static void
Lines( int width )
{
	int ang;
	int xcen = WIDTH/2;
	int ycen = HEIGHT/2;

	InitDrawLines();

/*
	for (xcen = WIDTH/2 - 100; xcen<WIDTH/2 + 100; xcen += 5) {
*/
		for (ang = 0; ang < 360; ang += 10)
		{
			float dx = cos(RAD((float)ang));
			float dy = sin(RAD((float)ang));
			unsigned int x1 = xcen + dx*(WIDTH/10.0),
			     y1 = ycen + dy*(HEIGHT/10.0);
			unsigned int x2 = xcen + dx*(WIDTH/2.5),
			     y2 = ycen + dy*(HEIGHT/2.5);

			DrawLine(x1, y1, x2, y2, width);
		}
/*
	}
*/
}


static void
Boxes( void )
{
	int size;
	int xcen = WIDTH/2;
	int ycen = HEIGHT/2;

	InitFillRectangles();

	for (size = 1 ; size < 9; size++)
	{
		int dx = (WIDTH/2)*size/10;
		int dy = (HEIGHT/2)*size/10;
		int x1 = xcen - dx, y1 = ycen - dy;
		int x2 = xcen + dx, y2 = ycen + dy;

		FillRectangle(x1, y1, x2-x1+1, y2-y1+1);
	}

}

int
main(int argc, char *argv[])
{
	int width = 1;
	if (argc>1) width = atoi(argv[1]);
	if (width<1) width = 1;

	Wait("About to initialise video");
	InitBlitter();
	Wait("About to draw lines");
	Lines(width);
#ifdef notdef
	Wait("About to draw boxes");
	Boxes();
#endif
}
