/************************************************************************/
/*                                                                      */
/* File: setupvid.c		                                        */
/*          								*/
/* Program to run on the Transputer loaded by test.			*/
/************************************************************************/

#include "helios.h"
#include "task.h"
#include "video.h"
#include "blossom.h"
#include <stdio.h>

void fillscr();
char* dummyscr;

WORD v_mode[NUM_MODES] = { 0x02,0x3C,0x36,0x11 }; /* CLUT enables */

WORD crt[8][9] =  { 
/* mode 0 */    { 208,158,160,16,1100,958,1000,4,3} ,		/* Hitachi */
/* mode 1 */	{ 320,254,268,21,797,766,774,4,0x10000000} ,	/* Hitachi */
/* mode 1 */	{ 352,254,284,21,896,766,824,4,3} ,		/* NEC pl */
/* mode 1 */	{ 320,254,268,21,797,766,774,4,0x00000000} ,	/* NEC xl */
/* mode 2 */	{ 210,158,176,20,544,478,500,3,0} ,		/* NEC pl */
/* mode 2 */	{ 210,158,176,20,544,478,500,3,0} ,		/* NEC xl */
/* mode 3 */	{ 800,511,616,88,544,478,500,3,0x00008005} ,	/* NEC pl */
/* mode 3 */	{ 800,511,616,88,544,478,500,3,0x00008005} ,	/* NEC xl */
		 };

setupvreg(index)
WORD index;
{
WORD i; 
WORD *x = 	VREG_BASE;
WORD *ptr1 = 	0x80000800;
WORD *ptr2 = 	0x80000804;

	for (i=0; i <= NUM_VREGS; i++) {
		*x++ = crt[index][i];
	}
}

void setup_video(mode,monitor)
WORD mode;	/* screen mode 0 - 3 */
WORD monitor;	/* monitor supplied 0 - 2 */
{
WORD *px; 	/* pointer for register stuffing */

	if (mode < 0 || mode >= NUM_MODES) return;

	switch (mode) {

		case 0:		setupvreg(MODE0);
				break;

		case 1:	switch (monitor) {

				case HITACHI:	setupvreg(MODE1_HITACHI);
						break;
				case NEC_PL:	setupvreg(MODE1_NEC_PL);
						break;
				case NEC_XL:	setupvreg(MODE1_NEC_XL);
						break;
				default:	/* error */
						break;
			}
			break;

		case 2:	switch (monitor) {

				case NEC_PL:	setupvreg(MODE2_NEC_PL);
						break;
				case NEC_XL:	setupvreg(MODE2_NEC_XL);
						break;
				default:	/* error */
						break;
			}
			break;

		case 3:	switch (monitor) {

				case NEC_PL:	setupvreg(MODE3_NEC_PL);
						break;
				case NEC_XL:	setupvreg(MODE3_NEC_XL);
						break;
				default:	/* error */
						break;
			}
			break;

		default: /*error */

			break;
	}
        px = MC_REG;
	*px = 2;		/* Enable video (Blossom) */
	px = VMODE;
        *px = v_mode[mode];	/* Enable the CLUT */
}

/*************************************************************
	Write colour to BT453 (modes 1,2) CLUT
	Assumes address has been set.
*************************************************************/
void set_colour(red,green,blue)
WORD red,green,blue;
{
WORD *px;
	px =C453WDATA; 
        *px= red;
	px =C453WDATA;
        *px = green;
	px =C453WDATA;
        *px = blue;
}

void set0colour(red,green,blue)
WORD red, green, blue;
{
WORD *px;
    px = 0x42582000;	/* address of BT454 data reg */
    *px = red;
    *px = green;
    *px = blue;
}

void setup_colour()
{
WORD i,bias,*px;

	px =C453WADDR;
        *px = 0L;		/* Set address to  zero */
	set_colour(~0,~0,~0);
	bias = 147;
	for(i=1; i <= 36; i++) {

		set_colour(bias + i*3,0,0);
		set_colour(0,bias + i*3,0);
		set_colour(0,0,bias + i*3);
		set_colour(bias + i*3,bias + i*3,0);
		set_colour(bias + i*3,0,bias + i*3);
		set_colour(0,bias + i*3,bias + i*3);
		set_colour(bias + i*3,bias + i*3,bias + i*3);
	}

	set_colour(~0,0,0);
	set_colour(0,0,~0);

	setup_0colour();
}	

setup_0colour()
{
WORD *px,i;

    px = 0x42580000;
    *px = 0L;
    set0colour(15L,15L,15L);
    for(i=0L;i<3;i++)
	set0colour(i<<2,i<<2,i<<2);
    for(i=0L;i<4;i++)
	set0colour(15L,i<<1,i<<1);
    for(i=0L;i<4;i++)
	set0colour(i<<1,15L,i<<1);
    for(i=0L;i<4;i++)
	set0colour(i<<1,i<<1,15L);

}

void clear_screen()
{
WORD	*i;
WORD	col = 0;
WORD	*start = (WORD *)MAP1;
WORD	*end   = (WORD *)MAP2;

	for(i=start;i<end;*i++=col);	/* Crude as they come! */
}

main(argc,argv)
int argc;
char *argv[];
{
	if (argc<3) {
		printf("No good!");
		return 1;
	}
	setup_video(atoi(argv[1]), atoi(argv[2]));

	setup_0colour();

	clear_screen();
}
