/* Trapezia for Blossom */

#include <stdio.h>
#include <math.h>
#include "/e/x11/server/include/blossom.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	1024

/* low-level routines */

static int pDstBase;

static void
wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

#define NibblesPerPixel		1
#define PixelSize		0
#define DstStride		(2048*2)

static void
FillTrap(int x1, int x2, int y1, int x3, int x4, int y2)
{
	int strideL, strideR;
	int nibDstL, nibDstR;
	int nibStrideL, subStrideL;
	int nibStrideR, subStrideR;
	BlitterPtr bPtr = BLITTER_BASE;
	int dx, dy = y2 - y1;

	RESET_BLITTER(bPtr);

	bPtr->PM = 0xFFFFFFFF;
	bPtr->SD = 0xFFFFFFFF;
	bPtr->BF = 12;

	if (dy<0) return;

	nibDstL = (pDstBase + y1*2048)*2 + x1;
	nibDstR = (pDstBase + y1*2048)*2 + x2;

	dx = x3 - x1;
	if (dx>=0) {
		strideL = dx*4096 / dy;
		nibStrideL = DstStride + strideL/4096;
		subStrideL = strideL&4095;
	} else {
		return;
	}
	dx = x4 - x2;
	if (dx>=0) {
		strideR = dx*4096 / dy;
		nibStrideR = DstStride + strideR/4096;
		subStrideR = strideR&4095;
	} else {
		return;
	}

	{
		int end1, end2, end3;
		end1 = nibDstL*4096 + strideL*dy;
		end2 = nibDstL*4096 + strideL*(dy-1);
		end3 = nibDstR*4096 + strideR*(dy-1);

		end1 = end1/4096;
		end2 = end2/4096;
		end3 = end3/4096;

		printf("end1 %x(%x), end2 %x(%x), end3 %x(%x)\n",
			end1, end1&0xFF8,
			end2, end2&0xFF8,
			end3, end3&0xFF8);
	}

	printf("FillTrap(%x, %x, %x, %x, %x, %x)\n",
		x1, x2, y1, x3, x4, y2);

	printf("L %x R %x SL %x.%03x SR %x.%03x Height %d\n",
		nibDstL, nibDstR, nibStrideL, subStrideL,
			nibStrideR, subStrideR, dy);

	wait("About to draw");
	{
		bPtr->LF = 0;

		bPtr->DS = nibDstL;
		bPtr->DT = nibStrideL;
		bPtr->DI = subStrideL;
		bPtr->SS = nibDstR;
		bPtr->ST = nibStrideR;
		bPtr->SI = subStrideR;
		bPtr->DL = dy;
		bPtr->BC =(BCgo + BCfcen + BCeols*0 +
		        	BCdps*PixelSize + BCsps*PixelSize
				+ BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
	}
}

static void
Traps(int width)
{
	int x1 = 128, x2 = x1 + width - 1;
	int y1 = 128, y2 = y1 + 1;
	int x3;

	for (x1=128 ; x1<199; x1++)
		FillTrap(x1, x1+1, y1, x1+1, x1+2, y2);
}

int
main(int argc, char *argv[])
{
	int width = 2;
	BlossomPtr blPtr = BLOSSOM_BASE;

/*
	if (argc<7) return -1;
*/
	if (argc>1) width = atoi(argv[1]);

	blPtr->MC = 0x4A;

	pDstBase = MAP1;

	Traps(width);
}
