/************************************************************************/
/*                                                                      */
/* File: vid.c		                                                */
/*          								*/
/* Program to run on the Transputer loaded by test.			*/
/* Tests the video system using the Charity chip			*/
/*                                                                      */
/* Authors:	RGM  Richard Miller, Perihelion				*/
/*		JPD  Jem Davies, Cambridge Systems Design		*/
/*									*/
/* Changes:								*/
/* 26/2/88  JPD	Add stripe drawing					*/
/* 25/2/88  RGM add mode 3 colour demo					*/
/* 25/2/88  JPD pick up mode and border width from a poke 		*/
/* 30/12/87 RGM Debugging for Boxes version				*/
/*                                                                      */
/* Copyright (c) 1988, Perihelion Ltd. All Rights Reserved.             */
/************************************************************************/

#include "helios.h"
#include "task.h"
#include "video.h"
#include "charity.h"

void fillscr();
char* dummyscr;

WORD v_mode[NUM_MODES] = { 0x02,0x3C,0x36,0x11 }; /* CLUT enables */

WORD crt[8][9] =  { 
/* mode 0 */    { 208,158,160,16,1100,958,1000,4,3} ,		/* Hitachi */
/* mode 1 */	{ 320,254,268,21,797,766,774,4,0x10000000} ,	/* Hitachi */
/* mode 1 */	{ 352,254,284,21,896,766,824,4,3} ,		/* NEC pl */
/* mode 1 */	{ 320,254,268,21,797,766,774,4,0x00000000} ,	/* NEC xl */
/* mode 2 */	{ 210,158,176,20,544,478,500,3,0} ,		/* NEC pl */
/* mode 2 */	{ 210,158,176,20,544,478,500,3,0} ,		/* NEC xl */
/* mode 3 */	{ 800,511,616,88,544,478,500,3,0x00008005} ,	/* NEC pl */
/* mode 3 */	{ 800,511,616,88,544,478,500,3,0x00008005} ,	/* NEC xl */
		 };

setupvreg(index)
WORD index;
{
WORD i; 
WORD *x = 	VREG_BASE;
WORD *ptr1 = 	0x80000800;
WORD *ptr2 = 	0x80000804;

	for (i=0; i <= NUM_VREGS; i++) {

/*		*x++  =  crt[index][i] ; */
		if (*ptr1 == 0)
			*ptr2  =  crt[index][i] ;
		*x++ = *ptr2++;
	}
}

void setup_video(mode,monitor)
WORD mode;	/* screen mode 0 - 3 */
WORD monitor;	/* monitor supplied 0 - 2 */
{
WORD *px; 	/* pointer for register stuffing */

	if (mode < 0 || mode >= NUM_MODES) return;

	switch (mode) {

		case 0:		setupvreg(MODE0);
				break;

		case 1:	switch (monitor) {

				case HITACHI:	setupvreg(MODE1_HITACHI);
						break;
				case NEC_PL:	setupvreg(MODE1_NEC_PL);
						break;
				case NEC_XL:	setupvreg(MODE1_NEC_XL);
						break;
				default:	/* error */
						break;
			}
			break;

		case 2:	switch (monitor) {

				case NEC_PL:	setupvreg(MODE2_NEC_PL);
						break;
				case NEC_XL:	setupvreg(MODE2_NEC_XL);
						break;
				default:	/* error */
						break;
			}
			break;

		case 3:	switch (monitor) {

				case NEC_PL:	setupvreg(MODE3_NEC_PL);
						break;
				case NEC_XL:	setupvreg(MODE3_NEC_XL);
						break;
				default:	/* error */
						break;
			}
			break;

		default: /*error */

			break;
	}
        px = MC_REG;
	*px = 2;		/* Enable video (Blossom) */
	px = VMODE;
        *px = v_mode[mode];	/* Enable the CLUT */
}

/*************************************************************
	Write colour to BT453 (modes 1,2) CLUT
	Assumes address has been set.
*************************************************************/
void set_colour(red,green,blue)
WORD red,green,blue;
{
WORD *px;
	px =C453WDATA; 
        *px= red;
	px =C453WDATA;
        *px = green;
	px =C453WDATA;
        *px = blue;
}

void set0colour(red,green,blue)
WORD red, green, blue;
{
WORD *px;
    px = 0x42582000;	/* address of BT454 data reg */
    *px = red;
    *px = green;
    *px = blue;
}

void setup_colour()
{
WORD i,bias,*px;

	px =C453WADDR;
        *px = 0L;		/* Set address to  zero */
	set_colour(~0,~0,~0);
	bias = 147;
	for(i=1; i <= 36; i++) {

		set_colour(bias + i*3,0,0);
		set_colour(0,bias + i*3,0);
		set_colour(0,0,bias + i*3);
		set_colour(bias + i*3,bias + i*3,0);
		set_colour(bias + i*3,0,bias + i*3);
		set_colour(0,bias + i*3,bias + i*3);
		set_colour(bias + i*3,bias + i*3,bias + i*3);
	}

	set_colour(~0,0,0);
	set_colour(0,0,~0);

	setup_0colour();
}	

setup_0colour()
{
WORD *px,i;

    px = 0x42580000;
    *px = 0L;
    set0colour(15L,15L,15L);
    for(i=0L;i<3;i++)
	set0colour(i<<2,i<<2,i<<2);
    for(i=0L;i<4;i++)
	set0colour(15L,i<<1,i<<1);
    for(i=0L;i<4;i++)
	set0colour(i<<1,15L,i<<1);
    for(i=0L;i<4;i++)
	set0colour(i<<1,i<<1,15L);

}

void clear_screen()
{
WORD	*i;
WORD	col = 0;
WORD	*start = (WORD *)MAP1;
WORD	*end   = (WORD *)MAP2;

	for(i=start;i<end;i++)	/* Crude as they come! */

		*i=col;
}


void whiteout()
{
	clear_screen();
}

void fillscr()
{
WORD	j,*i,*q,*e;	/* Pointers for screen memory */

WORD	col = 0;	/* Word to use as Fill */
WORD	lineinc = 512;	/* The Word Modulo for next line, ie: 2048 bytes */
WORD	linelen = 160;  /* Number of Words on a line, ie: 960 pixels */

WORD	*start = (WORD *)MAP1;		/* Start of MAP1 screen memory */
WORD	*end   = (WORD *)MAP1 + 512000;	/* End of MAP1 screen memory */

	for(j=0;j<=5;j++)
	{

		for(i=start;i<end;i=i+lineinc)	/* Quick, efficient(ish) cls! */
		{
			e = i+linelen;
			for(q=i;q<=e;)
			{
			*q++=col;	/* 1 */
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;	/* 8 */
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;
			*q++=col;	/* 16 */

			}
		}

		col=col+0x011111111;
	}


}



void draw_border(map,stride,width,height,bpp)
/* Draws a border for the mode */
WORD map,stride,width,height,bpp;
{
BYTE *ptr1,*ptr2;
BYTE col;
WORD i,j,border;

WORD *ptr3=0x80000908;
if (*ptr3) border=*ptr3;
   else border = 2;		/* Set border width in pixels */

col = 0xaa;

/* Draw top and bottom */

	ptr1 = map;   /* top left */
	ptr2 = map + (height-border)*stride; /* left edge, border pixels from bottom */
	for(i=0;i<width;i++)
	{
	    for(j=0;j<border;j++)
	    {
		*(ptr1+(j*stride)) = col;
		*(ptr2+(j*stride)) = col;
	    }
	    ptr1+=bpp;
	    ptr2+=bpp;
	}

/* Now draw sides */

	ptr1 = map;
	ptr2 = map + (width - border)*bpp;
	for (i=0;i<height;i++) 
	{
	   for (j=0;j<border;j++)
	      {
		*(ptr1+(j*bpp)) = col;
		*(ptr2+(j*bpp)) = col;
	      }
	   ptr1 += stride;
	   ptr2 += stride;
	}
}

WORD mode3colours[]={
0,
0x8,0x10,0x20,0x40,0x80,   /* seems to be blue */
0x800,0x1000,0x2000,0x4000,0x8000,  /* green */
0x080000,0x100000,0x200000,0x400000,0x800000,  /* red */
0x080808,0x101010,0x202020,0x404040,0x808080,  /* so.. grey to white */
0x000808,0x001010,0x002020,0x004040,0x008080,  /* B & G: cyan */
0x080008,0x100010,0x200020,0x400040,0x800080,  /* R & B: purple */
0x080800,0x101000,0x202000,0x404000,0x808000}; /* R & G: */

void draw_hstripes(mode,map,stride,width,height,bpp)
WORD mode,map,stride,width,height,bpp;
{

BYTE *ptr;
BYTE col;
WORD i,band,line,n;

band = 16;
col = 1;

for (line=0;line<height;line += band)
 {
    ptr=map+(line*stride);
    for(i=0;i<width;i++)
     {
	for (n=0;n<(band/2);n++)
	 {
	    if(mode!=3)
	        *((BYTE *)(ptr+(i*bpp)+(stride*n))) = col;
	    else
		*(WORD *)((ptr+(i*bpp)+(stride*n))) = mode3colours[col];
	 }
     }
    col++;
 }
}


void draw_vstripes(mode,map,stride,width,height,bpp)
WORD mode,map,stride,width,height,bpp;
{

BYTE *ptr;
BYTE col;
WORD i,band,line,n;

band = 16;
col = 1;

for (line=0;line<width;line += band)
 {
    ptr=map+(line*bpp);
    for(i=0;i<height;i++)
     {
	for (n=0;n<(band/2);n++)
	 {
	    if(mode!=3)
	        *((BYTE *)(ptr+(i*stride)+(bpp*n))) = col;
	    else
		*(WORD *)((ptr+(i*stride)+(bpp*n))) = mode3colours[col];
	 }
     }
    col++;
 }
}


/*******************************************************************
blittest. A simple test of the blitter.
*******************************************************************/

void blittest()
{
ccls();
}
/*
{
cqinit();
while (1) {
   cqblit(MAP1,MAP1,0,(600*2048),2048,2048,100,1024,0);
   cqblit(MAP1,MAP1,(600*2048),0,2048,2048,100,1024,0);
   }
}
*/
/*******************************************************************
cinit.  Initialise the Charity blitter.
*******************************************************************/
cqinit()
{
*((WORD *)(MC_REG)) = 0x02;	/* Enable video, reset blitter */
*((WORD *)(BC_REG)) = 0x1f;	/* Blitter control */
*((WORD *)(MC_REG)) = 0x4a;	/* Enable video, blitter */
}

/*******************************************************************
Blitter copy routine. Should replace Bytblt, except when using
internal RAM!!! Assumes forwards copy.
Must call cinit first. 
*******************************************************************/
cqblit(src,dst,soff,doff,sstride,dstride,dheight,dwidth,opcode)

WORD src,dst,soff,doff,sstride,dstride,dheight,dwidth,opcode;
{
WORD ca = 0;
WORD stuff;

*((WORD *)(TC_REG)) = 0x100000;		/* Force writes */
*((WORD *)(PA_REG)) = 0x00ffff;		/* No masks, no shift */

if ((WORD)dst & 0x80000000)
   ca = 0x02000;
if ((WORD)src & 0x80000000)
   ca += 0x1000;

stuff = (ca + 0x0a28);			/* +1 +stride */
   *((WORD *)(CA_REG)) = stuff;
   printf("Count alignment %lx \n",stuff);

stuff = ((dheight*dwidth)>>2);		/* Dest total */
   *((WORD *)(DT_REG)) = stuff;
   printf("Destination total %lx \n",stuff);

stuff = (dwidth>>2)-2;			/* Dest width */
   *((WORD *)(DW_REG)) = stuff;
   printf("Destination width %lx \n",stuff);

stuff = (((dstride-dwidth)>>2)+1);	/* Dest stride */
   *((WORD *)(DS_REG)) = stuff;
   printf("Destination stride %lx \n",stuff);

stuff = (((sstride-dwidth)>>2));	/* Source stride */
   *((WORD *)(SS_REG)) = stuff;
   printf("Source stride %lx \n",stuff);

*((WORD *)(PC_REG)) = 0xff;			/* Pipe control */
*((WORD *)(TV_REG)) = 0x00ffff00;		/* Tests. Always passed */

stuff = src+soff;			/* Source base */
   *((WORD *)(SB_REG)) = stuff;
   printf("Source base address %lx \n",stuff);

stuff = dst + doff;			/* Dest base address */
   *((WORD *)(DB_REG)) = stuff;		
   printf("Destination base address %lx \n",stuff);

*((WORD *)(BC_REG)) = 0x19;			/* Read source, QS0 */
}


void colours(void)
{
	WORD row, col, red, blue, green;
	WORD *otherScreen = (WORD*)0x80100000;
	WORD *screen = otherScreen;

	setup_video(3,NEC_PL);

		for (col=0; col<512; col++)	
   	 	   for (row=0; row<512; row++)
			*screen++ = ((row/2)<<16) + ((col/2)<<8);

		screen = otherScreen;
		bytblt((WORD*)screen,(WORD*)MAP3,0,0,2048,4096,512,2048,0);

		for(blue = 16; blue < 256; blue += 16) {

		  bytblt((WORD*)&blue,(WORD*)screen,0,0,0,4,512 * 512,1,0);
		  bytblt((WORD*)screen,(WORD*)MAP3,0,0,2048,4096,512,2048,0);
		}

		for (col=0; col<512; col++)	
			for (row=0; row<512; row++)
				*screen++ = ((row/2)<<16) + col/2;

		screen = otherScreen;
		bytblt((WORD*)screen,(WORD*)MAP3,0,0,2048,4096,512,2048,0);

		for(green = 16; green < 256; green += 16) {

		  bytblt((WORD*)&green,(WORD*)screen,0,1,0,4,512 * 512,1,0);
		  bytblt((WORD*)screen,(WORD*)MAP3,0,0,2048,4096,512,2048,0);
		}

		for (col=0; col<512; col++)	
			for (row=0; row<512; row++)
				*screen++ = ((row/2)<<8) + col/2;

		screen = otherScreen;
		bytblt((WORD*)screen,(WORD*)MAP3,0,0,2048,4096,512,2048,0);

		for(red = 16; red < 256; red += 16) {
		  bytblt((WORD*)&red,(WORD*)screen,0,2,0,4,512 * 512,1,0);
		  bytblt((WORD*)screen,(WORD*)MAP3,0,0,2048,4096,512,2048,0);
		}

}

/*****************************************************************
ccls()

Clear screen using PBS mode.
*****************************************************************/

WORD	WR_CCLS[NUM_BREGS] =

	{1,1,1,1,
	 1,1,1,1,
	 0,0,0,0,
	 1,0,1,1};

#define	WID_CCLS	(256-1)	/* Mode 1 screen width-1 (border) */
#define	HIT_CCLS	(768-4)	/* Mode 1 height-4 (border)	*/
				/* Note that height must be in
				   four line units. 		*/
WORD	VALS_CCLS[NUM_BREGS] =
	{0x10,		/* BC Reset, no reads		*/
	0x100000,	/* TC Force write cycles (not req?)	*/
	0xfffff,	/* PA Start, end, mid masks all true	*/
	0x0a28,		/* CA +x, +stride		*/

	WID_CCLS*HIT_CCLS/4,	/* Remember, PBS does 4 lines
					   each cycle   */
			/* DT (1M/4)/4 = 64K writes	*/
	MAP1,		/* DB = MAP1			*/
	WID_CCLS-2,	/* DW				*/
	0x800-WID_CCLS+1,
			/* DS = 8k(/4)			*/

	0,		/* SB Not used			*/
	0,		/* SS Not used			*/
	0,		/* ML Not implemented		*/
	0,		/* MB Not implemented		*/

	0x00,		/* SP Straight through		*/
	0,		/* CD Not implemented		*/
	0x55555555,	/* TV colour data		*/
	0xffffffff,	/* SD =stipple in PBS mode	*/
	};

ccls()
{
*((WORD *)(MC_REG)) = 0x02;	/* Enable video, reset blitter */
*((WORD *)(BC_REG)) = 0x1f;	/* Blitter control */
*((WORD *)(MC_REG)) = 0x4a;	/* Enable video, blitter */

blit(WR_CCLS,VALS_CCLS);
}


blit(wr_flags,vals)

WORD	*wr_flags[];
WORD	*vals[];
{

WORD	*creg = BC_REG;		/* Point to base of registers */
WORD	i;

for (i=NUM_BREGS-1;i>=0;i--) {	/* Write to BC_REG last */
   printf("Register %x data %x flag %x\n",creg,vals[i],wr_flags[i]);
   if (wr_flags[i])		/* If write flag */
      creg[i] = vals[i];	/* Write to that register */
   }
}
