/************************************************************************/
/*                                                                      */
/* File: xtest.c	                                                */
/*          								*/
/* Program to run on the Transputer loaded by test.			*/
/* Hardware test routines used for debugging the ABAQ (FAB1)		*/
/*                                                                      */
/* Authors:	RGM  Richard Miller, Perihelion				*/
/*		JPD  Jem Davies, Cambridge Systems Design		*/
/*									*/
/* Changes:								*/
/* 26/2/88  JPD	Add stripe drawing, set up video variables from mode	*/
/*		peeked from memory (set by test prog running on ST)	*/
/* 30/12/87 RGM								*/
/*                                                                      */
/* Copyright (c) 1988, Perihelion Ltd. All Rights Reserved.             */
/************************************************************************/

#include "helios.h"
#include "task.h"
#include "charity.h"
#include "video.h"


/*

	Bit	If set:
--------------------------------------------
	0	Enable video
 	1	main ram test
	2	video ram test
	3	video ram loop
	4	main ram loop,
	5	blits around memory
	6	Blitter test
	7	mode 3 colours demo
	8	Border on
	9	Vertical and Horizontal stripes
	10	short RAM test for Farm node (disables tests 0-9)
	11	whiteout in mode0
*/

int main(void)
{
WORD *ptr = 0x80000900;
WORD *ptr2 = 0x80000904;
WORD TEST;
WORD mode=*ptr2;
WORD map,stride,width,height,bpp;
TEST = *ptr;

farmreset();
printf("\n\nExternal (farm) nodes reset.\n");

printf("\n\nReading TEST no from $80000900,\nvideo mode from $80000904,\n");
printf("border width from $80000908\n\n");

switch(mode)
{
case 0:
    stride = 2048;
    width = 1280;
    height = 960;
    bpp = 1;		/* bytes per pixel (space taken by one pixel) */
    break;
case 1:
    stride = 2048;
    width = 1024;
    height = 768;
    bpp = 1;		/* bytes per pixel (space taken by one pixel) */
    break;
case 2:
    stride = 2048;
    width = 640;
    height = 480;
    bpp = 1;		/* bytes per pixel (space taken by one pixel) */
    break;
case 3:
    stride = 4096;
    width = 512;
    height = 480;
    bpp = 4;		/* bytes per pixel (space taken by one pixel) */
    break;
}
map=MAP0+(0x200000L*mode);


   if (TEST & 0x800)
        TEST &= ~0x7fe;

   if (TEST & 0x400)
	TEST &= ~0x3ff;	/* disable other tests if bit 10 set */

   while (1) {

	if (TEST&1){
	   setup_video(mode,NEC_PL);
	   setup_colour();
	   clear_screen();
	   if (TEST&256)
	      draw_border(map,stride,width,height,bpp);
	   if (TEST&512) {
	      draw_vstripes(mode,map,stride,width,height,bpp);
	      clear_screen();
	      draw_hstripes(mode,map,stride,width,height,bpp);}
	   }

	   if (TEST&2){
	      testmain(0x80040000L,0x80400000);
	      printf("Main Ram test done. \n");}

	   if (TEST&4){
	      testvid();
	      printf("Video Ram test done.\n");}

	   if (TEST&8) {
		vloop();
		printf("Special Video Ram test done.\n");}

	   if (TEST&16) {
		readmain();
		printf("Read main Ram\n");}

	   if (TEST&32) {
 		blits();
		printf("blits \n");}

	   if (TEST&64) {
		printf("Starting BLIT test.\n");
		blittest();
		printf("Finished BLIT test\n");}

	   if (TEST&128) {
		printf("Starting mode 3 colours demo.\n");
		colours();
		printf("Finished mode 3 colours demo.\n");}

	   if (TEST&1024)
		{
		    printf("\n256K to 1M RAM test:\n");
		    testmain(0x80040000L,0x80100000L);
		    printf("Short (farm node) RAM test finished\n");
		}

	   if (TEST&2048)
		whiteout();

	}
	return(1);
}


/******************************************************************************
   farmreset() - resets up to 32 farm nodes using patent reset/analyse scheme.
******************************************************************************/

farmreset()
{

    clk(32,1);		/* initial 32 pulses with slow high = address all */
    clk(1,0);		/* 1 pulse on fast with slow set low */
    clk(3,1);		/* 3 pulses with slow high = activate node Reset line */
    clk(5,0);		/* slow goes low again */

    clk(32,1);		/* address all nodes */
    clk(1,0);		
    clk(1,1);		/* release node Reset line fn. */
    clk(5,0);
}


/*****************************************************************************/

clk(num,slow)
int num, slow;
{
    int x;
    int fast = 2;
    int MCvalue = 0x00000000L;
    WORD *MCreg = 0x40000000L;

    slow = ~slow & 1;		/* invert slow in bit 0 */
    for(x=num; x>0; x--)
	{
	    *MCreg = MCvalue + slow;	/* set/reset slow as requested */

	    *MCreg = MCvalue + slow + fast;
	    *MCreg = MCvalue + slow; 		/* pulse fast */
	}
}


/*****************************************************************************/
