/* "zero-width" lines for Blossom */

#include <stdio.h>
#include <math.h>
#include "/e/x11/server/include/blossom.h"
#include "video.h"
#include "helios.h"

#define WIDTH	1280
#define HEIGHT	 960
#define WIDTH2	(WIDTH/2)
#define HEIGHT2	(HEIGHT/2)

/* low-level routines */

static int pDstBase;
static int total=0, fails=0;
static int delay=TRUE, quiet=FALSE;
static int firstang = 0;

static void
wait(char *s)
{
	if (s) printf("%s: ", s);
	printf("Press return to continue\n");
	while (getchar()!='\n');
}

#define NPP	2
#define PixelSize		0
#define DstStride		(2048*2)

#define ONE			4096
#define HALF			2048
#define MASK			(ONE-1)

static void
DrawLine(int x1, int y1, int x2, int y2)
{
	int L, R, stride;
	int nibDstL, subDstL,  nibDstR, subDstR;
	int dx, dy, nibStride, subStride;
	BlitterPtr bPtr = BLITTER_BASE;
	int dstStride = 4096;

	RESET_BLITTER(bPtr);

	bPtr->PM = 0xFFFFFFFF;
	bPtr->SD = 0xFFFFFFFF;
	bPtr->BF = 5;

	if (y2<y1) {
		int tmp;
		tmp = x1; x1 = x2; x2 = tmp;
		tmp = y1; y1 = y2; y2 = tmp;
	}

	dx = x2 - x1;
	dy = y2 - y1;

	if (!quiet) printf("Draw Line (%d,%d)->(%d,%d) dx=%d, dy=%d\n",
					x1, y1, x2, y2, dx, dy);

	if (dx>=0) {
		L = x1*ONE + HALF;
		if (dx>dy) {
			stride = (dx+1)*ONE / (dy+1);
			R = L + stride - ONE;
		} else {
			stride = dx*ONE / dy;
			R = L;
		}
		nibStride = dstStride + (stride/ONE)*NPP;
		subStride = stride&MASK;
	} else {
		R = x1*ONE + HALF;
		if (-dx>dy) {
			stride = (-dx+1)*ONE / (dy+1);
			L = R - stride + ONE;
		} else {
			stride = -dx*ONE / dy;
			L = R;
		}
		nibStride = dstStride - (stride/ONE)*NPP - 1;
		subStride = MASK - (stride&MASK);
	}
	nibDstL = pDstBase*2 + y1*dstStride + (L/ONE)*NPP;
	subDstL = L&MASK;
	nibDstR = pDstBase*2 + y1*dstStride + (R/ONE)*NPP;
	subDstR = R&MASK;

	if (!quiet) printf("L %x.%03x R %x.%03x Stride %x.%03x Height %d\n",
		nibDstL, subDstL, nibDstR, subDstR, nibStride, subStride, dy);

	bPtr->LT = nibStride;
	bPtr->LI = subStride;

	bPtr->DS = nibDstL;
	bPtr->DF = subDstL;
	bPtr->SS = nibDstR;
	bPtr->SF = subDstR;
	bPtr->DL = dy;

	if (delay) wait("About to draw");
	bPtr->BC = BCgo + BCfcen + BCeols*0 +
	        	BCdps*1 + BCsps*1 +
			BCdst +
			BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0;
}

static void
NewDrawLine(int x1, int y1, int x2, int y2)
{
	int L, R, stride;
	int nibDstL, subDstL,  nibDstR, subDstR;
	int dx, dy, adx, ady, nibStride, subStride;
	BlitterPtr bPtr = BLITTER_BASE;
	int bc;
	int dstStride = 4096*2;

	RESET_BLITTER(bPtr);

	bPtr->PM = 0xFFFFFFFF;
	bPtr->SD = 0xFFFFFFFF;
	bPtr->BF = 5;

	dx = x2 - x1;
	dy = y2 - y1;
	adx = (dx<0) ? -dx:dx;
	ady = (dy<0) ? -dy:dy;

	if (!quiet) printf("Draw Line (%d,%d)->(%d,%d) dx=%d, dy=%d\n",
					x1, y1, x2, y2, dx, dy);

	if ( dx>=0 ) {
		if ( dy>=0 ) {
			L = x1*ONE + HALF;
			if (adx>ady) {
				stride = (adx+1)*ONE / (ady+1);
				R = L + stride - ONE;
			} else {
				stride = adx*ONE / ady;
				R = L;
			}
			nibStride = dstStride + stride/ONE;
			subStride = stride&MASK;
		} else {
			R = x1*ONE + HALF;
			if (adx>ady) {
				stride = (adx+1)*ONE / (ady+1);
				L = R + stride - ONE;
			} else {
				stride = adx*ONE / ady;
				L = R;
			}
			nibStride = dstStride + stride/ONE;
			subStride = stride&MASK;
		}
	} else {
		if ( dy>=0 ) {
			R = x1*ONE + HALF;
			if (adx>ady) {
				stride = (adx+1)*ONE / (ady+1);
				L = R - stride + ONE;
			} else {
				stride = adx*ONE / ady;
				L = R;
			}
			nibStride = dstStride - stride/ONE + 1;
			subStride = MASK - (stride&MASK);
		} else {
			L = x1*ONE + HALF;
			if (adx>ady) {
				stride = (adx+1)*ONE / (ady+1);
				R = L - stride + ONE;
			} else {
				stride = adx*ONE / ady;
				R = L;
			}
			nibStride = dstStride + stride/ONE;
			subStride = stride&MASK;
		}
	}
	nibDstL = pDstBase*2 + y1*dstStride + L/ONE;
	subDstL = L&MASK;
	nibDstR = pDstBase*2 + y1*dstStride + R/ONE;
	subDstR = R&MASK;
	if (!quiet) printf("L %x.%03x R %x.%03x Stride %x.%03x Height %d\n",
		nibDstL, subDstL, nibDstR, subDstR, nibStride, subStride, ady);

	if (delay) wait("About to draw");
	bPtr->LT = nibStride;
	bPtr->LI = subStride;

	bPtr->DS = nibDstL;
	bPtr->DF = subDstL;
	bPtr->SS = nibDstR;
	bPtr->SF = subDstR;
	bPtr->DL = ady;

	bc = BCgo + BCfcen + BCeols*0 +
        	BCdps*PixelSize + BCsps*PixelSize +
		BCdst;
	if (dy>=0)
		bc += BCdyc*0 + BCsyc*0 + BCdxc*1 + BCsxc*1;
	else
		bc += BCdyc*2 + BCsyc*2 + BCdxc*3 + BCsxc*3;

	bPtr->BC = bc;
	total++;
	if (total%1000 == 0) {
		printf("%d\r", total);
		fflush(stdout);
	}
}

static void
OldDrawLine(int x1, int y1, int x2, int y2)
{
	int L, R, stride;
	int nibDstL, subDstL,  nibDstR, subDstR;
	int dx, dy, nibStride, subStride;
	BlitterPtr bPtr = BLITTER_BASE;
		int bc;
	int trouble = FALSE;

	RESET_BLITTER(bPtr);

	bPtr->PM = 0xFFFFFFFF;
	bPtr->SD = 0xFFFFFFFF;
	bPtr->BF = 5;

	if (y2==y1) return;

	if (y2<y1) {
		int tmp;
		tmp = x1; x1 = x2; x2 = tmp;
		tmp = y1; y1 = y2; y2 = tmp;
	}

	dx = x2 - x1;
	dy = y2 - y1;

	printf("Draw Line (%d,%d)->(%d,%d) dx=%d, dy=%d\n",
					x1, y1, x2, y2, dx, dy);

	L = x1*ONE + HALF;

	if (dx>=0) {
	 	int end0, end1, end2, end3;
		stride = (dx+1)*ONE / (dy+1);
		if (dx>dy) {
			R = L + stride - ONE;
		} else {
			R = L;
		}
		end0 = L + stride*dy;
		end1 = R + stride*dy;
		end2 = L + stride*(dy-1);
		end3 = R + stride*(dy-1);
		end0 = end0/ONE;
		end1 = end1/ONE;
		end2 = end2/ONE;
		end3 = end3/ONE;
		if (dy>0 && (end0&0xFF8)==(end1&0xFF8) && (end0&0xFF8)==(end3&0xFF8) &&
		    (end2&0xFF8)==((end3-8)&0xFF8))
			trouble = TRUE;
		if (trouble) {
			printf("end0 %xend1 %x end2 %x end3 %x\n",
				end0, end1, end2, end3);
		}
trouble = FALSE;
		nibDstL = (pDstBase + y1*2048)*2 + L/ONE;
		subDstL = L&MASK;
		nibDstR = (pDstBase + y1*2048)*2 + R/ONE;
		subDstR = R&MASK;
		nibStride = DstStride + stride/ONE;
		subStride = stride&MASK;
	} else {
		return;
	}

	printf("L %x.%03x R %x.%03x Stride %x.%03x Height %d\n",
		nibDstL, subDstL, nibDstR, subDstR, nibStride, subStride, dy);

	wait("About to draw");
	bPtr->LT = nibStride;
	bPtr->LI = subStride;

	bPtr->DS = nibDstL;
	bPtr->DF = subDstL;
	bPtr->SS = nibDstR;
	bPtr->SF = subDstR;

	bc =(BCgo + BCfcen + BCeols*0 +
        	BCdps*PixelSize + BCsps*PixelSize
		+ BCdst
		+ BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 );
/*
	if (trouble) {
		bPtr->DL = dy-1;
		bPtr->BC = bc;
		bPtr->DL = 1;
		bPtr->BC = bc;
		fails++;
	} else {
*/
		bPtr->DL = dy;
		bPtr->BC = bc;
/*
	}
*/
	total++;
	if (total%1000 == 0) {
		fprintf(stderr, "%d %d %d\r", total, fails, fails*1000/total);
		fflush(stderr);
	}
}

#define PI		3.14159265
#define RAD(ang)	((ang)*(PI/180.0))
#define DEG(ang)	((ang)*(180.0/PI))

static void
CircleLines()
{
	int ang;
	int xcen = WIDTH/2 - 100;
	int dxcen = 3;
	int ycen = HEIGHT/2;

	for (;;) {
		for (ang = firstang; ang < 360; ang += 1)
		{
			float dx = cos(RAD((float)ang));
			float dy = sin(RAD((float)ang));
			unsigned int x1 = xcen + dx*(WIDTH/10.0),
			     y1 = ycen + dy*(HEIGHT/10.0);
			unsigned int x2 = xcen + dx*(WIDTH/2.5),
			     y2 = ycen + dy*(HEIGHT/2.5);

			DrawLine(x1, y1, x2, y2);
		}
		xcen += dxcen;
		if (xcen>WIDTH/2+100 || xcen<WIDTH/2-100) dxcen = -dxcen;
	}
}

static void
SquareLines()
{
	int i;
	int w = WIDTH, h = HEIGHT;

	for (;;) {
		for (i=0; i<w; i++)	DrawLine(  0,   0,   i, h-1);
		for (i=h-1; i>=0; i--)	DrawLine(  0,   0, w-1,   i);
		for (i=0 ; i<h; i++)	DrawLine(w-1,   0,   0,   i);
		for (i=0; i<w; i++)	DrawLine(w-1,   0,   i, h-1);
		for (i=w-1; i>=0; i--)	DrawLine(w-1, h-1,   i,   0);
		for (i=0 ; i<h; i++)	DrawLine(w-1, h-1,   0,   i);
		for (i=h-1; i>=0; i--)	DrawLine(  0, h-1, w-1,   i);
		for (i=w-1; i>=0 ;i--)	DrawLine(  0, h-1,   i,   0);
	}
}

static void
RandomLines( )
{
	for (;;)
		DrawLine(rand()%WIDTH, rand()%HEIGHT,
			 rand()%WIDTH, rand()%HEIGHT);
}

int
main(int argc, char *argv[])
{
	int random = FALSE;
	int square = FALSE;
	BlossomPtr blPtr = BLOSSOM_BASE;

	blPtr->MC = 0x4A;
	pDstBase = MAP0;

	if (argc>1) {
		int i, c;
		for (i=0;(c=argv[1][i]) != 0; i++) {
			switch (c) {
			case 'q':
				quiet = TRUE;
				break;
			case 'n':
				delay = FALSE;
				break;
			case 'r':
				random = TRUE;
				break;
			case 's':
				square = TRUE;
				break;
			case 'a':
				firstang = 100;
				break;
			default:
				printf("Ignoring option %c\n", argv[1][i]);
			}
		}
	}
	if (random)
		RandomLines();
	else if (square)
		SquareLines();
	else
		CircleLines();
}
