/*
 *
 * Copyright 1988 by Perihelion Software Ltd.
 * 
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting
 * documentation, and that the name of Perihelion Software Ltd.
 * not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written
 * prior permission.  Perihelios Software Ltd. makes no
 * representations about the suitability of this software
 * for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * Perihelion  Software  Ltd  disclaims all  warranties with
 * regard to this software, including all implied warranties
 * of merchantability  and fitness for a particular purpose.
 * In no event shall Perihelion Software Ltd be  liable  for 
 * any  special,  indirect or  consequential damages  or any
 * damages  whatsoever resulting from loss of use,  data  or
 * profits,  whether in an action of contract, negligence or
 * other  tortious action,  arising out of  or in connection 
 * with the use or performance of this software.
 */

/*****************************************************************

	Blossom registers

	RGM 17 October 1988

	C structure version

	THG 24 October 1988

*****************************************************************/

#ifndef BLOSSOM_H
#define BLOSSOM_H

#define NUM_VREGS 	9				/* Number of video registers */
#define VREG_BASE 	0x40000040			/* Base of video registers */
#define BLOSSOM_BASE	((BlossomPtr)0x40000000)	/* Base of everything */
#define BLITTER_BASE	((BlitterPtr)0x40000080)	/* Base of blitter registers */

typedef struct _Blossom
{
	int	MC;		/* Master control 	*/
	int	EA;		/* Event line address 	*/
	int	TS;		/* Test (leave alone!) 	*/
	int	OB;		/* I/O bits 		*/

	int	T0;
	int	T1;		/* writes as EV 	*/

	int	pad1[ 0xA ];

	int	HT;
	int	HD;
	int	HS;
	int	HW;

	int	VT;
	int	VD;
	int	VS;
	int	VW;
	int	VC;
} Blossom, *BlossomPtr;

typedef struct _Blitter
{
	int	DS;
	int	DF;
	int	DT;
	int	DI;
	int	DW;
	int	DL;

	int	MS;
	int	MB;

	int	SS;
	int	SF;
	int	ST;
	int	SI;

	int	BF;
	int	TC;
	int	PM;
	int	TV;

	int	SD;		/* also SP */
	int	DD;
	int	MD;
	int	BC;

	int 	pad3[ 0x2C ];

	int	LS;
	int	LF;
	int	LT;
	int	LI;
} Blitter, *BlitterPtr;

typedef struct BCreg {
	int	SYC  : 2 ;
	int	SXC  : 2 ;
	int	SPS  : 2 ;
	int	SQ   : 2 ;
	int	DYC  : 2 ;
	int	DXC  : 2 ;
	int	DPS  : 2 ;
	int	DQ   : 2 ;
	int	MPR  : 4 ;
	int	MPS  : 2 ;
	int	EOLS : 1 ;
	int	FCEN : 1 ;
	int	MSK  : 1 ;
	int	DST  : 1 ;
	int     SRC  : 1 ;
	int	STP  : 1 ;
	int	SSH  : 1 ;
	int	SSP  : 1 ;
	int     RESB : 1 ;
	int	GO   : 1 ;
} BCreg ;

#define PlusStride	0
#define PlusCounter	1
#define MinusStride	2
#define MinusCounter	3

extern void BloPlot( int ds, int ss ) ;
extern void BloEmulate( BlitterPtr blo, BCreg *bc ) ;
extern void BCInit( BCreg *bc ) ;

/*****************************************************************/

#define	BCgo	0x80000000
#define	BCresb	0x40000000
#define	BCssp	0x20000000
#define	BCssh	0x10000000

#define	BCstp	0x08000000
#define	BCsrc	0x04000000
#define	BCdst	0x02000000
#define	BCmsk	0x01000000

#define	BCfcen	0x00800000
#define	BCeols	0x00400000

#define	BCmps	0x00100000
#define	BCmpr	0x00010000

#define	BCdq	0x00004000
#define	BCdps	0x00001000
#define	BCdxc	0x00000400
#define	BCdyc	0x00000100
#define	BCsq	0x00000040
#define	BCsps	0x00000010
#define	BCsxc	0x00000004
#define	BCsyc	0x00000001

/* BLOSSOM equivalents of X's GX values */

#define BFclear		0
#define BFand		8
#define BFandReverse	4
#define BFcopy		12
#define BFandInverted	2
#define BFnoop		10
#define BFxor		6
#define BFor		14
#define BFnor		1
#define BFequiv		9
#define BFinvert	5
#define BForReverse	13
#define BFcopyInverted	3
#define BForInverted	11
#define BFnand		7
#define BFset		15

#define RESET_BLITTER(bPtr) \
		(bPtr)->MD = 0xFFFFFFFF; \
		(bPtr)->TC = 1; \
		(bPtr)->MB = 0xFF00; \
		(bPtr)->BC = 0;

#define MAP0	0x00000000
#define MAP1	0x00200000
#define MAP2	0x00400000
#define MAP3	0x00600000

#endif /* BLOSSOM_H */
