/* glocal.h	- Header file for local low level graphics operations
 *
 * Copyright	:	TransTiriel	(c)1990
 *
 * Author	:	M.J. Charnley-Fisher
 *
 * Revisions	:
 *
 *	1.0A	10Mar90	MJCF	Created
 *
 * Description	:
 *
 * The following routines are used to draw and manipulate graphics
 *
 * -------------------------------------------------------------------------------- *
 */
#ifndef __glocal_h
#	define __glocal_h

#ifndef __blossom_h
#	include <blossom.h>
#endif

#ifndef __graphics_h
#	include <graphics.h>
#endif

/* The following defines are used to set various debug options */
#define	WINDDEBUG	0x00000001
#define POLYDEBUG	0x00000002
#define LINEDEBUG	0x00000004
#define COPYDEBUG	0x00000008

/* Defines used by the subsequent macros - fixed constants */
#define BitsPerWord		32
#define BitsPerNibble		4
#define NibblesPerByteShift	1
#define NibbleFractions		2048
#define BlitFractions		4096

/* The pointer to the blitter structure */
extern BlitterPtr	blitter ;

extern TScreen	CurrentScreen ;

/* The following global variables are set up by InitBlossomGraphics */
extern int	NibblesPerPixel,	/* Number of nibbles per pixel */
		NibblesPerPixelShift,	/* Amount to shift to convert */
		NibbleFractionShift,	/* Shift to use to get Nibbles from fraction */
		NibbleFractionMask,	/* Mask to use to extract pixel fraction */
		StrideinNibbles,	/* The stride of the screen in nibbles */
		StrideinNibblesShift,
		StrideinNibblesMask,
		ScreenBaseinNibbles,	/* Base address in nibbles */
		PixelFractionShift,	/* Amount to shift to convert */
		HalfPixelinFractions,	/* Half a pixel in fractions */
		BlitFractionMask ;	/* Mask to use to extract blit fraction */

/* The following are used where graphics are done by the T8 */
extern int	StrideinBytes ,		/* Stride of screen in bytes */
		StrideinBytesShift,	/* Stride in bytes */
		StrideinBytesMask,
		ScreenBaseinBytes,
		ByteFractionMask,	/* Extract bytes from fractions */
		ByteFractionShift ;	/* Amount to shift to get to bytes */

/* The following macros are used to alter between formats */
#define PixelstoFractions( x )	( (x) << PixelFractionShift	)
#define FractionstoPixels( x )	( (x) >> PixelFractionShift	)
#define WholeNibbles( x )	( (x) >> NibbleFractionShift	)

#define YtoNibbAddr( x )	( (x) << StrideinNibblesShift	)
#define NibbAddrtoY( x )	( (x) >> StrideinNibblesShift	)
#define NibbAddrtoX( x )	( (x) &  StrideinNibblesMask	)

#define YtoByteAddr( x )	( (x) << StrideinBytesShift	)
#define ByteAddrtoY( x )	( (x) >> StrideinBytesShift	)
#define ByteAddrtoX( x )	( (x) &  StrideinBytesMask	)
#define WholeBytes( x )		( (x) >> ( NibbleFractionShift + NibblesPerByteShift ) )

#define NibblestoBytes( x )	( (x) >> NibblesPerByteShift	)
#define BytestoNibbles( x )	( (x) << NibblesPerByteShift	)

#define GetBlitFractions( x )	( (x) &  BlitFractionMask	)

#define ExpandColour( x )	{ int n = CurrentScreen.depth ; \
				  while( n < BitsPerWord )  (x) |= ((x) << n), n <<= 1 ; }

/* Used by InitBlit */
#define	TypePolygon	0
#define TypeLine	1
#define TypeRect	2
#define TypeCopy	3

/* Select different low level polygon blit functions */
extern void PolyBlitB( int tlx, int blx, int trx, int brx, int y, int dy, TGCValues * ) ;
extern void PolyBlitT( int tlx, int blx, int trx, int brx, int y, int dy, TGCValues * ) ;
extern void (*PolyBlit)() ;

extern void LineBlitB( int x, int y, int dy, int m, TGCValues * ) ;
extern void LineBlitT( int x, int y, int dy, int m, TGCValues * ) ;
extern void (*LineBlit)() ;

extern void RectBlitB( int x, int y, int w, int h, TGCValues * ) ;
extern void RectBlitT( int x, int y, int w, int h, TGCValues * ) ;
extern void (*RectBlit)() ;

extern void CopyBlitB( int sx, int sy, int w, int h, int dx, int dy, TGCValues * ) ; 
extern void CopyBlitT( int sx, int sy, int w, int h, int dx, int dy, TGCValues * ) ;
extern void (*CopyBlit)() ;

extern int  InitBlit( BlitterPtr, TGCValues *, int type ) ;

/* -------------------------------------------------------------------------------- */
#endif
/* ----------------------------------- glocal.h ----------------------------------- */
