/* graphics.c	window support graphics */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <myvideo.h>

#include "graphics.h"

#define DELAY	50000
#define BRIGHT	150

static CLUT defClut[16] = 	{	{	0, 	~0,	~0,	~0	}, 	/* White */
				{	1,	0,	0,	0	}, 	/* Black */
				{	2,	~0,	0,	0	}, 	/* Red */
				{	3,	0,	~0,	0	}, 	/* Green */
				{	4,	0,	0,	~0	}, 	/* Blue */
				{	5,	0,	~0,	~0	}, 	/* Cyan */
				{	6,	~0,	~0,	0	}, 	/* Yellow */
				{	7,	~0,	0,	~0	}, 	/* Magenta */
				{	8,	~0,	~0,	~0	}, 	/* Foreground White */
				{	9,	BRIGHT,	BRIGHT,	BRIGHT	}, 	/* Grey */
				{	10,	~0,	BRIGHT,	BRIGHT	}, 	/* Light Red */
				{	11,	BRIGHT,	~0,	BRIGHT	}, 	/* Light Green */
				{	12,	BRIGHT,	BRIGHT,	~0	}, 	/* Light Blue */
				{	13,	BRIGHT,	~0,	~0	}, 	/* Light Cyan */
				{	14,	~0,	~0,	BRIGHT	}, 	/* Light Yellow */
				{	15,	~0,	BRIGHT,	~0	} } ; 	/* Light Magenta */
				
static CLUT curClut[16] ;	/* Used to save curClut CLUT when zeroing for alarm etc. */

/* graphics_init() just does any necessary initialisation */

void graphics_init()
{
	cmode = &( vmode[VIDEO_MODE1] ) ;
}


/* clut_writ() loads the appropriate entry in the clut with the red, green, blue values */

void clut_writ(CLUT color )
{
	unsigned char *col  = (unsigned char *)C453WDATA ;
	unsigned char *slot = (unsigned char *)C453WADDR ; 
	
	*slot = color.entry ;
	*col = color.red ;
	*col = color.green ;
	*col = color.blue ;
}


/* clut_read() reads the entry for a clut and returns a structure with the entry */

CLUT clut_read(unsigned char entry )
{
	unsigned char *col = (unsigned char *)C453RDATA ;
	unsigned char *slot = (unsigned char *)C453RADDR ;
	
	CLUT color ;
	
	color.entry = *slot = entry ;
	color.red = *col ;
	color.green = *col ;
	color.blue = *col ;
	
	return color ;
}


/* clut_dflt() loads the first 16 entries with the GEM defClut values */

void clut_dflt()
{
	int i ;
	
	for ( i = 0 ; i < sizeof defClut / sizeof( CLUT ) ; i++ )	clut_writ( defClut[i] ) ;
}


/* clut_load() restores a saved set of entries */

void clut_load()
{
	int i ;
	
	for ( i = 0 ; i < sizeof defClut / sizeof( CLUT ) ; i++ )	clut_writ( curClut[i] ) ;
}


/* clut_save() saves the curClut contents of the clut */

void clut_save()
{
	unsigned char i ;
	
	for ( i = 0 ; i < sizeof defClut / sizeof( CLUT ) ; i++ )	curClut[i] = clut_read( i ) ;
}

/* clut_zero() blanks the clut() */

void clut_zero()
{
	unsigned char i ;
	
	CLUT	color ;
	
	color.red = color.green = color.blue = 0 ;
	
	for ( i = 0 ; i < sizeof defClut / sizeof( CLUT ) ; i++ )
	{
		color.entry = i ;
		clut_writ( color ) ;
	}
}


/*
 * box() simply draws a box at absolute co-ordinates specified by x, y, w, h
 * it draws it in the colour specified by 'col' relative to the base (base)
 *
 * the stride is also passed as an argument
 */
 
void box ( char *base, int stride, int x, int y, int w, int h, char val )
{
	char *addr = base + (y * stride) + x;
	*addr = val;

	bytblt( addr, addr+1, 0,0, 0,1, w-1,1, 0 );
	bytblt( addr, addr+stride, 0,0, 0,stride, h-1,1, 0 );
	
	addr += w - 1;
	bytblt( addr, addr+stride, 0,0, 0,stride, h-1,1, 0 );
	
	addr = base + ( (y+h-1) * stride ) + x;
	bytblt( addr, addr+1, 0,0, 0,1, w-1,1, 0 );
}


/* hvline() is similar to box except that if h = 0 it is vertical and if w = 0 
 * it is horizontal (the hv.. indicates that only 90 are catered for 
 */

void hvline ( char *base, int stride, int x, int y, int w, int h, char val )
{
	char *addr = base + (y * stride) + x;
	*addr = val;

	if ( h == 0 )	bytblt( addr, addr+1, 0,0, 0,1, w-1,1, 0 );
	else 		bytblt( addr, addr+stride, 0,0, 0,stride, h-1,1, 0 );
}


/* byt_blkfill() simply fills the specified block with the appropriate
 * value - it is not limited to graphics.  
 *
 * For example if base = scr.map & stride = 2048 it will fill the 
 * co-ordinates from x to x+width and y to y+height with the colour
 * in value 
 */
 
void byt_blkfill(char *base, int stride, int x, int y, int w, int h, char val )
{
	char *addr;
	int offset;

	/* Poke the int into the first element */
	offset = (y * stride) + x;   
	addr = base + offset;
	*addr = val;
   
	/* Perform the area fill */
	bytblt( base,base, offset,offset+1, 0,1, w-1,1, 0);
	bytblt( base,base, offset,offset+stride, 0,stride, h-1,w, 0);
}


/* byt_line() is an optimised algorithm for drawing lines. It is designed to 
 * take advantage of the bytblt() instruction where possible
 */
 
void byt_line(char *base, int stride, int x1, int y1, int x2, int y2, char col )
{
	char *addr ;
	
	int dx, dy ;
	
	if ( y1 == y2 )							/* horizontal line */
	{
		if ( x2 < x1 )
		{
			dx = x1 - x2 ;
			x1 = x2 ;
		}
		else	dx = x2 - x1 ;
		
		addr = base + stride * y1 + x1 ;
		*addr = col ;

		if ( dx > 1 )	bytblt( addr, addr+1, 0,0, 0,1, dx,1, 0 ) ;
		
		return ;
	}

	else if ( y2 < y1 )
	{
		dy = y1 - y2 ;
		y1 = y2 ;
		
		dx = x1 - x2 ;
		x1 = x2 ;
	}
	else
	{
		dy = y2 - y1 ;
		dx = x2 - x1 ;
	}

	addr = base + stride * y1 + x1 ;

	if ( dx != 0 )
	{
		int s1 ;
		
		if ( dx < 0 )
		{
			dx = -dx ;
			s1 = -1 ;
		}
		else	s1 = 1 ;

		if ( dx != dy )									/* Calculate points */
		{
			int 	i, dx2, dy2 ; 
			int 	err = 2 * dy - dx ,	/* These variables are to keep loop to minimum */
				s2 = stride ;
			int	*chs1 = &s1 ,
				*chs2 = &s2 ;
			
			if ( dy > dx )
			{
				int t = dx ;
				
				dx = dy ;
				dy = t ;
			}
			else
			{
				chs1 = &s2 ;	
				chs2 = &s1 ;
			}

			dx2 = 2 * dx ;
			dy2 = 2 * dy ;
			
			for ( i = 0 ; i < dx ; i++ )						/* Main doer (kept small) */
			{
				*addr = col ;
				
				while ( err >= 0 )
				{
					addr += *chs1 ;
					err -= dx2 ;
				}
				
				addr += *chs2 ;
				err += dy2 ;
			}
		}
		else										/* 45 degree line */ 
		{
			*addr = col ;
			if ( dy > 1 )	bytblt( addr, addr + stride + s1, 0,0, 0,stride + s1, dy,1, 0 ) ;
		}
	}
	else											/* Vertical line */
	{
		*addr = col ;
		if ( dy > 1 ) 	bytblt( addr, addr + stride, 0,0, 0,stride, dy,1, 0 ) ;
	}

}


/* byt_linestyl() is an algorithm for drawing lines with a linestyl. It does not
 * unfortunately take advantage of the bytblt() instruction
 */
 
void byt_linestyl(char *base, int stride, int x1, int y1, int x2, int y2, int d0, int d1, int d2, char col )
{
	int 	dx ,
		dy = y2 - y1 ;
	
	int 	s1, s2, chg, i, err ;
	
	int	dsum1 = d0 + d1 ,
		dtot = d0 + dsum1 + d2 ,
		dsum2 = dtot - d0 ,
		n, rem ; 

	char 	*addr ;
	
	if 	( dy < 0 )
	{
		dy = -dy ;
		y1 = y2 ;

		dx = x1 - x2 ;
		x1 = x2 ;
		s2 = 1 ;
	}
	else	
	{
		if ( dy == 0 )	s2 = 0 ;
		else		s2 = 1 ;
		
		dx = x2 - x1 ;
	}

	if 	( dx < 0 )
	{
		dx = -dx ;
		s1 = -1 ;
	}
	else if ( dx > 0 )	s1 = 1 ;
	else			s1 = 0 ;
	
	if ( dy > dx )
	{
		int t = dx ;
		
		dx = dy ;
		dy = t ;
	
		chg = 1 ;
	}
	else	chg = 0 ;
	
	err = 2 * dy - dx ;
	n = y1 * stride + x1 ;
	rem = n % dtot ;
	addr = base + n ;
	
	for ( i = 0 ; i < dx ; i++ )
	{
		if ( rem < d1 || ( rem >= dsum1  && rem < dsum2 ) )	*addr = col ;
				
		while ( err >= 0 )
		{
			if ( chg == 1 )	addr += s1 ;
			else		addr += ( stride * s2 ) ;
			
			err -= ( 2 * dx ) ;
		}
		
		if ( chg == 1 )	addr += ( stride * s2 ) ;
		else		addr += s1 ;
		
		err += ( 2 * dy ) ;
		
		rem = (++n) % dtot ;
	}
}


/* byt_polyline() is the modified Bresenham algorithm used to calculate the x positions
 * for each line
 *
 * The y position is the index into the 'array' pointed to by 'ptr' offset by ymin
 * (which is calculated in the calling routine
 *
 * Dir controls whether the pointer is moved up by 2 or down by 2
 */
 
short *byt_polyline( short *ptr, int x1, int y1, int x2, int y2, int dir )
{
	if ( y1 == y2 )		/* They're both the same so just store the x-coordinate */
	{
		*ptr = x1 ;
		return 	( ptr ) ;
	}
	else if ( x1 == x2 )	/* The x co-ordinates are the same so just run through the ys storing x */
	{
		if ( dir < 0 )
		{
			int t = y1 ;
			
			y1 = y2 ;
			y2 = t ;
		}
		
		while ( y1++ <= y2 )
		{
			*ptr = x1 ;
			ptr += dir ;
		}
		
		ptr -= dir ;
	}
	else
	{
		int i, s1, s2, y_chg = 0, dx, dy, err, chg, dx2, dy2 ;
		
		if 	( x1 > x2 )	s1 = -1 ;
		else 			s1 = 1 ;
		
		s2 = dir / 2 ;
	
		dx = ( x2 - x1 ) * s1 ;
		dy = ( y2 - y1 ) * s2 ;
		
		if ( dx > dy )
		{
			int t = dx ;
			
			dx = dy ;
			dy = t ;
			chg = 1 ;
		}
		else	chg = 0 ;
		
		err = 2 * dy - dx ;
		dx2 = dx * 2 ;
		dy2 = dy * 2 ;
		
		*ptr = x1 ;
		
		for ( i = 0 ; i <= dx ; i++ )
		{
			/* Depending on the direction store the min/max value of x at a given y */
			while ( y_chg )
			{
				ptr += dir ;
				*ptr = x1 ;
				y_chg -= s2 ;
			}

			if ( dir > 0 )
			{
				if ( x1 < *ptr )	*ptr = x1 ;
			}
			else	if ( x1 > *ptr )	*ptr = x1 ;
				
			while ( err >= 0 )
			{
				/* y changes so mark the y_chg flag */
				if ( chg == 1 )		x1 += s1 ;
				else			y_chg += s2 ;
				
				err -= dx2 ;
			}
			
			if ( chg == 1 )		y_chg += s2 ;
			else			x1 += s1 ;
			
			err += dy2 ;
		}
	}
	
	/* Return the last position of the pointer */
	return ( ptr ) ;
}


/* byt_scanfill() takes a list of edge x co-ordinates and maps them to the screen
 * starting at ymin and ending at ymax 
 *
 * The data is held in the array 'ptr'
 */
	
void byt_scanfill( char *base, int stride, int ymin, int ymax, short *ptr, char col )
{
	int dx ;
	
	char *addr ;

	base += ymin * stride ;
	
	for ( ; ymin <= ymax ; ymin++ )
	{
		addr = base + *(ptr++) ;
		dx = *ptr - *(ptr - 1) ;

		*addr = col ;
		bytblt( addr, addr+1, 0,0, 0,1, dx,1, 0 ) ;

		ptr++ ;
		base += stride ;
	}
}

/* byt_polyfill() is a general purpose polygon filler the only limitation being that
 * polygons with concave sections in the y plane will not behave correctly
 *
 * the data is passed in xy and the number of sides held in 'sides'
 */

void byt_polyfill(char *base, int stride, char col, int sides, int xy[1] )
{
	int i, j, x1, y1, left, dir, pos, xmin, xmax, ymin, ymax ;

	short *scn, *ptr ;
	
	sides *= 2 ;		/* Up the value of sides to the number of co-ordinates */
	y1 = x1 = 1 ;		/* x1 and y1 are used to count the number of equal values (test for a line) */
	
	pos = 0 ;
	xmin = xmax = xy[0] ;
	ymin = ymax = xy[1] ;
	
	/* Find out the minimum and maximum y positions ( & then x if 2 the same ) */
	for ( i = 2 ; i < sides ; i += 2 )
	{
		if ( xy[i] == xy[0] )	x1++ ;
		
		if ( xy[i+1] < ymin )
		{
			pos = i ;
			xmin = xy[i] ;
			ymin = xy[i+1] ;
		}
		else	if ( xy[i+1] == ymin )
		{
			y1++ ;			/* Use y1 to count the number of equal ys */
			if ( xy[i] < xmin )
			{
				pos = i ;
				xmin = xy[i] ;
				ymin = xy[i+1] ;
			}
			
			if ( xy[i] > xmax )	xmax = xy[i] ;
		}
		
		if ( xy[i+1] > ymax )	ymax = xy[i+1] ;
	}
	
	/* If we have a straight line then just do a line */
	if 	( x1 == sides / 2 )	hvline( base, stride, xy[0], ymin, 1, ymax - ymin + 1, col ) ;
	else if ( y1 == sides / 2 )	hvline( base, stride, xmin, ymin, xmax - xmin + 1, 1, col ) ;
	else
	{
		/* Allocate the space for the scan_line offsets */
		scn = malloc( ( ymax - ymin + 1 ) * sizeof(short) * 2 + 4 ) ;
	
		if ( scn == NULL )	
		{
			fputs("byt_polyfill : Insufficient memory\n", stderr ) ;
			return ;
		}
		
		/* We now need to choose which adjacent point is to the left */
		if ( pos == 0 )	i = sides - 2 ;
		else		i = pos - 2 ;
		
		if ( xy[i] < xy[(pos+2) % sides] )	/* The left side is anti-clockwise */
		{
			left = 1 ;
		}
		else				/* Reverse i */
		{
			i = ( pos + 2 ) % sides ;
			left = 0 ;
		}
		
		ptr = scn ;
		dir = 2 ;
		
		x1 = xy[pos] ;	y1 = xy[pos+1] ;
	
		/* For each edge calculate the change in x for each change in y */
		for ( j = 0 ; j < sides ; j += 2 )
		{
			ptr = byt_polyline( ptr, x1, y1, xy[i], xy[i+1], dir ) ;
	
			x1 = xy[i] ;	y1 = xy[i+1] ;
			
			if ( left )
			{
				if ( i == 0 ) 	i = sides - 2 ;
				else		i -= 2 ;
			}
			else	i = ( i + 2 ) % sides ;
			
			if ( xy[i+1] >= y1 )	
			{
				if ( dir == -2 )		ptr-- ;
				dir = 2 ;
			}
			else
			{
				if ( dir == 2 )			ptr++ ;
				dir = -2 ;
			}
		}
		
		byt_scanfill( base, stride, ymin, ymax, scn, col ) ;
		free ( scn ) ;
	}
}


/* byt_polygon() just does a simple polygon using lines without filling */

void byt_polygon( char *base, int stride, char col, int sides, int xy[1] )
{
	int i = 0 ;
	
	while ( sides-- > 1 )
	{
		byt_line( base, stride, xy[i], xy[i+1], xy[i+2], xy[i+3], col ) ;
		i += 2 ;
	}

	/* And the last line */
	byt_line( base, stride, xy[i], xy[i+1], xy[0], xy[1], col ) ;
}


/* clip() simply clips a provided (x,y,w,h) set to a boundary defined by
 * (xmin,ymin,xmax,ymax). It returns a value to indicate which sides were
 * outside the region
 */
 
int clip(int *x, int *y, int *w, int *h, int xmin, int ymin, int xmax, int ymax)
{
	int clipped = 0 ;
	
	if ( *x > xmax )		return ( RIGHT | ALL_OUT );
	if ( *y > ymax )		return ( BOTTOM | ALL_OUT );
	
	if ( *x < xmin )
	{
		*w -= xmin - *x ;
		if ( *w <= 0 )		return ( LEFT | ALL_OUT );
		
		*x = xmin ;
		clipped = LEFT ;
	}
	
	if ( *y < ymin )
	{
		*h -= ymin - *y ;
		if ( *h <= 0 )		return ( TOP | ALL_OUT );
		
		*y = ymin ;
		clipped = clipped | TOP ;
	}
	
	if ( *x + *w > xmax + 1 )
	{
		*w = xmax - *x + 1 ;
		clipped = clipped | RIGHT ;
	}
	
	if ( *y + *h > ymax + 1 )
	{
		*h = ymax - *y + 1 ;
		clipped = clipped | BOTTOM ;
	}
	
	return clipped ;
}


/* clip_pointid() identifies if a point is inside or outside a rectangular
 * area
 */
 
int clip_pointid( x, y, xmin, ymin, xmax, ymax )
float x, y, xmin, ymin, xmax, ymax ;
{
	int result = 0 ;
	
	if 	( x < xmin )		result = CLIP_XMIN ;
	else if ( x > xmax )		result = CLIP_XMAX ;
	
	if 	( y < ymin )		result += CLIP_YMIN ;
	else if ( y > ymax )		result += CLIP_YMAX ;
	
	return( result ) ;
}


/* clip_rect() identifies the interference between two windows
 */
 
int clip_rect( x1, y1, x2, y2, xmin, ymin, xmax, ymax )
int x1, y1, x2, y2, xmin, ymin, xmax, ymax ;
{
	if ( x1 > xmax  ||  x2 < xmin  ||  y1 > ymax  ||  y2 < ymin )
	{
		return( -1 ) ;
	}
	else
	{
		int result = 0 ;
		
		if ( x1 < xmin )		result = CLIP_XMIN ;
		if ( x2 > xmax )		result += CLIP_XMAX ;
	
		if ( y1 < ymin )		result += CLIP_YMIN ;
		if ( y2 > ymax )		result += CLIP_YMAX ;
	
		return( result ) ;
	}
}


/* clip_line() is a faithful reproduction of the Cohen-Sutherland clipping
 * algorithm (documented all over the place)
 *
 *	the line to be clipped runs from (x1,y1) to (x2,y2)
 *	the bounding box is described by xmin, ymin, xmax, ymax
 */

void clip_line( x1, y1, x2, y2, xmin, ymin, xmax, ymax )
float *x1, *y1, *x2, *y2, xmin, ymin, xmax, ymax ;
{
	int i, i1, i2 ;
	float x, y ;

	/* Codify the end points */
	i1 = clip_pointid( *x1, *y1, xmin, ymin, xmax, ymax ) ;
	i2 = clip_pointid( *x2, *y2, xmin, ymin, xmax, ymax ) ;

	while ( i1 != 0  ||  i2 != 0 )
	{
		if ( (i1 & i2) != 0 )		/* Line completely out */
		{
			*x1 = *x2 ;
			*y1 = *y2 ;
			break ;			/* Loop exit point */
		}
		
		i = i1 ;
		if ( i == 0 )	i = i2 ;
		
		if 	( i & CLIP_XMIN )
		{
			y = *y1 + (*y2 - *y1) * (xmin - *x1) / (*x2 - *x1) ;
			x = xmin ;
		}
		else if ( i & CLIP_XMAX )
		{
			y = *y1 + (*y2 - *y1) * (xmax - *x1) / (*x2 - *x1) ;
			x = xmax ;
		}
		else if ( i & CLIP_YMIN )
		{
			x = *x1 + (*x2 - *x1) * (ymin - *y1) / (*y2 - *y1) ;
			y = ymin ;
		}
		else
		{
			x = *x1 + (*x2 - *x1) * (ymax - *y1) / (*y2 - *y1) ;
			y = ymax ;
		}
		
		if ( i == i1 )
		{
			*x1 = x ;
			*y1 = y ;
			i1 = clip_pointid( x, y, xmin, ymin, xmax, ymax ) ;
		}
		else
		{
			*x2 = x ;
			*y2 = y ;
			i2 = clip_pointid( x, y, xmin, ymin, xmax, ymax ) ;
		}
	}
}


/* clip_rectfill() fills a rectangle breaking it down into the number
 * of blocks as necessary
 * The x,y co-ordinate represent the block to be drawn
 * The p,q co-ordinates the block to draw around
 */

void clip_rectfill( base, stride, x1, y1, x2, y2, p1, q1, p2, q2, col )
char *base ;
int stride, x1, y1, x2, y2, p1, q1, p2, q2, col ;
{
	int clip_r ;
	
	/* Identify the interaction between the two rectangles */
	clip_r = clip_rect( x1, y1, x2, y2, p1, q1, p2, q2 ) ;

	if ( clip_r > 0 )
	{	
		if ( clip_r & CLIP_XMIN )
		{
			byt_blkfill( base, stride, x1, y1, p1-x1, y2-y1+1, col ) ;
			x1 = p1 ;
		}

		if ( clip_r & CLIP_XMAX )
		{
			byt_blkfill( base, stride, p2+1, y1, x2-p2, y2-y1+1, col ) ;
			x2 = p2 ;	
		}
		
		if ( clip_r & CLIP_YMIN )
		{
			byt_blkfill( base, stride, x1, y1, x2-x1+1, q1-y1, col ) ;
		}

		if ( clip_r & CLIP_YMAX )
		{
			byt_blkfill( base, stride, x1, q2+1, x2-x1+1, y2-q2, col ) ;
		}
	}
	else if ( clip_r < 0 )
	{
		byt_blkfill( base, stride, x1, y1, x2-x1+1, y2-y1+1, col ) ;
	}
} 


/* regen_id() routine identifies which area of memory may be copied
 * and which part has to be updated 
 *
 * It is primarliy intended for screen operations
 *
 * The source and destination areas are assumed to be set up
 * in the REGEN block pointed to by 'regn'
 *
 * The regn->to.w regn->to.flag, & regn->to.h variables are ignored
 *
 * (xmin,ymin, xmax,ymax) specify the clipping box
 * Source areas outside the clip region which are brought into the
 * clip region are flagged for regeneration
 * 
 * The results are stored back into the REGEN structure
 */

void regen_id(REGEN *regn, int xmin, int ymin,  int xmax, int ymax)
{
	short temp;
	
	/* Initialise the flags in the regen structure */
	regn->from.flag = 0;
	regn->regen1.flag = 0;
	regn->regen2.flag = 0;
	
	/* Find out if the destination box is within the viewport
	 * Clip or return as necessary
	 */
	if ( regn->to.x > xmax || regn->to.y > ymax )	return ;
	
	if ( regn->to.x < xmin )
	{
		if ( regn->to.x + regn->from.w <= xmin )	return ;
		else	
		{
			temp = xmin - regn->to.x;
			regn->to.x = xmin;
			regn->from.w -= temp;
			regn->from.x += temp;
		}
	}
		
	if ( regn->to.y < ymin )
	{
		if ( regn->to.y + regn->from.h <= ymin )	return ;
		else
		{
			temp = ymin - regn->to.y;
			regn->to.y = ymin;
			regn->from.h -= temp;
			regn->from.y += temp;
		}
	}
	
	if ( regn->to.x + regn->from.w > xmax + 1 )	regn->from.w = xmax - regn->to.x + 1;
	if ( regn->to.y + regn->from.h > ymax + 1 )	regn->from.h = ymax - regn->to.y + 1;	

	/* If the source is totally outside the viewport regenerate it all */
	if ( 	(regn->from.x + regn->from.w) <= xmin  ||  regn->from.x > xmax  ||  
		(regn->from.y + regn->from.h) <= ymin  ||  regn->from.y > ymax 	)
	{
		regn->regen1.flag = 1;

		regn->regen1.x = regn->to.x ;
		regn->regen1.y = regn->to.y ;
		regn->regen1.w = regn->from.w ;
		regn->regen1.h = regn->from.h ;
	}
	else						/* Copy as much of the block as possible */
	{
		regn->from.flag = 1 ;
		
		if ( regn->from.x < xmin )				/* source outside left edge */
		{
			regn->regen1.flag = 1;
			
			/* Adjust the widths & set the x co-ordinate */
			regn->regen1.w = xmin - regn->from.x;
			regn->from.w -= regn->regen1.w;
			regn->regen1.x = regn->to.x;

			/* Offset the copy origins */
			regn->from.x = xmin;
			regn->to.x += regn->regen1.w;
			
			/* Set the constant values */
			regn->regen1.y = regn->to.y;
			regn->regen1.h = regn->from.h;
		}
		else if ( regn->from.x + regn->from.w > xmax + 1 )	/* source outside right edge */
		{
			regn->regen1.flag = 1;
			
			regn->regen1.w = regn->from.x + regn->from.w - xmax -1 ;
			regn->from.w -= regn->regen1.w;
			regn->regen1.x = regn->to.x + regn->from.w;
			
			regn->regen1.y = regn->to.y;
			regn->regen1.h = regn->from.h;
		}

		/* Do the same for y */
		if ( regn->from.y < ymin )				/* source outside top edge */
		{
			regn->regen2.flag = 1 ;

			regn->regen2.h = ymin - regn->from.y;
			regn->from.h -= regn->regen2.h;
			regn->regen2.y = regn->to.y;
			
			regn->from.y = ymin;
			regn->to.y += regn->regen2.h;

			regn->regen2.x = regn->to.x;
			regn->regen2.w = regn->from.w;
		}
		else if ( regn->from.y + regn->from.h > ymax + 1 )	/* source outside bottom edge */
		{
			regn->regen2.flag = 1 ;
			
			regn->regen2.h = regn->from.y + regn->from.h - ymax - 1 ;
			regn->from.h -= regn->regen2.h;
			regn->regen2.y = regn->to.y + regn->from.h ;

			regn->regen2.x = regn->to.x;
			regn->regen2.w = regn->from.w;
		}
	}
}


/* byt_blkcopy() copies a block of memory - it handles overlapping areas and
 * squeezes to fit within the specified buffer size
 * 
 * Normally it is used for screen copies but it can be used elsewhere
 */

void byt_blkcopy(char *base, int stride, int x1, int y1, int w, int h, int x2, int y2)
{
	char 	method = ALL_CLR ;

	int	tx = x1,	ty = y1,
		tw = w,		th = h;
		
	int	offset,
		from = y1 * stride + x1,
		to = y2 * stride + x2 ,
		rows = h, cols = w,
		*adjust, *no_change, *basis ;

	if ( from == to ) 	return ;	/* Dont bother with this if they are the same */
	
	/* Find out if the two areas interfere - setting precedence to left-right */
	method = clip( &tx, &ty, &tw, &th, x2, y2, x2+w-1, y2+h-1 );
	
	if ( (method & ALL_OUT) > 0 )		/* No overlap - a straight copy */
	{
		bytblt( base, base, from,to, stride,stride, h,w, 0 );
		return ;
	}
	else if ( (method & RIGHT) > 0 )
	{
		method = NORMAL | BY_COLS ;
		cols = w - tw ;
	}
	else if ( (method & LEFT) > 0 )
	{
		method = REVERSE | BY_COLS ;
		cols = w - tw ;
	}
	else if ( (method & BOTTOM) > 0 )
	{
		method = NORMAL | BY_ROWS ;
		rows = h - th ;
	}
	else
	{
		method = REVERSE | BY_ROWS ;
		rows = h - th ;
	}
	
	/* If we have got this far we have to do an incremental copy so set up which variables to adjust */
	if ( (method & BY_COLS) > 0 )
	{
		offset = cols ;
		basis = &cols ;
		no_change = &h ;
		adjust = &w ;
	}
	else
	{
		offset = rows * stride ;
		basis = &rows ;
		no_change = &w ;
		adjust = &h ;
	}

	/* We need to start from the end and work backwards if it is a REVERSE copy */
	if ( (method & REVERSE) > 0 )
	{
		if ( (method & BY_COLS) > 0 )
		{
			x1 += w - cols ;
			x2 += w - cols ;
		}
		else
		{
			y1 += h - rows ;
			y2 += h - rows ;
		}
		
		from = y1 * stride + x1;
		to = y2 * stride + x2;
		offset = -offset;
	}
	
	/* Now we can do the copy */
	while ( h > 0  &&  w > 0 )
	{
		bytblt( base, base, from, to, stride, stride, rows,cols, 0 );

		*adjust -= *basis ;
			
		if ( *adjust < *basis )
		{
			if ( offset < 0 )	offset = (offset / *basis) * *adjust ;
			*basis = *adjust ;
		}
		
		from += offset ;
		to += offset ;
	}
}

	
/* byt_blkshade() shdaes the block with the appropriate intensity (expressed in %) */

void byt_blkshade( char *base, int stride, int x, int y, int w, int h, char col, char intens )
{
	int i;
	char *addr = base + y * stride + x ;
	
	if ( intens == 0 )	intens = 1 ;
	else			intens = 100 / intens ;

	/* start the pattern */
	*addr = col ;
	
	/* Do the first line */
	bytblt( addr, addr + intens, 0,0, 1,1, w-intens,1, 0 );

	/* If the height is greater than 1 do the next line */
	for ( i = 2 ; i <= h && i <= intens ; i++ )
		bytblt( addr, addr + (i - 1) * stride + i - 1, 0,0, 1,1, w-i,1, 0 );

	/* Now fill up the block */
	if ( i < h )	bytblt( addr, addr + (i-1) * stride, 0,0, stride,stride, h-i,w, 0 ); 
}
