/* gextern.h	- Header file for extern routines to access low level graphics operations
 *
 * Copyright	:	TransTiriel	(c)1990
 *
 * Author	:	M.J. Charnley-Fisher
 *
 * Revisions	:
 *
 *	1.0A	10Mar90	MJCF	Created
 *
 * Description	:
 *
 * The following routines are used to draw and manipulate graphics
 *
 * -------------------------------------------------------------------------------- */

#ifndef __graphics_h
#	define __graphics_h

/* Structure used to send in multiple co-ordinates, it is intended to be identical to
 * the X structure of the same name
 */

typedef struct TPoint {
	short	x ;
	short	y ;
} TPoint ;

typedef int		TID ;
typedef int		TBool ;	
typedef int		TDummy ;

typedef TID 		TDrawable ;
typedef TID		TPixmap ;
typedef TID		TGContext ;
typedef TID		TWindow ;
typedef TID		TColormap ;

typedef unsigned long	TVisualID ;

typedef struct TFont {
	int	dummy ;
} TFont ;

typedef struct TGCValues {
	int		function ;
	unsigned long	plane_mask ;
	unsigned long	fg ;
	unsigned long	bg ;
	int		line_width ;
	int		line_style ;
	int		cap_style ;
	int		join_style ;
	int		fill_style ;
	int		fill_rule ;
	int		arc_mode ;
	TPixmap		tile ;
	TPixmap		stipple ;
	int		ts_x_origin ;
	int		ts_y_origin ;
	TFont		font ;
	int		subwindow_mode ;
	TBool		graphics_exposures ;
	int		clip_x_origin ;
	int		clip_y_origin ;
	TPixmap		clip_mask ;
	int		dash_offset ;
	char		dashes ;
} TGCValues ;

typedef struct TGCstruct {
	TDummy		*ext_data ;
	TGContext	gid ;
	TBool		rects ;
	TBool		dashes ;
	unsigned long	dirty ;
	TGCValues	values ;	
} *TGC ;

typedef struct TScreen {
	TDummy		*ext_data ;
	char		*display ;		/* Base of screen */
	TWindow		root ;
	int		width, height ;
	int		mwidth, mheight ;
	int		depth ;			/* Depth of screen in bits */
	TDummy		*depths ;
	int		stride ;
	TDummy		*root_visual ;
	TGC		default_gc ;
	TColormap	cmap ;
	unsigned long	white_pixel ;
	unsigned long	black_pixel ;
	TDummy		max_maps, min_maps ;
	TDummy		backing_store ;
	TDummy		save_unders ;
	TDummy		root_input_mask ;
} TScreen ;

typedef struct TWindowAttributes {
	int		x, y ;
	int		width, height ;
	int		border_width ;
	int		depth ;
	TWindow		root ;
	int		class ;
	TDummy		bit_gravity ;
	TDummy		win_gravity ;
	int		backing_store ;
	unsigned long	backing_planes ;
	unsigned long	backing_pixel ;
	TDummy		save_under ;
	TColormap	colormap ;
	TBool		map_installed ;
	int		map_state ;
	TDummy		all_event_masks ;
	TDummy		your_event_mask ;
	long		do_not_propogate_mask ;
	TBool		override_redirect ;
	TScreen		*screen ;
} TWindowAttributes ;
	
typedef struct TRectangle {
	short		x, y ;
	unsigned short	width, height ;
} TRectangle ;

/* These are defines used by the various routines */

/* Polygon/line mode */
#define	Origin			0
#define Previous		1

/* Polygon shape */
#define Complex			0
#define Nonconvex		1
#define Convex			2

/* GC Bit function */
#define	Clear			0
#define And			1
#define	AndReverse		2
#define Copy			3
#define AndInverted		4
#define NoOp			5
#define Xor			6
#define Or			7
#define Nor			8
#define Equiv			9
#define Invert			10
#define OrReverse		11
#define CopyInverted		12
#define OrInverted		13
#define Nand			14
#define Set			15

/* GC Line Style */
#define LineSolid		0
#define LineOnOffDash		1
#define LineDoubleDash		2

/* GC Cap Style */
#define CapNotLast		0
#define CapButt			1
#define CapRound		2
#define CapProjecting		3

/* GC Join Style */
#define JoinMiter		0
#define JoinRound		1
#define JoinBevel		2

/* GC Fill Style */
#define	FillSolid		0
#define FillTiled		1
#define FillStippled		2
#define FillOpaqueStippled	3

/* GC Fill-Rule */
#define EvenOddRule		0
#define WindingRule		1

/* GC SubWindow Mode */
#define ClipByChildren		0
#define IncludeInferiors	1

/* GC Clip Mask */
#define ClipNone		0

/* GC Arc-Mode */
#define	Chord			0
#define PieSlice		1

/* Specials to select processor */
#define TransputerT800		0
#define BlossomBlitter		1

/* These functions are Helios supplied routines used by the graphics systems */ 
extern void bytblt( char *, char *, int,int, int,int, int,int, int ) ;

/* These are the graphics functions themselves */
extern void	TDrawPoint( TDrawable, TGC, int x, int y ) ;
extern void	TDrawLine( TDrawable, TGC, int x1, int y1, int x2, int y2 ) ;

extern void	TDrawRectangle( TDrawable, TGC, int x, int y, unsigned width, unsigned height ) ;
extern void	TFillRectangle( TDrawable, TGC, int x, int y, unsigned width, unsigned height ) ;

extern void	TDrawLines( TDrawable, TGC, TPoint *, int npoints, int mode ) ;
extern int 	TFillPolygon( TDrawable, TGC, TPoint *, int npoints, int shape, int mode ) ;
 
extern void	TClearArea( TDrawable, TGC, int x, int y, unsigned width, unsigned height ) ;
extern void	TCopyArea( TDrawable src, TDrawable dst, TGC, int sx, 
						int sy, unsigned w, unsigned h, int dx, int dy ) ;


extern TGC	TCreateGC( TDrawable ) ;
extern void	TFreeGC( TDrawable, TGC *gc ) ;

extern TWindow	TCreateWindow( TWindow par, int x, int y, unsigned w, unsigned h, unsigned bg ) ;
extern TBool	TQueryWindowAttributes( TWindow window, TWindowAttributes *attributes ) ;
extern void	TSetWindowAttributes( TWindow window, TWindowAttributes *attributes ) ;
extern void	TDestroyWindow( TWindow window ) ;
extern void	TClearWindow( TWindow window ) ;

extern TWindow	TOpenDisplay( int mode, int processor ) ;
extern void	TCloseDisplay( TWindow ) ;

/* -------------------------------------------------------------------------------- */
#endif
/* ----------------------------------- graphics.h --------------------------------- */
