#include <stdlib.h>
#include <stdio.h>

#define FRAC_SHIFT	12 	/* 4096 */
#define HALF_FRAC	2048
#define FRACTION	4095
#define ScreenStride	4096
#define NibbStrideShift	12

void polygon( int lx, int lm, int rx, int rm, int y, int dy )
{
	int	lf, ls ;
	int	rf, rs ;

	/* Initialise the x and fractions for both sides */
	lx >>= FRAC_SHIFT ;
	rx >>= FRAC_SHIFT ;
	lf = HALF_FRAC - 1 ;	/* Prevents problems if straight line */
	rf = HALF_FRAC + 1 ;
	
	/* First the left hand side (Note the difference in comparisons to force to go to left) */
	if( lm > 0 )
	{
		if( lm > FRACTION )	lm = lm + ( FRACTION - lm ) / ( dy + 1 ) ;
	
		ls = lm >> FRAC_SHIFT ;
		lm = lm - ( ls << FRAC_SHIFT ) ;
		ls += ScreenStride ;
	}
	else
	{
		if( lm > FRACTION )	lm = lm + ( FRACTION - lm ) / ( dy + 1 ) ;
		
		ls = lm >> FRAC_SHIFT ;
		if( ls ) 		lx += ( 1 - ls ) ;	/* Go to lhs of step */
		
		lm = lm - ( ls << FRAC_SHIFT ) ;
		
		if( lm )
		{
			ls = ScreenStride -ls - 1 ;
			lm = FRACTION - lm ;
		}
		else	ls = ScreenStride - ls ;
	}

	/* And similarly for the right hand side (the ++rm is to force the slope to the right 
	 * to prevent collision with the lhs)
	 */
	if( ++rm >= 0 )
	{
		if( rm > FRACTION )	rm = rm + ( FRACTION - rm ) / ( dy + 1 ) ;
	
		rs = rm >> FRAC_SHIFT ;
		if( rs )		rx += ( rs - 1 ) ;	/* Go to rhs of step */
		 
		rm  = rm - ( rs << FRAC_SHIFT ) ;
		rs += ScreenStride ;
	}
	else
	{
		rm = -rm ;
		if( rm > FRACTION )	rm = rm + ( FRACTION - rm ) / ( dy + 1 ) ;
		
		rs = rm >> FRAC_SHIFT ;
		rm = rm - ( rs << FRAC_SHIFT ) ;
		
		if( rm )
		{
			rs = ScreenStride - rs - 1 ;
			rm = FRACTION - rm ;
		}
		else	rs = ScreenStride - rs ;
	}
	
	y <<= NibbStrideShift ;
	lx += y ;
	rx += y ;
	
	while( dy >= 0 )
	{
/*		if( ( lx >> NibbStrideShift ) == y  ||  dy == 0 ) */ 
			printf( "(%4d,%4d) at %4d\n", 
				lx % ScreenStride, rx % ScreenStride, lx >> NibbStrideShift ) ;

		if( rx < lx  ||  ( rx == lx  &&  rf <= lf ) )
			printf( "Overlap at %4d ( %7d, %7d), li = %d, ri = %d\n", 
				lx >> NibbStrideShift, lx, rx, lm, rm ) ;

		lx += ls ;
		lf += lm ;
		if( lf >= FRACTION )
		{
			lx++ ;
			lf -= FRACTION ;
		}

		rx += rs ;
		rf += rm ;
		if( rf >= FRACTION )
		{
			rx++ ;
			rf -= FRACTION ;
		}
		
		dy-- ;
	} 
}

int main( void )
{
	int		tlx, trx, blx, brx, ty, by ;
	static char	*s, buf[BUFSIZ] = "d" ;

	while( *buf )
	{
		fputs( "> ", stdout ) ;	fflush( stdout ) ;
		gets( buf ) ;
		
		if( *buf )
		{
			s = buf ;
			
			tlx = strtol( s, &s, 0 ) << FRAC_SHIFT ;
			if( s ) trx = strtol( s, &s, 0 ) << FRAC_SHIFT ;
			if( s ) blx = strtol( s, &s, 0 ) << FRAC_SHIFT ;
			if( s ) brx = strtol( s, &s, 0 ) << FRAC_SHIFT ;
			if( s ) ty  = strtol( s, &s, 0 ) ;
			if( s ) by  = strtol( s, &s, 0 ) ;
			
			by  = by - ty ;

			if( !by )	continue ;
			
			blx = ( blx - tlx ) / by ;
			brx = ( brx - trx ) / by ;
			
			polygon( tlx, blx, trx, brx, ty, by ) ;
		}
		else	break ;
	}
}
