/* This modified line drawing algorithm WILL draw lines at any angle - The routine
 * supplied as an example works as the algorithm predicts (The gaps are all due to
 * not switching counters when deltaX is greater than deltaY). However, this algorithm
 * does not support arbitrarly thick lines (neither did the other one really!). As in X,
 * there is lots of sense in separating out the routines to draw 'thin' lines and 'thick'
 * lines.
 */
#include "blossom.h"

/* The following defines are fixed for Mode 1 - Note that shifts are used instead of multiplies
 * because they are faster
 */
#define ScreenWidthinNibbles	12	/* Amount to shift to get fractions */
#define	PixelSize		1	/* Code for Pixel size in nibbles (1 = 8 bits) */

#define ScreenBaseinNibbles	0x00400000	/* Map1 is at 0x00200000 * 2 to get to nibbles */

/* The following defines index the arrays of masks used to control the X-Y counters */
#define PlusStride		0	/* Add stride to current value */
#define PlusCounter		1	/* Increment the current value */
#define MinusStride		2	/* Subtract the stride from the current value */
#define MinusCounter		3	/* Decrease the current counter */

/* These are the arrays used to select a suitable masks for controlling direction of fill.
 * Note that, unlike the transputer, Blossom is able to fill in any direction - hence the
 * four values
 */
static int BC_YDestControl[4] = { 0x00000000, 0x00000101, 0x00000202, 0x00000303 } ;
static int BC_XDestControl[4] = { 0x00000000, 0x00000404, 0x00000808, 0x00000C0C } ;

/* The remainder of the control register is fixed for line drawing so, for performance,
 * elimate the calculations. Note that since the source data is not read there is no need
 * to worry about those bits. Note that the effect of this pattern is to do the following
 *
 *	BCgo		Start Blitter Engine 
 *	BCfcen		Enable Fraction Carry
 *	BCdps+BCsps	Set Destination/Source Pixel Size to 8 bits
 *
 * The other bits used in the old example did (or didn't do!) the following :
 *
 *	BCeols*0	Set the mode to trapezium/line draw (rather than block fill)
 *	BCdxc, BCdyc	See BC_YDestControl & BC_XDestControl above
 *	BCsxc, BCsyc	Same for source (combined with above)
 */
static int BCforLines = BCgo + BCfcen + BCeols*0 + (BCdps + BCsps) * PixelSize ; 

/* Create pointers to the blossom control register */
BlitterPtr blitter = BLITTER_BASE ;
BlossomPtr blossom = BLOSSOM_BASE ;

/* Do our own define for initialising the blitter to what we want. Note the following differences
 * from the original sample code :
 *
 *	MC	Master Control 		- Should already be active.
 *	MB	Mask Bit Control 	- Mask Data not read so no need to change
 *	MD	Mask Data		- As Above
 *	DD	Destination Data 	- (Colour) Argument to DrawThinLine
 *	BC	Blitter Control		- Bits set by DrawThinLine (see above)
 *
 * For your info :
 *
 *	BF	BitWise Function	- default to 0xA - Do no bitwise operation on data
 *	TC	Test Control		- default to 0x1 - Force write (No tests)
 *	PM	Planar Mask		- default to 0xFFFFFFFFL - write all bits
 */
#define InitLineBlit( bptr )	( bptr->BF = 0xA, bptr->TC = 0x1, bptr->PM = 0xFFFFFFFF,\
				  bptr->MB = 0xFF00, bptr->MD = 0xFFFFFFFF, bptr->BC = 0 )

/* -------------------------------------------------------------------------------- */
#include <stdio.h>

char	buf[BUFSIZ] ;

int ask( char *txt )
{
	printf( txt ) ; fflush( stdout ) ;
	gets( buf ) ;
	if( *buf == 'y'  ||  *buf == 'Y' )	return( 1 ) ;
	else					return( 0 ) ;
}

/* ................................................................................ */

static void OldLine( int x1, int y1, int x2, int y2, unsigned char colour )
{
	int nibDstL, subDstL, nibDstR, subDstR ;
	int dx, dy, stride, nibStride, subStride ;
	int DDcolour ;
	unsigned char *pos = (unsigned char *)&DDcolour ;

	int width = 1 ;
	
	*(pos++) = colour ;
	*(pos++) = colour ;
	*(pos++) = colour ;
	*pos = colour ;

#define DstStride	4096	
#define NibblesPerPixel	2

	blossom->MC = 0x4A ;
	InitLineBlit( blitter ) ;
	
	if( y1 == y2 )	return ;
	
	if( y2 < y1 )
	{
		int tmp ;
		tmp = x1 ; x1 = x2 ; x2 = tmp ;
		tmp = y1 ; y1 = y2 ; y2 = tmp ;
	}
	
	nibDstL = 0x00200000 * 2 + y1 * ScreenWidthinNibbles + x1 * 2 ;
	subDstL = 0 ;
	
	if( dx >= 0 )
	{
		stride = dx * DstStride / dy ;
		subStride = stride & 4095 ;
		nibStride = DstStride + (stride/DstStride)*NibblesPerPixel ;
		
		stride = stride * width ;
		nibDstR = nibDstL + (stride/DstStride) * NibblesPerPixel ;
		subDstR = stride & 4095 ;
	}
	else
	{
		stride = (-dx) * DstStride / dy ;
		subStride = DstStride - stride & 4095 ;
		nibStride = DstStride - (stride/DstStride)*NibblesPerPixel ;
		
		stride = stride * width ;
		nibDstR = nibDstL + (stride/DstStride) * NibblesPerPixel ;
		subDstR = stride & 4095 ;
	}
	
	printf("O:DL = %d, LS = %d, LF = %d, LT = %d, LI = %d\nSS = %d, SF = %d, DD = %d, BC = %d\n",
		dy-1, nibDstL, subDstL, nibStride, subStride, nibDstR, subDstR, DDcolour, 
		BCforLines + BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 ) ;
	
	if( ask("Do blit ? >" ) )
	{
		blitter->DL = dy -1 ;
		blitter->LS = nibDstL ;
		blitter->LF = subDstL ;
		blitter->LT = nibStride ;
		blitter->LI = subStride ;
		blitter->SS = nibDstR ;
		blitter->SF = subDstR ;
		blitter->DD = DDcolour ;
		blitter->BC = BCforLines + BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 ;
	}
}

/* -------------------------------------------------------------------------------- */

static void DrawThinLine( int x1, int y1, int x2, int y2, unsigned char colour )
{
	int 		DDcolour ;	/* Used to store the colour (sent to DD) */
	unsigned char 	*ptr ;		/* Used to load the colour into DDcolour */
	int		NibDestAddr ;	/* Destination address in nibbles */
	int		StrideFrac ;	/* Stride Fraction (Solid lines never be a stride in nibbles) */
	int		CountControl ;	/* X and Y control for insertion into BC */
	int		dx, dy ;	/* Absolute differences in x and y */ 

	/* First load colour into DDcolour - dont use a loop - takes longer */
	ptr = (unsigned char *)&DDcolour ;
	*(ptr++) = colour ;
	*(ptr++) = colour ;
	*(ptr++) = colour ;
	*ptr	 = colour ;

	/* Work out the address of the start (x1 + x1 because need address in nibbles == 2 * x1) */
	NibDestAddr = ScreenBaseinNibbles + y1 << ScreenWidthinNibbles + x1 + x1 ;
	
	/* Now decide on the directions and amounts to count for the counters. Whilst this may
	 * look more than the previous example in terms of volume it is actually significantly
	 * more efficient. There is one extra comparison but a massive reduction in the number
	 * of divides, multiplies, and equates
	 */
	if( x1 < x2 )
	{
		dx = x2 - x1 + 1 ;
		
		if( y1 < y2 )
		{
			dy = y2 - y1 + 1 ;
			
			if( dy > dx )
			{
puts("Doing correct one\n" ) ;
				StrideFrac = ( dx << ScreenWidthinNibbles ) / dy ;
				CountControl = BC_XDestControl[PlusCounter] ;
				CountControl |= BC_YDestControl[PlusStride] ;
			}
			else
			{
				StrideFrac = ( dy << ScreenWidthinNibbles ) / dx ;
				CountControl = BC_XDestControl[PlusCounter] ;
				CountControl |= BC_YDestControl[PlusStride] ;
			}
		}
		else
		{
			dy = y1 - y2 + 1 ;
			
			if( dy > dx )
			{
				StrideFrac = ( dx << ScreenWidthinNibbles ) / dy ;
				CountControl = BC_XDestControl[PlusStride] ;
				CountControl |= BC_YDestControl[MinusCounter] ;
			}
			else
			{
				StrideFrac = ( dy << ScreenWidthinNibbles ) / dx ;
				CountControl = BC_XDestControl[PlusCounter] ;
				CountControl |= BC_YDestControl[MinusStride] ;
			}
		}
	}
	else
	{
		dx = x1 - x2 + 1 ;
		
		if( y1 < y2 )
		{
			dy = y2 - y1 + 1 ;
			
			if( dy > dx )
			{
				StrideFrac = ( dx << ScreenWidthinNibbles ) / dy ;
				CountControl = BC_XDestControl[MinusStride] ;
				CountControl |= BC_YDestControl[PlusCounter] ;
			}
			else
			{
				StrideFrac = ( dy << ScreenWidthinNibbles ) / dx ;
				CountControl = BC_XDestControl[MinusCounter] ;
				CountControl |= BC_YDestControl[PlusStride] ;
			}
		}
		else
		{
			dy = y1 - y2 + 1 ;
			
			if( dy > dx )
			{
				StrideFrac = ( dx << ScreenWidthinNibbles ) / dy ;
				CountControl = BC_XDestControl[MinusStride] ;
				CountControl |= BC_YDestControl[MinusCounter] ;
			}
			else
			{
				StrideFrac = ( dy << ScreenWidthinNibbles ) / dx ;
				CountControl = BC_XDestControl[MinusCounter] ;
				CountControl |= BC_YDestControl[MinusStride] ;
			}
		}
	}
	
	printf("N:DL = %d, LS = %d, LF = %d, LT = %d, LI = %d\nSS = %d, SF = %d, DD = %d, BC = %d\n",
		dy-1, NibDestAddr, 0, ScreenWidthinNibbles, StrideFrac, NibDestAddr,
		StrideFrac, DDcolour, BCforLines + BCdxc*1 + BCdyc*0 + BCsxc*1 + BCsyc*0 ) ;
		
	/* Now we can fire off the blitter */
	
	if( ask("Do blit ? >" ) )
	{
	blossom->MC = 0x4A ;				/* Initialise the master control register */
	InitLineBlit( blitter ) ;			/* Initialise the blitter */
	blitter->DL = dy - 1 ;				/* Set the length/height */
	blitter->LS = blitter->SS = NibDestAddr ;	/* Both dest and source are the same */
	blitter->LF = 0 ;				/* It was always zero */
	blitter->LT = ScreenWidthinNibbles ;		/* The nibble stride is always this */
	blitter->LI = blitter->SF = StrideFrac ;	/* Both of these are the same (width = 1) */
	blitter->DD = DDcolour ;			/* Set the colour */
	blitter->BC = BCforLines | CountControl ;	/* Start the whole lot up */
	}
}

/* -------------------------------------------------------------------------------- */

#include <math.h>
#include <stdlib.h>

/* Now back to the original code */
#define	WIDTH	1024
#define HEIGHT	768
#define PI	3.14159265
#define RAD(a)	((a)*(PI/180.0))

/* ................................................................................ */

static void Lines( unsigned char colour )
{
	int i ;
	int xcen = WIDTH/2 ;
	int ycen = HEIGHT/2 ;
	
	for( i = 180 ; i < 360 ; i++ )
	{
		float	dx = cos( RAD((float)i) ) ;
		float	dy = sin( RAD((float)i) ) ;
		int	x1 = xcen + dx * (WIDTH/10.0) ;
		int	y1 = ycen + dy * (HEIGHT/10.0) ;
		int	x2 = xcen + dx * (WIDTH/2.5) ;
		int	y2 = ycen + dy * (HEIGHT/2.5) ;
		
		DrawThinLine( x1, y1, x2, y2, colour ) ;
	}
}

/* ................................................................................ */

int main( void )
{
	char *s = buf ;
	
	*s = 'A' ;
	
	while( *buf != 'q'  &&  *buf != 'Q' ) 
	{
		putchar( '>' ) ; fflush( stdout ) ;
		gets( buf ) ;
		
		s = buf ;
		
		if( *s >= '0'  &&  *s <= '9' )
		{
			unsigned char	col = strtol( s, &s, 10 ) ;
			int		x1  = strtol( s, &s, 10 ) ;
			
			if( x1 )
			{
				int	y1 = strtol( s, &s, 10 ) ;
				int	x2 = strtol( s, &s, 10 ) ;
				int	y2 = strtol( s, &s, 10 ) ;
				
				if( y1  &&  x2  &&  y2 )
				{
					printf( "OldLine( %d, %d, %d, %d, %d )\n", 
									x1, y1, x2, y2, col ) ;
					OldLine( x1, y1, x2, y2, col ) ;

					printf( "DrawThinLine( %d, %d, %d, %d, %d )\n", 
									x1, y1, x2, y2, col ) ;
					DrawThinLine( x1, y1, x2, y2, col ) ;
				}
				else	puts( "Incorrect input, No zero values allowed" ) ;
			}
			else
			{
				printf( "Lines( %d )\n", col ) ;
				Lines( col ) ;
			}
		}
		else puts("Invalid input ..") ;
	}
}

