/* window.c	: Low level routines for initialising and opening windows (+ misc)
 *
 * Copyright	:	TransTiriel	(c)1990
 *
 * Author	:	M.J. Charnley-Fisher
 *
 * Revisions	:
 *
 *	1.0A	10Mar90	MJCF	Created
 *
 * Description	:
 *
 * -------------------------------------------------------------------------------- */

#include <stdio.h>

#include <glocal.h>

/* Address for the blitter */
BlitterPtr	blitter 		= BLITTER_BASE ;

/* The following global variables are set up by InitBlossomGraphics */
int		NibblesPerPixel,	/* Number of nibbles per pixel */
		NibblesPerPixelShift,	/* Amount to shift to convert */
		NibbleFractionShift,	/* Shift to use to get Nibbles from fraction */
		NibbleFractionMask,	/* Mask to use to extract pixel fraction */
		StrideinNibbles,	/* The stride of the screen in nibbles */
		StrideinNibblesShift,
		StrideinNibblesMask,
		ScreenBaseinNibbles,	/* Base address in nibbles */
		PixelFractionShift,	/* Amount to shift to convert */
		HalfPixelinFractions,	/* Half a pixel in fractions */
		BlitFractionMask ;	/* Mask to use to extract blit fraction */

/* The following are used where graphics are done by the T8 */
int		StrideinBytes ,		/* Stride of screen in bytes */
		StrideinBytesShift,	/* Stride in bytes */
		StrideinBytesMask,
		ScreenBaseinBytes,
		ByteFractionMask,	/* Extract bytes from fractions */
		ByteFractionShift ;	/* Amount to shift to get to bytes */

TScreen		CurrentScreen ;		/* Current screen */

static int	BCforTrapeziumFill ;

/* -------------------------------------------------------------------------------- */
#if ( GDEBUG & WINDDEBUG )

/* Routine used to print out global data if debug set */
void PrintStaticGlobalVars( void )
{
	printf( "BitsPerNibble         = %d\n", BitsPerNibble ) ;
	printf( "NibblesPerByteShift   = %d\n", NibblesPerByteShift ) ;
	printf( "BitsPerWord           = %d\n", BitsPerWord ) ;
	printf( "NibbleFractions       = %d\n", NibbleFractions ) ;
	printf( "BlitFractions         = %d\n", BlitFractions ) ;
	putchar( '\n' ) ;
	printf( "ScreenBase            = %#x\n", (int)CurrentScreen.display ) ;
	printf( "ScreenStride          = %d\n", CurrentScreen.stride ) ;
	printf( "ScreenHeight          = %d\n", CurrentScreen.height ) ;
	printf( "ScreenWidth           = %d\n", CurrentScreen.width ) ;
	printf( "ScreenDepth           = %d\n", CurrentScreen.depth ) ;
	putchar( '\n' ) ;
	printf( "NibblesPerPixel       = %d\n", NibblesPerPixel ) ;
	printf( "NibblesPerPixelShift  = %d\n", NibblesPerPixelShift ) ;
	printf( "NibbleFractionMask    = %d\n", NibbleFractionMask ) ;
	printf( "NibbleFractionShift   = %d\n", NibbleFractionShift ) ;
	printf( "StrideinNibbles       = %d\n", StrideinNibbles ) ;
	printf( "StrideinNibblesShift  = %d\n", StrideinNibblesShift ) ;
	printf( "StrideinNibblesMask   = %d\n", StrideinNibblesMask ) ;
	printf( "ScreenBaseinNMibbles  = %#x\n", ScreenBaseinNibbles ) ;
	printf( "PixelFractionShift    = %d\n", PixelFractionShift ) ;
	printf( "HalfPixelinFractions  = %d\n", HalfPixelinFractions ) ;
	printf( "BlitFractionMask      = %d\n", BlitFractionMask ) ;
	putchar( '\n' ) ;
	printf( "StrideinBytes         = %d\n", StrideinBytes ) ;
	printf( "StrideinBytesShift    = %d\n", StrideinBytesShift ) ;
	printf( "StrideinBytesMask     = %d\n", StrideinBytesMask ) ;
	printf( "ByteFractionMask      = %d\n", ByteFractionMask ) ;
	printf( "ByteFractionShift     = %d\n", ByteFractionShift ) ;
}
#endif
/* -------------------------------------------------------------------------------- */

TWindow TOpenDisplay( int mode, int processor )
{
	int i ;
	
	switch( mode )
	{
		case	0 :	/* Mode 0, etc */
		{
			/* Fill in for mode 0 when ready */
		}
		
		default	:	/* 1024 by 768, 8-bit pixels, 256 Colours */
		{
			CurrentScreen.display	= (char*)0x00200000 ;
			CurrentScreen.stride	= 2048 ;
			CurrentScreen.height	= 768 ;
			CurrentScreen.width	= 1024 ;
			CurrentScreen.depth	= 8 ;
		}
	}

	ScreenBaseinBytes = (int)CurrentScreen.display ;
		
	NibblesPerPixel = 1 ;
	for( i = BitsPerNibble ; i < CurrentScreen.depth ; i += BitsPerNibble )
		NibblesPerPixel++ ;
		
	NibblesPerPixelShift = 0 ;
	for( i = NibblesPerPixel ; i > 1 ; i >>= 1 )
		NibblesPerPixelShift++ ;
	
	NibbleFractionMask = NibbleFractions - 1 ;

	NibbleFractionShift = 0 ;
	for( i = NibbleFractions ; i > 1 ; i >>= 1 )
		NibbleFractionShift++ ;
		
	PixelFractionShift 	= NibbleFractionShift + NibblesPerPixelShift ;
	HalfPixelinFractions	= ( 1 << ( PixelFractionShift - 1 ) ) - 1 ;

	StrideinNibbles		= CurrentScreen.stride << NibblesPerPixelShift ;
	StrideinNibblesMask	= StrideinNibbles - 1 ;
	
	StrideinNibblesShift	= 0 ;
	for( i = StrideinNibbles ; i > 1 ; i >>= 1 )
		StrideinNibblesShift++ ;

	ScreenBaseinNibbles	= ScreenBaseinBytes << NibblesPerByteShift ;

	BlitFractionMask	= BlitFractions - 1 ;
	
	StrideinBytes		= StrideinNibbles >> NibblesPerByteShift ;
	StrideinBytesMask	= StrideinBytes - 1 ;
	StrideinBytesShift	= StrideinNibblesShift - NibblesPerByteShift ;
	ByteFractionMask	= ( NibbleFractions << NibblesPerByteShift ) - 1 ;
	ByteFractionShift	= NibbleFractionShift + NibblesPerByteShift ;

	if( processor == TransputerT800 )
	{
		LineBlit	= LineBlitT ;
		PolyBlit	= PolyBlitT ;
		RectBlit	= RectBlitT ;
		CopyBlit	= CopyBlitT ;
	}
	else
	{
		LineBlit	= LineBlitB ;
		PolyBlit	= PolyBlitB ;
		RectBlit	= RectBlitB ;
		CopyBlit	= CopyBlitB ;
	}

	BCforTrapeziumFill 	= BCgo + BCfcen*1 + BCeols*0
				+ ( BCdps + BCsps ) * NibblesPerPixelShift + BCdxc*1
				+ BCdyc*0 + BCsxc*1 + BCsyc*0 + BCdst*1 ;
				
#if ( GDEBUG & WINDDEBUG )
	PrintStaticGlobalVars() ;
#endif
}

/* -------------------------------------------------------------------------------- */

void bytblt2( char *s, char *d, int os, int od, int ss, int ds, int w, int h, int op )
{
	ss = (int)s - ScreenBaseinBytes ;
	os = ByteAddrtoY( ss ) ;
	od = ByteAddrtoX( ss ) ;
	d  = d ;
	ds = ds ;
	h  = h ;
	op = op ;
	 	
	printf("bytblt @ ( %d,%d ) for %d to %d\n", od, os, w, od + w ) ;
}

/* -------------------------------------------------------------------------------- */

TGC	TCreateGC( TDrawable drawable )
{
	static struct TGCstruct gcs ;
	
	drawable = drawable ;
	
	gcs.values.function 	= BFcopy ;
	gcs.values.plane_mask	= 0xFFFFFFFF ;
	gcs.values.fg		= 1 ;
	gcs.values.bg		= 0 ;
	gcs.values.fill_rule	= EvenOddRule ;
	
	return( &gcs ) ;
}

/* -------------------------------------------------------------------------------- */

int	InitBlit( BlitterPtr blitter, TGCValues *values, int type )
{
	type = type ;		/* Currently not used */

	ExpandColour( values->fg ) ;

	blitter->PM = values->plane_mask ;
	blitter->BC = BCssp + BCssh ;
	blitter->BF = values->function ;
	blitter->SD = values->fg ;
	blitter->TC = 0x1 ;
	
	return( BCforTrapeziumFill ) ;
}

/* --------------------------------- window.c ------------------------------------- */ 
/* -------------------------------------------------------------------------------- */
