/* index.h - Header file for TransTiriel Extendible Tree Hashing System
 *
 * Copywrite	: 	TransTiriel		(c) 1989
 * Author	:	M.J. Charnley-Fisher	Nov 1989
 *
 * Revisions	:
 *
 * 1.00A	Nov 89	MJCF	Created for Helios 1.1
 *
 *
 * Description	:
 * Based on ....
 *
 * -------------------------------------------------------------------------------- *
 */
 
typedef HashHdr {
	short		Sector ;	/* Index to appropriate Sector 	*/
	unsigned char	Offset ;	/* Offset within Sector		*/
	char		Flag ;		/* Used for various purposes depending on context */
} HashHdr ;

/* Various pre-defined values for Flag */
static const unsigned char TypeMaskSet	= 0xE0u ;	/* Mask for data types (11100000) */
static const unsigned char TypeMaskClr	= 0x1Fu ;	/* Mask for data types (00011111) */
static const unsigned char TypeTable	= 0x20u ;	/* Hash Table 	       (00100000) */
static const unsigned char TypeList	= 0x40u ;	/* List (4 elements)   (01000000) */
static const unsigned char TypeStack	= 0x60u ;	/* Stack (usually date)(01100000) */
static const unsigned char TypeData	= 0x80u ;	/* Pointer to data     (10000000) */
static const unsigned char TypeIndex	= 0xA0u ;	/* Index to Sectors    (10100000) */
static const unsigned char TypeHdr	= 0xC0u ;	/* Table Header/desc   (11000000) */
static const unsigned char TypeSem	= 0xE0u ;	/* Semaphore	       (11100000) */

static const unsigned char DataFlagMask = 0x18u ;	/* Mask for two flags  (00011000) */
static const unsigned char Flag2DiffSet	= 0x10u ;	/* Set 2 different     (00010000) */
static const unsigned char Flag2DiffClr = 0xEFu ;	/* Used to AND clear   (11101111) */
static const unsigned char Flag2SameSet = 0x08u ;	/* Set 2 the same      (00001000) */
static const unsigned char Flag2SameClr = 0xF7u ;	/* Clear 2 the same    (11110111) */

static const unsigned char CountMaskSet	= 0x07u ;	/* Used to mask counter(00000111) */
static const unsigned char CountMaskClr = 0xF8u ;	/* Used to clear count (11111000) */
static const unsigned char Count01	= 0x01u ;	/* 			00000001  */
static const unsigned char Count02	= 0x02u ;	/*			00000010  */
static const unsigned char Count04	= 0x03u ;	/*			00000011  */
static const unsigned char Count08	= 0x04u ;	/*			00000100  */
static const unsigned char Count16	= 0x05u ;	/*			00000101  */
static const unsigned char Count32	= 0x06u ;	/*			00000110  */
static const unsigned char Count64	= 0x07u ;	/*			00000111  */
 	
typedef struct TableElem {
	HashHdr		Pos[4] ;	/* 4 headers to point to each ASCI code 0-256/4   */
} TableElem ;

typedef TableElem ListElem ;		/* List element has the same structure as a TableElem */

typedef struct DataElem {
	HashHdr		Hdr ;		/* Pointer back to caller */
	HashHdr		Next ;		/* Pointer to next data element */
	Date		Date ;		/* Date element last modified */
	void		*Data ;		/* Pointer/index to actual data */
} DataElem ;

typedef struct IndexElem {
	HashHdr		Hdr ;		/* Type of Sector pointed to by Sector */
	HashHdr		Type ;		/* Pointer to next free element in list */
	Semaphore	*Lock ;		/* Pointer to locking semaphore */
	word		*Sector ;	/* Pointer to associated Sector */
} IndexElem ;

typedef struct HdrElem {
	HashHdr		Hdr ;		/* Describing it as a 
	int		Nentries ;	/* Total Number of entries in table */
	Semaphore	*Lock ;		/* Pointer to Semaphore */
	word		*Sector ;	/* 
typedef struct StackElem {
	HashHdr		Hdr ;		/* Default header structure - back to owner */
	HashHdr		Head ;		/* Head of List of associated elements */
	HashHdr		Tail ;		/* Tail */
	HashHdr		Earth ;		/* Earth (as in Helios) */
} StackElem ;

typedef struct SemElem {
	HashHdr		Hdr ;		/* Header back to owner */
	Semaphore	Sem ;		/* Actual semaphore */
} SemElem ;
