#include <stdio.h>
#include <stdlib.h>

static const int block_size = 1024 ;

#define MAX_BLOCKS	32

void err( char *text )
{
	if ( !text )	text = "Unknown message" ;
	fprintf( stderr, "Error : %s\n", text ) ;
}

int hash_prime( int size )
{
	int *block[MAX_BLOCKS] ;			/* Memory blocks */
	int bp ;					/* Pointer to current block */
	int *p, *lp ;					/* Pointer to next/last free slots */
	int n = 3 ;					/* Start number for primes */
	int nelem = block_size / sizeof( int ) ;	/* Size of each block in integers */
	int hp = 3 ;					/* The current highest prime */

	/* Initialise the block pointers */
	for( bp = 0 ; bp < MAX_BLOCKS ; bp++ )
		block[bp] = NULL ;
	bp = 0 ;
			
	/* Loop until the returned prime is greater than the requested size */
	while( hp <= size )
	{
		/* If we have allocated too many blocks */
		if( bp >= MAX_BLOCKS )
		{
			err("prime : Requested size too big" ) ;
			break ;
		}
		
		/* Or if memory cannot be allocated */
		else if( ( block[bp] = malloc( block_size ) ) == NULL )
		{
			err("prime : Insufficient memory" ) ;
			break ;
		}
		
		else
		{
			p = block[bp] ;
			lp = p + nelem ;
			
			/* If it is the first block, initialise first two slots */
			if( bp++ == 0 )
			{
				*(p++) = 2 ;
				*(p++) = 3 ;
			}
		}
		
		/* Now loop until the prime is reached or we need another block */
		while( p < lp )
		{
			int i = 0 ;		/* Block index */
			int *k = block[i++] ;	/* Pointer to first slot in the block */
			int *lk = k + nelem ;	/* Pointer to last slot in the block */
			
			n += 2 ;		/* Increment the potential number by 2 */
			k++ ;			/* Point to the 2nd element in the block (3) */
	
			/* If the result is 0 not prime - break */
			while( n % *k )
			{
				/* if quotient <= previous prime current is a prime (see Knuth) */
				if( n / *k <= *k )
				{
					*(p++) = hp = n ;
					break ;
				}

				/* Otherwise move to next previous prime, checking block */
				else if( ++k > lk )
				{
					k = block[i++] ;
					lk = k + nelem ;
				}
			}
			
			/* Break out if the requested size is reached */
			if( hp > size )	break ;
		}
	}
	
	/* Free the memory */
	for( bp = 0 ; bp < MAX_BLOCKS ; bp++ )
	{
		if( block[bp] == NULL )	break ;
		else	free( block[bp] ) ;
	}
	
	/* Return the result */
	return( hp ) ;
}


int main( int argc, char *argv[] )
{
	int n = atoi( argv[1] ) ;
	
	argc = argc ;
	printf( "hash_prime( %d ) = %d\n", n, hash_prime( n ) ) ;
}
