#include <stdio.h>
#include <attrib.h>
#include <nonansi.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <time.h>

#define NO_REGS 	8		/* The number of regs */
#define NO_MODES	10		/* The number of modes supported by program */
#define MAX_HMODE	3		/* Maximum hardware mode (used with NO_MODES) */
#define MAX_SMODE	4		/* Maximum mode used by software (Mode 0 B/W) */
#define BIT12MX		4095		/* Maximum value able to be held in 12 bits */
#define BIT08MX		255		/* Maximum value able to be held in 8 bits */
#define REG_BASE	0x40000000	/* Base for register addresses */
#define REG_OFFSET	0x0040/4	/* Offset for video control registers */

/* This line will appear in the Xrc file where the addition was added (at top of file) */
#define INSERT	"# Added by the TransTiriel utility 'ss' on %s"		/* %s is the date */

static struct {
	char  *name ;			/* program name; for error messages */
	short cmode ;			/* if no arguments are provided, default mode */
	short smode ;			/* Start mode */
	char  *xrc ;			/* The name of the Xrc file (defaulted) */
	char  *bak ;			/* Name of backup file */
	char  *rserv ;			/* Server used to identify graphics processor */
	char  buf[BUFSIZ] ;		/* buffer used for various purposes */
	short dx ;			/* offset for digit */
	int   mult ;			/* Multiplication factor (a function of dx) */
	short y	;			/* current Row position (and index to desc array) */
	short msgy ;			/* Row for message */
	char  msgflg ;			/* Flag to indicate if message sent at last request */
	short *cscn ;			/* Pointer to current screen parameters */
	short *sscn ;			/* Pointer to startup screen parameters */
	int   *addr ;			/* Address for registers */
} prg = { "", 1, -1, "/helios/xrc", NULL, "/window", "", 0, 0, 0, NULL, NULL } ;

/* Note that all strings should be the same length & be padded with blanks */
static struct {
	char	*name ;			/* Name for register */
	char	*text ;			/* Description for register */
	short	max ;			/* Maximum value permitted for register */
} desc[] = {	{ "HT",	"Horizontal total (words)", BIT12MX },
		{ "HD", "Horizontal displayed    ", BIT12MX },
		{ "HS",	"Horizontal sync position", BIT12MX },
		{ "HW",	"Horizontal sync width   ", BIT08MX },
		{ "VT",	"Vertical total (rows)   ", BIT12MX },
		{ "VD", "Vertical displayed      ", BIT12MX },
		{ "VS",	"Vertical sync position  ", BIT12MX },
		{ "VW",	"Vertical sync width     ", BIT08MX },
		{ "HM", "Maximum height (forInfo)", BIT12MX },
		{ "VM", "Maximum width  (forInfo)", BIT12MX },
		{ "  ",	"Video MODE              ", NO_MODES-1 } } ;

/* The following definitions force all prints to the same pattern */
#define PRINT_TEXT	"\r\t\t %s : %s = %4d"
#define PRINT_LINE(x) 	PRINT_TEXT, desc[x].name, desc[x].text, prg.cscn[x]
	
/* Used to index respective arrays */
#define HT 	0
#define HD 	1
#define HS 	2
#define HW 	3
#define VT 	4
#define VD 	5
#define VS 	6
#define VW 	7
#define HM 	VW+1		/* HM = Max value for HD ( width in pixels / (pixels/word) ) */
#define VM 	VW+2		/* VM = Max value for VD ( height in pixels ) */
#define CMODE	VW+3		/* For the mode */

#define NO_FLDS	NO_REGS+2	/* for dimensioning array */

static short dflt[NO_MODES+1][NO_FLDS] = { 		/* default register values */
/*		  HT  HD  HS  HW  VT  VD  VS  VW  HM, VM                      */
		{ 208,158,160,16,1100,958,1000,4,160,960 },	/* Mode 0 */
		{ 340,254,272,21, 860,766, 800,4,256,768 },	/* Mode 1 */
		{ 210,158,176,20, 544,478, 500,3,160,480 },	/* Mode 2 */
		{ 800,511,616,88, 544,478, 500,3,512,480 },	/* Mode 3 */
		{ 208,158,160,16,1100,958,1000,4,  0,  0 },	/* Mode 4 (0) */
		{ 340,254,272,21, 860,766, 800,4,  0,  0 },	/* Mode 5 (1) */
		{ 210,158,176,20, 544,478, 500,3,  0,  0 },	/* Mode 6 (2) */
		{ 800,511,616,88, 544,478, 500,3,  0,  0 },	/* Mode 7 (3) */
		{ 208,158,160,16,1100,958,1000,4,  0,  0 },	/* Mode 8 (0) */
		{ 340,254,272,21, 860,766, 800,4,  0,  0 },	/* Mode 9 (1) */
		{   0,  0,  0, 0,   0,  0,   0,0,  0,  0 } } ;	/* From Xrc file */

/* NOTE : The purpose of adding HM and VM is to provide sensible limitations on the
 * values selected by the user. The rules which are followed are as follows :
 *
 *	if ( HM > 0 )
 *	{
 *		0  <= HD <= HM		** The driving limitation
 *		HD <= HT <= 4095
 *		HD <= HS <= HT
 *		0  <= HW <= (HS-HD)
 *	}
 *
 * Similarly for VT, VD, VS, & VW. Modes 0 through 3 are thereforer controlled, others are
 * not (except to the maximum values permitted by the size of the register).
 *
 * The following text will be used in the printf() statement (the same ones will be used 
 * for the vertical values). Note that they appear in pairs and in the same order as
 * the registers (this ordering is used in the program) :
 */

char *messages[] = { 	"%s cannot go below 0, cycled to %d (%s)", 	/* HD < 0	*/
			"%s is limited to %d in this mode by (%s)",	/* HD > HM 	*/
			"%s will not go below %d unless you change %s", /* HT < HD 	*/
			"%s will not go above %d (%s) - cycled",	/* HT > 4095    */
			"%s will not go below %d unless you change %s", /* HS < HD 	*/
			"%s will not go above %d unless you change %s", /* HS > HT	*/
			"%s cannot go below 0, cycled to %d (%s)",	/* HW < 0 	*/
			"%s will not go above %d unless you change %s"};/* HW > (HS-HD) */

static short vals[NO_MODES][NO_REGS] ;		/* actual register values for each mode */

/* Introduction to program (should be terminated by a NULL entry), Note that I would have
 * used Function keys for HT, HD, etc, except for a 'feature' of the Perihelion X terminal
 * emulator which returns user defined strings (even when using raw input!!!)
 */
static char *banner[] = { 	
		" SCREEN SETUP - An ATW Utility by TransTiriel (author M.J. Charnley-Fisher)",
		" ==========================================================================",
		" ",
		" Keys have the following effects ('*' means 'controlled') :",
		"\t0    Mode 0*     1    Mode 1*     2    Mode 2*     3    Mode 3*",
		"\t4    Mode 0      5    Mode 1      6    Mode 2      7    Mode 3",
		"\t8    Mode 0      9    Mode 1                       q(Q) Quit",
		"\tt    Goto HT     d    Goto HD     s    Goto HS     w    Goto HW",
		"\tT    Goto VT     D    Goto VD     S    Goto VS     W    Goto VW",
		"\tUp   Move Up     Down Move Down   Left Move Left   Rght Move Right",
		"\t,(<) Decrease    .(>) Increase    (note effect of Left & Rght)",
		"\ti(I) Initialise current mode      r(R) Restore start-up values",
		" ",
		NULL } ;
		
/* Should the decision be taken to change the above keys the routine get_input() will
 * also have to be changed
 */

/* ================================================================================
 *
 * check_root() looks for the entry <prg.rserv>, gets its context, and checks it
 * against the current processor. If the two are not the same, the program assumes that
 * it is not running on the root processor and transfers itself using the 'remote' command
 */
 
void check_root( int argc, char *argv[] )
{
	Object *object ;
	int error ;

	/* Find the root device */	
	if( ( object = Locate( NULL, prg.rserv ) ) == NULL )
	{
		fprintf( stderr, "%s Unable to locate %s\n", prg.name, prg.rserv ) ;
		exit( 1 ) ;
	}

	/* Get the current machine name */
	if ( ( error = MachineName( prg.buf ) ) < 0 )
	{
		fprintf( stderr, "%s Error getting MachineName\n", prg.name ) ;
		exit( 1 ) ;
	}
	
	/* Check to see if the paths are the same */
	if( strncmp( prg.buf, object->Name, strlen( prg.buf ) ) != 0 )
	{
		int i ;			/* Loop counter */
		char *ptr ;		/* string pointer */
		
		/* Print a message */
		printf("%s Not running on root processor - relocating ...\n\n", prg.name ) ;
		
		/* They are not so relocate - first build command */
		strcpy( prg.buf, "remote " ) ;		/* Initialise */
		strcat( prg.buf, object->Name ) ;	/* Add root processor path */
		ptr = strrchr( prg.buf, c_dirchar ) ;	/* Find last directory delim */
		*ptr++ = ' ' ;				/* Replace with space */
		*ptr = '\0' ;				/* Get rid of rest (ie name) */
		strcat( ptr, prg.name ) ;		/* Add current program name */

		for ( i = 1 ; i < argc ; i++ )		/* Add any arguments */
		{
			strcat( ptr, " " ) ;
			strcat( ptr, argv[i] ) ;
		}
			
		/* Echo command to the screen and send to shell */
		putchar( '\t' ) ;
		puts( prg.buf ) ;
		puts( "Please wait ...\n" ) ;
		system( prg.buf ) ;
		
		/* Exit this version of the program */
		exit( 0 ) ;
	}
	
	/* Otherwise we are OK so just return */
}

/* --------------------------------------------------------------------------------
 *
 * raw_input() puts the console into the correct mode for getting input from the
 * keyboard
 */
 
void raw_input()
{
	Attributes attr ;

	/* Get the current settings for the console */	
	if( GetAttributes( Heliosno( stdin ), &attr ) < 0 )
	{
		fprintf( stderr, "%s Error getting terminal attributes\n", prg.name ) ;
		exit( 1 ) ;
	}
	
	/* Set the appropriate attributes to the required values */
	RemoveAttribute( &attr, ConsoleEcho ) ;
	AddAttribute( &attr, ConsoleRawInput ) ;
	
	/* Implement the changes */
	if( SetAttributes( Heliosno( stdin ), &attr ) < 0 )
	{
		fprintf( stderr, "%s Error setting terminal to raw input\n", prg.name ) ;
		exit( 1 ) ;
	}
}

/* --------------------------------------------------------------------------------
 *
 * handle_args() processes the input arguments - it returns the name of the xrc file
 * (which will usually be /helios/xrc) 
 */
 
void handle_args( int argc, char *argv[] )
{
	int i ;		/* Loop counter */
	int invalid ;	/* Error code */
	
	/* Look at each argument */
	for ( i = 1 ; i < argc ; i++ )
	{
		char *c = argv[i] ;
		
		invalid = 0 ;
		
		if ( *c++ == '-' )		/* Check for option intro */
		{
			switch( *c )
			{
				case 'm' :	/* Set the mode to a start value */
				case 'M' :
				{
					if ( ! *(++c) )
					{
						if ( i < argc )
						{
							if ( *argv[i+1] != '-' )
								prg.smode = atoi( argv[++i] ) ;
							else
								invalid = 'm' ;
						}
						else	invalid = 'm' ;
					}
					else	prg.smode = atoi( c ) ;
					
					prg.smode = abs( prg.smode % NO_MODES ) ;
					prg.sscn  = dflt[prg.smode] ;
					break ;
				}
				
				case 'f' :	/* Change the name of the default file */
				case 'F' :
				{
					if ( ! *(++c) )
					{
						if ( i < argc )
						{
							if ( *argv[i+1] != '-' )
								prg.xrc = argv[++i] ;
							else
								invalid = 'f' ;
						}
						else	invalid = 'f' ;
					}
					else	prg.xrc = c ;
					
					break ;
				}
				
				default :	/* No other valid choices */
				{
					static int printed = 0 ;
					
					if ( *c != 'h'  &&  *c != 'H' )	invalid = 'd' ;
					
					if ( printed )	break ;
				
					printf("Use %s <options>, where options are :\n", 
										argv[0] ) ;
					puts("\t-h              This help message");
					puts("\t-m<mode>        Start-up mode");
					puts("\t-f<pathname>    Name of X startup file (xrc)");
					printed = 1 ;
					break ;
				}
			}
		}
		else	invalid = '-' ;
		
		/* If invalid was set, print an appropriate message */
		switch ( invalid )
		{
			case 'm' : 
				fprintf( stderr, "%s Invalid Mode %s\n", argv[0], argv[i] ) ;
				break ;
			case 'f' :
				fprintf( stderr, "%s Invalid File %s\n", argv[0], argv[i] ) ;
				break ;
			case '-' :
			case 'd' :
				fprintf( stderr, "%s Invalid Option %s\n", argv[0], argv[i] ) ;
				break ;
			default :
				break ;
		}
	}
}

/* --------------------------------------------------------------------------------
 *
 * read_file() reads the supplied XRC file and finds the uppercase MODE and SCREEN entries
 * If the mode has not been overwritten by command line input then it will initialise the
 * choice to the mode specified in this file (if it can find it!). While doing the read it
 * also copies the contents to a back-up file <file>.bak
 */
 
void read_file()
{
	FILE *fi, *fo ;			/* input and output files */
	short mode = -1 ;		/* Tempory store for mode */
	
	/* Try and open the input file */
	if ( ( fi = fopen( prg.xrc, "r" ) ) == NULL )
	{
		fprintf( stderr, "%s, Could not read %s, using dflt\n", prg.name, prg.xrc ) ;
		prg.xrc = NULL ;
		prg.bak = NULL ;
		return ;
	}
	
	/* Allocate space for the backup file name */
	if( ( prg.bak = malloc( strlen( prg.xrc ) + 4 ) ) != NULL )
	{
		char *c1, *c2 ;
		int i ;
		
		strcpy( prg.bak, prg.xrc ) ;

		/* Now add the .bak extension (doing any necessary checks) */
		if ( ( c1 = strrchr( prg.bak, c_dirchar ) ) == NULL )	c1 = prg.bak ;
		else							c1++ ;
	
		if ( ( c2 = strrchr( c1, '.' ) ) != NULL )	*c2 = '\0' ;
	
		if ( ( i = strlen( c1 ) + 4 ) > NAME_MAX )	*(c1+NAME_MAX-4-1) = '\0' ;
		strcat( c1, ".bak" ) ;
	
		/* Now open up this file */
		if ( ( fo = fopen( prg.bak, "w" ) ) == NULL )
		{
			fprintf( stderr, "%s, %s backup not opened\n", prg.name, prg.bak ) ;
			prg.bak = NULL ;
		}
	}
	else
	{
		fprintf( stderr, "%s No memory for backup file\n", prg.name ) ;
		prg.bak = NULL ;
		fo = NULL ;
	}

	/* Now we can start reading the file (and writing to the backup if there is one) */
	printf("%s Reading X resource file %s, please wait ...\n", prg.name, prg.xrc ) ;
	 
	while( fgets( prg.buf, BUFSIZ, fi ) )
	{
		char *c ;

		/* Copy the line to the backup file */		
		if( fo )
		{
			if( fputs( prg.buf, fo ) )
			{
				fprintf( stderr, "%s Error %d writing to %s\n", 
								prg.name, errno, prg.bak ) ;
				clearerr( fo ) ;
				fclose( fo ) ;
				remove( prg.bak ) ;
				prg.bak = NULL ;
				fo = NULL ;
			}
		}
		
		/* If we selected a mode with command line option, read next line */
		if ( prg.smode >= 0 )	continue ;
		
		/* Otherwise get rid of comment */
		if ( ( c = strchr( prg.buf, '#' ) ) != NULL )	*c = '\0' ;

		/* Check for mode */
		if ( ( c = strstr( prg.buf, "MODE" ) ) != NULL )
		{
			c = strchr( c, '=' ) + 1 ;
			mode = abs( atoi( c ) % NO_MODES ) ;
		}
		else if ( ( c = strstr( prg.buf, "SCREEN" ) ) != NULL )
		{
			int i ;
			
			prg.sscn = dflt[NO_MODES] ;
	
			c = strchr( c, '=' ) + 1 ;
			c = strtok( c, " \t," ) ;
			
			for( i = 0 ; i < NO_REGS ; i++ ) 
			{
				if ( c )
				{
					prg.sscn[i] = abs( atoi( c ) % desc[i].max ) ;
					if( prg.sscn[i] == 0 )	c = NULL ;
					else			c = strtok( NULL, " \t," ) ; 
				}
				else	prg.sscn[i] = 0 ;
			}
		}
	}

	/* if there was an error say so */	
	if ( !feof( fi ) )
	{
		fprintf( stderr, "%s Error %d reading %s\n", prg.name, errno, prg.xrc ) ;
		prg.xrc = NULL ;
	}
		
	/* And if the file couldn't be closed say so */
	clearerr( fi ) ;
	if( fclose( fi ) )
	{
		fprintf( stderr, "%s Error %d closing %s\n", prg.name, errno, prg.xrc ) ;
		prg.xrc = NULL ;
	}
	
	/* Likewise, if the backup file is still open */
	if( fo )
	{
		if ( fclose( fo ) )
		{
			fprintf( stderr, "%s Error %d closing %s\n", 
							prg.name, errno, prg.bak ) ;
			remove( prg.bak ) ;
			prg.bak = NULL ;
		}
		
		if ( prg.xrc == NULL )
		{
			remove( prg.bak ) ;
			prg.bak = NULL ;
		}
	}
	
	/* Finally update the mode and screen parameters, if applicable */
	if ( mode > 0  &&  prg.smode < 0 )
	{
		int i ;
		short *from ;
		
		prg.smode = mode ;
		
		/* Just add any non-zero parameters */
		from = dflt[mode] ;
		for( i = 0 ; i < NO_FLDS ; i++ )
			if( prg.sscn[i] <= 0 )	prg.sscn[i] = from[i] ;
	}
}

/* --------------------------------------------------------------------------------
 *
 * redraw() redraws the whole screen. Since it is effectively in full screen mode 
 * it has to control where the cursor is
 */
 
void redraw()
{
	int i ;
	
	/* initialise the co-ordinate system */
	prg.y = 0 ;
	
	/* Print the menu and banner */
	printf("\f\n\x9b%dm",7);
	i = 0 ;
	while( banner[i] )
	{
		puts( banner[i++] ) ;
		if( i == 2 ) printf("\x9b%dm",0);
	}
	
	/* Print the mode */
	printf("\x9b%dm",7);
	printf( PRINT_TEXT, desc[CMODE].name, desc[CMODE].text, prg.cmode ) ;
	printf("\x9b%dm\n", 0 ) ;
	
	/* Now print the values for each register, recording the position.
	 * The register itself also needs to be updated
	 */
	for( i = 0 ; i < NO_REGS ; i++ )
	{
		*(prg.addr+i) = prg.cscn[i] ;
		printf( PRINT_LINE(i) ) ;
		putchar('\n') ;
		prg.y++ ;
	}

	
	/* Remember the line number for messages */
	prg.msgy = prg.y ;
	
	/* Move the cursor to the first position and initialise the necessary variables */
	prg.dx = 1 ;
	prg.mult = 1 ;
	printf("\x9b%dA", prg.y);
	prg.y = 0;
	printf( PRINT_LINE(0) ) ;
	printf( "\x9b%dD",1 ) ;	fflush( stdout ) ;
}

/* --------------------------------------------------------------------------------
 *
 * move() is responsible for moving the cursor about the screen
 */
 
void move( int dirn, int reg )
{
	if ( dirn >= 0 )
	{
		switch( dirn )
		{
			case 'A' : 	if( --prg.y >= 0 )
						printf( "\x9b%dA",1 ) ;
					else
						prg.y++ ;
					break ;
					
			case 'B' : 	if( ++prg.y < NO_REGS )
						printf( "\x9b%dB",1 );
					else
						prg.y-- ;
					break ;
					
			case 'C' :	if ( prg.mult > 1 )
					{
						prg.mult /= 10 ;
						prg.dx-- ;
						printf( "\x9b%dC",1 ) ;
					}
					break ;
					
			case 'D' :	if ( prg.mult < desc[prg.y].max )
					{
						prg.mult *= 10 ;
						prg.dx++ ;
						printf( "\x9b%dD",1 ) ;
					}
					break ;

			default :	break ;
		}
		
		while( prg.mult > desc[prg.y].max  &&  prg.mult > 1 )
		{
			prg.mult /= 10 ;
			prg.dx-- ;
			printf( "\x9b%dC",1 ) ;
		}
	}
	else
	{
		int y = prg.y ;
		
		prg.y = reg % NO_REGS ;
		
		( y < prg.y ) ? printf("\x9b%dB",prg.y-y) : printf("\x9b%dA",y-prg.y );
		printf( PRINT_LINE( prg.y) ) ;
		printf( "\x9b%dD",1 ) ;
		prg.mult = 1 ;
		prg.dx = 1 ;
	}
	
	fflush( stdout ) ;	
}	

/* --------------------------------------------------------------------------------
 *
 * message() puts a message at the bottom of the screen
 */
 
void message( char *text )
{
	int diff = prg.msgy - prg.y ;

	/* set the flag to indicate a message sent */
	if( *text )	prg.msgflg = 1 ;
	else		prg.msgflg = 0 ;		
	
	printf("\r\x9b%dB\x9bK\t%s", diff, text ) ;
	printf("\x9b%dA", diff ) ;
	printf( PRINT_LINE( prg.y ) ) ;
	printf( "\x9b%dD", prg.dx ) ;
	fflush( stdout ) ;
}

/* --------------------------------------------------------------------------------
 *
 * check() is responsible for checking that the values do not go outside limits
 * provided as input. It is also the routine which actually updates the register
 */
 
void check( reg, pos, min, max, minmsg, maxmsg, minreg, maxreg, flag, mult )
int reg, pos, min, max, flag, mult ;
char *minmsg, *maxmsg, *minreg, *maxreg ;
{
	short n = prg.cscn[reg] ;

	if ( n < min )
	{
		if( reg == pos)	
		{
			n += ( ( max - n ) / mult ) * mult ;

			if( minmsg )
			{
				sprintf( prg.buf, minmsg, desc[reg].name, 
							(flag>0)?max:min, minreg ) ;
				message( prg.buf ) ;
			}
			else	message( "Cycled through minimum" ) ;
		}
		else	n = min ;
	}
	else if ( n > max )
	{
		if( reg == pos)
		{
			n -= ( ( n - min ) / mult ) * mult ;
		
			if( maxmsg )
			{
				sprintf( prg.buf, maxmsg, desc[reg].name, 
							(flag<0)?min:max, maxreg ) ;
				message( prg.buf ) ;
			}
			else	message( "Cycled through maximum" ) ;
		}
		else	n = max ;
	}
	
	if ( n != prg.cscn[reg]  ||  reg == pos )
	{
		int *addr ;
		
		/* Update the structure and the register */
		prg.cscn[reg] = n ;
		addr = prg.addr + reg ;
		*addr = n ;
		
		/* Update the screen */
		if( reg < pos )
		{
			int n = pos-reg;
			
			printf("\x9b%dA",n);
			printf( PRINT_LINE( reg ) );
			printf("\x9b%dB",n);
		}
		else if ( reg > pos )
		{
			int n = reg-pos;
			printf("\x9b%dB",n);
			printf( PRINT_LINE( reg ) ) ;
			printf("\x9b%dA",n);
		}
		else	printf( PRINT_LINE( reg ) ) ;
		
		/* Move the cursor back */
		printf("\x9b%dD", prg.dx ) ;
	}
}		
	
/* --------------------------------------------------------------------------------
 *
 * adjust() alters the value in the register
 */
 
void adjust( int dirn )
{
	int base, bmax ;
	
	prg.cscn[prg.y] = prg.cscn[prg.y] + ( dirn * prg.mult ) ;
	
	/* Work out whether we are Horizontal or Vertical */
	if ( prg.y <= HW )	base = HT ;
	else			base = VT ;
	
	bmax = HM+base/VT ;
	
	/* Check to see if the limits rules apply */
	if( dflt[prg.cmode][bmax] > 0 )
	{
		int reg = prg.y ;
		int mult = prg.mult ;

		if( reg == base+1 )
		{
			check(	reg,			prg.y,	
				0, 			dflt[prg.cmode][bmax],
				messages[0],		messages[1],
				desc[bmax].name,	desc[bmax].name,  1,mult ) ;
			
/*			reg = base ; */
			mult = 1 ;
		}
		
		if( reg == base )
		{
			check(  reg,			prg.y,
				prg.cscn[base+1],	desc[reg].max,
				messages[2],		messages[3],
				desc[base+1].name,	"Register Limit",  0,mult ) ;
				
/*			reg = base+2 ; */
			mult = 1 ;
		}
		
		if( reg == base+2 )
		{
			check(	reg,			prg.y,
				prg.cscn[base+1],	prg.cscn[base],
				messages[4],		messages[5],
				desc[base+1].name,	desc[base].name,   0,mult );
			
/*			reg = base+3 ; */
			mult = 1 ;
		}
		
		if( reg == base+3 )
		{
			char *msgvar ;
			
			if ( base == HT ) 	msgvar = "(HS-HD)" ;
			else			msgvar = "(VS-VD)" ;
			
			check(  reg,			prg.y,
				0,			prg.cscn[base+2] - prg.cscn[base+1],
				messages[6],		messages[7],
				msgvar,			msgvar,		   1,mult ) ;
		}
		
	}
	else	check( prg.y, prg.y, 0, desc[prg.y].max, 0,0,0,0, 0, prg.mult ) ;

	/* Finally flush all output */
	fflush( stdout ) ;
}
	
/* --------------------------------------------------------------------------------
 *
 * get_input() is the main loop for getting information from the user
 */
 
void get_input()
{
	int i, c, loop ; 
	
	loop = 1 ;
	
	redraw() ;
	
	while( loop )
	{
		c = getchar() ;
		
		if( prg.msgflg ) 	message("") ;	/* Clear the message */
		
		if( c >= '0'  &&  c <= '9' )
		{
			prg.cmode = c - '0' ;
			prg.cscn = vals[prg.cmode] ;
			redraw() ;
		}
		else	switch( c )
		{
			case '\x9b' :	move( getchar(), -1 ) ;	break ;
			case '<'    :
			case ','    :	adjust( -1 ) ;		break ;
			case '>'    :
			case '.'    :	adjust( +1 ) ;		break ;
			case 'q'    :
			case 'Q'    :	putchar('\f') ;
					loop = 0 ;		break ;
			case 't'    :	move( -1, HT ) ;	break ;	
			case 'd'    :	move( -1, HD ) ;	break ;	
			case 's'    :	move( -1, HS ) ;	break ;	
			case 'w'    :	move( -1, HW ) ;	break ;	
			case 'T'    :	move( -1, VT ) ;	break ;	
			case 'D'    :	move( -1, VD ) ;	break ;	
			case 'S'    :	move( -1, VS ) ;	break ;	
			case 'W'    :	move( -1, VW ) ;	break ;	
			case 'i'    :
			case 'I'    :	for( i = 0 ; i < NO_REGS ; i++ )
						prg.cscn[i] = dflt[prg.cmode][i] ;
					redraw() ;
					break ;
			case 'r'    :	prg.cmode = prg.smode ;
					prg.cscn = vals[prg.cmode] ;
					for( i = 0 ; i < NO_REGS ; i++ )
						prg.cscn[i] = prg.sscn[i] ;
					redraw() ;
					break ;
			default	    :	message("Key not recognised" ) ;
					break ; 
		}
	}
}

/* --------------------------------------------------------------------------------
 *
 * write_file() is actually reponsible for updating the file
 */
 
void write_file( char *date )
{
	FILE *fi, *fo ;
	int i ;
	
	printf( "Updating the file %s, please wait ...\n", prg.xrc ) ;

	/* Open the files */	
	if ( ( fi = fopen( prg.bak, "r" ) ) == NULL )
	{
		fprintf( stderr, "%s Could not re-open file %s\n", prg.name, prg.bak ) ;
		return ;
	}
	
	if ( ( fo = fopen( prg.xrc, "w" ) ) == NULL )
	{
		fprintf( stderr, "%s Could not re-open file %s\n", prg.name, prg.xrc ) ;
		fclose( fi ) ;
		return ;
	}
	
	/* Print the new data to the fresh xrc file */
	fprintf( fo, INSERT, date ) ;
	if ( !ferror( fo ) ) 	fprintf( fo, "MODE  \t=\t%d\n", prg.cmode ) ;
	if ( !ferror( fo ) ) 	fputs( "SCREEN\t=\t", fo ) ;
	for( i = 0 ; i < NO_REGS ; i++ )
	{
		if ( !ferror( fo ) )	fprintf( fo, "%d ", prg.cscn[i] ) ;
		else break ;
	}
	if ( !ferror( fo ) )	fputs( "\n\n", fo ) ;
	
	/* Now transfer to contents of the backup file */
	while( fgets( prg.buf, BUFSIZ, fi )  &&  !ferror( fo ) )
	{
		char *c ;
		
		/* Check for upper case MODE or SCREEN */
		if( ( c = strstr( prg.buf, "MODE" ) ) != NULL )
		{
			for( i = 0 ; i < 4 ; i++, c++ )
				*c = tolower( *c ) ;
		}
		else if( ( c = strstr( prg.buf, "SCREEN" ) ) != NULL )
		{
			for( i = 0 ; i < 6 ; i++, c++ )
				*c = tolower( *c ) ;
		}
		
		/* Write to the output file */
		fputs( prg.buf, fo ) ;
	}
	
	/* If there was an error say so */
	if( !feof( fi ) )
		fprintf( stderr, "%s Error %d reading %s\n", prg.name, ferror(fi), prg.bak ) ;
		
	if( ferror( fo ) )
		fprintf( stderr, "%s Error %d writing %s\n", prg.name, ferror(fo), prg.xrc ) ;
		
	/* Try and close the files */
	if( fclose( fo ) )
	{
		fprintf( stderr, "%s Error %d closing %s\n", prg.name, errno, prg.xrc ) ;
		fprintf( stderr, "Please restore original from %s\n", prg.bak ) ;
	}
	
	if( fclose( fi ) )
	{
		fprintf( stderr, "%s Error %d closing %s\n", prg.name, errno, prg.bak ) ;
		fprintf( stderr, "Check %s, you may need to use floppy backup", prg.xrc) ;
	}
}

/* --------------------------------------------------------------------------------
 *
 * update_file() is responsible for getting user feedback about updating the file
 */
 
void update_file()
{
	int changes = 0 ;
	
	/* Check for any changes */
	if( prg.cmode == prg.smode )
	{
		int i ;
		
		for( i = 0 ; i < NO_REGS ; i++ )
		{
			if( prg.cscn[i] != prg.sscn[i] )
			{
				changes = 1 ;
				break ;
			}
		}
	}
	else	changes = 1 ;
	 
	if( changes  &&  prg.xrc  &&  prg.bak )
	{
		int c, i ;
		time_t timer ;
		char *date ;
		
		puts("\nYou may now update the X resource file. Note that a backup") ;
		puts("file (with a .bak extension) will retain the old data until" ) ;
		puts("you run this program again\n" );
		
		printf(" Enter <y> to update X resource file %s > ", prg.xrc ) ;
		fflush( stdout ) ;
		
		c = toupper( getchar() ) ;
		printf("%c\n", c ) ;
		
		if( c != 'Y' )
		{
			printf("\n\n Please confirm (y) > " ) ;	fflush( stdout ) ;
			c = toupper( getchar() ) ;
			printf("%c\n", c ) ;
			if( c == 'Y' )	return ;
		}
		
		/* Set up the date */
		if( time( &timer ) != -1 )	date = ctime( &timer )  ;
		else				date = "Unknown Date\n" ;

		if ( prg.cmode > MAX_SMODE )	prg.cmode %= (MAX_HMODE+1) ;
		
		printf("\nThe following lines will be added to %s :\n\n\t", prg.xrc ) ;

		printf( INSERT, date ) ;
		printf( "\tMODE  \t=\t%d\n", prg.cmode ) ;
		fputs( "\tSCREEN\t=\t", stdout ) ;
		for( i = 0 ; i < NO_REGS ; i++ )
			printf("%d ", prg.cscn[i] ) ;
		puts("\n");
		
		if( prg.cmode != prg.smode )
		{
			puts("Note that the new mode is different to the startup mode");
			printf("Strart-up Mode = %d, new Mode = %d\n", prg.smode, prg.cmode ) ;
		}
		
		printf("\n Please confirm that you wish to add these changes (y) > " ) ;
		fflush( stdout ) ;
		
		c = toupper( getchar() ) ;
		printf("%c\n", c ) ;
		
		if( c == 'Y' )	write_file( date ) ;
		else		puts("\nChanges ignored\n");
		
	}
	else
	{
		if ( ! changes )
			puts("There were no changes, therefore no file updates\n");
		else	
			puts("There are no files to update due to an error early on\n" ) ;
	}
}
		
/* ================================================================================
 *
 * The main() entry point. Input arguments are taken in the form -<char><option>
 */
 
int main( int argc, char *argv[] )
{
	int i,j ;

	/* Clear the screen */
	putchar ( '\f' ) ;	
	
	/* Initialise the prog name for errors and the register location */
	prg.name = argv[0] ;
	prg.addr = (int*)REG_BASE + REG_OFFSET ;
		
	/* Make sure we are running on root processor */ 
	check_root( argc, argv ) ;

	/* Handle the input arguments */
	handle_args( argc, argv ) ;
	
	/* Read and copy the XRC file to XRC.BAK */
	read_file() ;

	/* copy the defaults across to the working copy */
	for( j = 0 ; j < NO_MODES ; j++ )
	{
		for( i = 0 ; i < NO_REGS ; i++ )
			vals[j][i] = dflt[j][i] ;
	}

	/* Initialise the Mode, copy over parameters if not defaults */
	if ( prg.smode >= 0 )
	{
		prg.cmode = prg.smode ;
		prg.cscn = vals[prg.cmode] ;
		
		for( i = 0 ; i < NO_REGS ; i++ )
			prg.cscn[i] = prg.sscn[i] ;
	}
	else
	{
		prg.smode = prg.cmode ;
		prg.sscn = dflt[prg.smode] ;
		prg.cscn = vals[prg.cmode] ;
	}
	
	/* Get user input - go raw first */
	raw_input() ;
	get_input() ;
	
	/* Now update the Xrc file */
	update_file() ;
	
	/* Finish */
	printf("Finishing ....\n" ) ;
}
