/* ------------------------------------------------------------------------ */

 * fastmath.c - TransTiriel

 *

 * Modified by M.J. Charnley-Fisher for Helios and the Transputer Jul 89

 */



/*_ math.c	18-Oct-85	Walter Bright	*/

/* Copyright (C) 1985 by Northwest Software	*/

/* All Rights Reserved				*/



/* Algorithms from Cody and Waite, "Software Manual For The Elementary	*/

/* Functions".								*/





#define A0		-0.64124943423745581147e+2f

#define A1		 0.16383943563021534222e+2f

#define A2		-0.78956112887491257267e+0f



#define B0		-0.76949932108494879777e+3f

#define B1	 	 0.31203222091924532844e+3f

#define B2		-0.35667977739034646171e+2f

#define B3	 	 0.10000000000000000000e+1f



#define C1		 355.0f/512.0f

#define C2 		-2.121944400546905827679e-4f



#define SQRTPNT5	 0.70710678118654752440f

#define LOG10E		 0.43429448190325182765f



static int sign_mask = 0x00000001L << 31 ;



float alog(float y)

{ 

	float z, x=y ;

  	int n;



	if ( y<=0 )

	{

		return( -1E37 ) ;

	}

	else

	{

		union {

			float		f ;

			unsigned char 	c[4] ;

			int		i ;

		} u ;

		int sign_bit ;

		

		u.f = x ;

		sign_bit = u.i & sign_mask ;

		u.i <<= 1 ;

		n = u.c[3] - 126 ;

		u.c[3] -= n ;

		u.i >>= 1 ;

		u.i |= sign_bit ;

		z = u.f ;

	}

  	

  	x = z - 0.5f ;

  	if (z > SQRTPNT5)

  	{

  		x -= 0.5f ;

  	}

  	else

  	{

  		z = x ;

  		n-- ;

  	}



	z = x /(z * 0.5f + 0.5f);

  	x = z * z;

  	z += z * x * (((A2*x+A2)*x+A1)*x+A0)/((((B3*x+B3)*x+B2)*x+B1)*x+B0);

  	x = n;

  	x = (x * C2 + z) + x * C1;

  	return (x);

}



float alog10( float x )

{

	return( _alog( x ) * LOG10E ) ;

}



